#!/usr/bin/perl
#
###################################################################################
#	
#	LASTPIECE_P (Local Alignment-STate Probabilities that Insertion-type and dEletion-type gaps Co-Exist, type P):
#	    A package of programs to compute the probabilities of gapped segments in each of which an insertion-type gap and a deletion-type gap co-exist, (which were referred to as case-(iv) gapped segments by Ezawa (2016a?)),
#	    under a stochastic model of sequence evolution with biologically realistic insertions/deletions.
#	   (Written almost exclusively in Perl.)
#
#	Version 0.3: Copyright (C) 2020 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
#
###################################################################################
#
# swap_indexes_A.pl:
#	This servant script swaps one of the indexes for labelling each row of the input data
#	with the index for labelling each input file.
#	It was created by taking advantage of some parts of the framework of 
#	another servant script, 'graft_convolution.pl'.
#	[CREATED on 2020/02/22.]
#
#	[NOTE:] The script assumes that the row label is of the form, "\d+:\d+", where "\d+" represents a string of decimal digits.
#	and that it is separated from the data fields by a tab ("\t"). 
#
# graft_convolution.pl:
#	This servant script performs the "graft-convolution" of TWO multiplication factors of the form:
#	muP_{1st or 2nd} [ (L_F, t_F)  | (L_I, t_I) ],
#	that the number of sites is L_I at the initial time (=t_I) and L_F at the final time (=t_F),
#	with 1 <= L_I, L_F <= $UB_CT_SITES.
#	Here, the "graft-convolution" means the following operation:
#	muP_{Output} [ (L_F, t_F) | (L_I, t_I) ]
#	     = \sum_{Lm = 1}^{$UB_CT_SITES} \int_{t_I}^{t_F} d{}tm  [
#	    muP_{1st} [ (Lm, tm)  | (L_I, t_I) ] * muP_{2nd} [ (L_F, t_F)  | (Lm, tm) ] 
#		].
#	The script was created by taking advantage of the main framework of another servant script,
#	'cmpt_mu_Dcr2Adel.pl' (in "./2nd_ii/mu_Dcr2Adel/").
#	[CREATION STARTED on 2020/02/14, ENDED on 2020/02/dd.]
#	

	##### (IMPORTED w/ substantial modifications from '../../graft_convolution.pl' on 2020/02/22) (HD & CL & CHK) #####

use strict;
use warnings;

use MyFile_manipu qw(
	open_input_file2
	open_output_file
	list_sessions_ext
	);


$| = 1; # Flush the buffer for standard output as soon as it receives data.

my $bin_perl = "/usr/bin/perl"; # The Perl interpreter. # (Added on 2020/01/29). #

my $program_path = $0;
my ($program_name) = ($program_path =~ m!([^/]+)$!);

	# Controlling the level of test outputs. #
	# (0: silent; 1: very quiet; 2: quiet; 3: medium; 4: verbose; 5: very verbose) 
our $TEST_LEVEL = 2;
##our $TEST_LEVEL = 5;


	####### Take in the Command-Line Arguments. #######


		# Modified in creation of 'swap_indexes_A.pl' (on 2020/02/22) (CL) #

unless (@ARGV > 6) {
	print STDERR "\nUsage:  [perl] $program_name  {label_for_table_corner} {{1,2}_when_selecting_rows_w/_the_specified_{1st,2nd}_row_index} {specified_value_of_the_row_index} {path/to/the/input/directory} {prefix_for_input_file_names} {suffix_for_input_file_names}  {path/to/the/output/file} ";
	print STDERR "\n\n";
	die "$!\n\n";
}
my ($label_corner, $which_row_index, $spec_value_row_index, $path_indir, $pref_infile, $suff_infile, $path_ofile) = @ARGV[0 .. 6];

=For the predecessor, 'graft_convolution.pl'. (CL)
unless (@ARGV > 7) {
	print STDERR "\nUsage:  [perl] $program_name {initial_time (non-negative number)} {final_time (positive number, greater than initial_time)} {number_of_sub_time_intervals (positive integer)} {upper-bound_of_number_of_sites} {input_file_for_indel_parameters} {input_file_for_muPs_2B_graft-convoluted_1st} {input_file_for_muPs_2B_graft-convoluted_2nd}  {output_file} ";
	print STDERR "\n\n";
	die "$!\n\n";
}
our ($TIME_INIT, $TIME_FIN, $CT_SB_INTVLS, $UB_CT_SITES_CMPT) = @ARGV[0 .. 3];
my ($infile_id_prms, $infile_muPs_in_1st, $infile_muPs_in_2nd, $ofile_muPs_out) = @ARGV[3+1 .. 7];
=cut # END of "For the predecessor, 'graft_convolution.pl'. (CL)"

		# END of "Modified in creation of 'swap_indexes_A.pl' (on 2020/02/22) (CL)" #


unless (($which_row_index =~ m/^\d+$/) and (($which_row_index == 1) or ($which_row_index == 2)) ) {
	print STDOUT "[ERROR '$program_name'] ILLEGAL INPUT: \$which_row_index = '$which_row_index', while it HAS TO BE '1' or '2' : $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}


	### Check the existence & permission mode of the input directory. ###

unless ( ( -d $path_indir ) and ( -r $path_indir ) ) {
	print STDOUT "[ERROR '$program_name'] CanNOT FIND a readable input directory, \"$path_indir\": $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}

	### Retrieve the labels of all the input files. ###

my ($list_flabels) =  list_sessions_ext ($path_indir, $pref_infile, $suff_infile, '-');
unless (defined $list_flabels) {
	print STDOUT "[ERROR '$program_name'] The subroutine, 'list_sessions_ext', FAILED on the input directory, \"$path_indir\": $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}

my @std_list_flabels = sort { $a <=> $b } @{$list_flabels} ; # Assume that the file labels are numbers (especially integers). #

my $ct_infiles = @std_list_flabels;

	### Check the existence & permission mode of the input files. ###

foreach my $flabel (@std_list_flabels) { # Preliminary foreach-loop (over file-labels). #
	my $infile = join ('', $path_indir, '/', $pref_infile, $flabel, $suff_infile);

    unless ( ( -f $infile ) and ( -r $infile ) ) {
	print STDOUT "[ERROR '$program_name'] CanNOT FIND a readable input file, \"$infile\" (w/ the label, \"$flabel\"): $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
    }
}  # End of the preliminary foreach-loop (over file-labels). #

	##### END of "(IMPORTED w/ substantial modifications from '../../graft_convolution.pl' on 2020/02/22) (HD & CL & CHK)" #####


	### Declare the Beginning of the Script ###
	##### (IMPORTED w/ substantial modifications from '../../graft_convolution.pl' on 2020/02/22) (DC) #####

print STDOUT "\n<<< Program name: $program_name (path: $program_path) >>>\n\n";

print STDOUT "Label for the Table-Corner:  \$label_corner = '$label_corner' .\n\n";

print STDOUT "\$which_row_index = $which_row_index;\n";
print STDOUT " this means that the rows with {the ", (($which_row_index == 1) ? '1st' : '2nd' ), " row-index} = \$spec_value_row_index (=$spec_value_row_index)  will be selected as input.\n\n";


print STDOUT "\n<< Input Files (#{files} = $ct_infiles): >>\n\n";

print STDOUT ' ' x 3, join ('', $path_indir, '/', $pref_infile, '{file-label}', $suff_infile), " ,\n";
print STDOUT "  where {file-label} = ", join (', ', @std_list_flabels), " .\n\n";


print STDOUT "\n<< Output File: >>\n\n"; 

print STDOUT ' ' x 3, "$path_ofile .\n\n\n";

print STDOUT "\n<< Global Variables specifying the levels of tests: >>\n\n"; # Added on 2020/01/30. (DC2) #

print STDOUT "\n", '$TEST_LEVEL = ', $TEST_LEVEL, " .\n\n";

	##### END of "(IMPORTED w/ substantial modifications from '../../graft_convolution.pl' on 2020/02/22) (DC)" #####



	###  Main Process: 
	###  Read in the input files & Select rows w/ the specified value of the index,
	###  Swap the specified index w/ the file-label,
	### & Output the results.
	### (ADDED in creation of 'swap_indexes_A.pl' on 2020/02/22). (MAIN) ###

$/ = "\n"; # Set the input-record-separator to be the newline character.

print STDOUT "\n<<<<<  Main Process: Read in the input files & Select rows w/ the specified value of the index, Swap the specified index w/ the file-label, & Output the results >>>>>\n\n\n";

		# (Preliminary) Open a file-handle for outputting to the output file. #

my $fh_out = open_output_file ($path_ofile, 0);
unless (defined $fh_out) {
	print STDOUT "[ERROR '$program_name'] The subroutine 'open_output_file' FAILED to open a file-handle for the file \"$path_ofile\": $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}


for (my $k = 0; $k < $ct_infiles; $k++) { # Main outer for-loop (over input files). #


    my $flabel = $std_list_flabels[$k];

    my $infile = join ('', $path_indir, '/', $pref_infile, $flabel, $suff_infile);

   (print STDOUT "\n< The $k th input file (w/ label = '$flabel'; {path/to/the/input/file} = $infile) >\n\n") if ($TEST_LEVEL > 1);

    my $fh_in = open_input_file2 ($infile);
    unless (defined $fh_in) {
	print STDOUT "[ERROR '$program_name'] The subroutine 'open_input_file2' FAILED to open a file-handle for the $k th input file, \"$infile\": $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
    }

    if ($k == 0) { # For the 1st input file... # 

	while (my $line = <$fh_in>) { # Preceding inner while-loop (to read in the field-label line). #
	    if ($line =~ m/^\#/) {
		$line =~ s/^[^\t\n]+//;

			# output the specified corner-label
			# along with the labels for the fields (i.e., columns). #
		print $fh_out $label_corner, $line;
		print $fh_out "\n";
		last;
	    }
	} # End of the Preceding inner while-loop (to read in the field-label line). #

    } # End of "if ($k == 0) { # For the 1st input file... #  ...}"

    while (my $line = <$fh_in>) { # Main inner while-loop (over input lines). #

	my ($indx1, $indx2) = ($line =~ m/^(\d+)\:(\d+)/);

	unless (defined $indx2) { next; } # Skip non-data lines. #

	my $indx_for_slct = ($which_row_index == 1) ? $indx1 : $indx2;

	if ($indx_for_slct == $spec_value_row_index) {

		# Use ONLY those lines with the specified value of the row index. #

	    $line =~ s/^[^\t\n]+//; # Remove the old row-label. #

	    my $row_label = ($which_row_index == 1) ? join (':', $flabel, $indx2) : join (':', $indx1, $flabel) ; # Create a new row-label. #

	    print $fh_out $row_label, $line ;
	}

## my ($label_corner, $which_row_index, $spec_value_row_index, $path_indir, $pref_infile, $suff_infile, $path_ofile) = @ARGV[0 .. 6];

    } # End of the Main inner while-loop (over input lines). #

    close ($fh_in);

} # End of the Main outer for-loop (over input files). #


close ($fh_out);

print STDOUT "\n\n === All The Processes of '$program_name' FINISHED !! ===\n\n\n";

	### END of "(ADDED in creation of 'swap_indexes_A.pl' on 2020/02/22). (MAIN)" ###


