#!/usr/bin/perl
#
###################################################################################
#	
#	LASTPIECE_P (Local Alignment-STate Probabilities that Insertion-type and dEletion-type gaps Co-Exist, type P):
#	    A package of programs to compute the probabilities of gapped segments in each of which an insertion-type gap and a deletion-type gap co-exist, (which were referred to as case-(iv) gapped segments by Ezawa (2016a?)),
#	    under a stochastic model of sequence evolution with biologically realistic insertions/deletions.
#	   (Written almost exclusively in Perl.)
#
#	Version 0.3: Copyright (C) 2020 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
#
###################################################################################
#
# cmpt_muP0.pl:
#	This servant script computes the multiplication factors,
#	mu_{P0} [ (L, t) | (L_I, t_I) ],
#	of the site-count (L) at a time (t), conditioned on the second site-count (L_I) at the initial time (t_I),
#	under the "base" rate operator (\hat{Q}^{ID}_0 (i; t)) in the new perturbative method.
#	It was created by taking advantage of the main framework of another servant script,
#	'cmpt_muP_case_i2iii.pl' (in "../muP_case_i2iii/") .
#	[CREATION STARTED on 2020/02/03, ENDED on 2020/02/04.]
#
#	[NOTE: The multiplication factors, mu_{P0} [ (L, t) | (L_I, t_I) ],
#		(and those that appear in the subsequent sessions (except the 'PreFinal' ones)
#		DIFFER from the multiplication factors of full gapped segments
#		(labeled case-(i), (ii), (iii), and (iv))
#		in that they are NOT re-scaled by (the reciprocal of) the corresponding "no-indel" probabilities.
#		In this sense, they are more like simple conditional probabilities.]
#


	### (IMPORTED w/ modifications from '../muP_case_i2iii/cmpt_muP_case_i2iii.pl' on 2020/02/03) (HD & CL & CHK) ###
		### (IMPORTED w/ modifications from '../../../lastpiece.alpha.pl' on 2020/01/30) (HD) ###

use strict;
use warnings;

use MyFile_manipu qw(
	open_input_file2
	open_output_file
	list_sessions_ext
	);

use MyLASTPIECE_IO qw(
	readin_indel_params
	readin_indel_rates
	print_muP0s
	);
	# print_muP_case_i
	# print_muP_case_ii_or_iii 
	#
	# Added 'readin_indel_params' & 'readin_indel_rates' on 2020/01/31. #
	# Added 'print_muP_case_i' & 'print_muP_case_ii_or_iii' on 2020/02/01. # => Removed on 2020/02/03.
	# Added 'print_muP0s' on 2020/02/03.

##use MyLASTPIECE_Sppl qw(
##	cal_delta_exp_del_5p2
##	);


##sub create_dir_rw (\$); # Added on 2020/01/29.


$| = 1; # Flush the buffer for standard output as soon as it receives data.

my $bin_perl = "/usr/bin/perl"; # The Perl interpreter. # (Added on 2020/01/29). #

my $program_path = $0;
my ($program_name) = ($program_path =~ m!([^/]+)$!);

	# Controlling the level of test outputs. #
	# (0: silent; 1: very quiet; 2: quiet; 3: medium; 4: verbose; 5: very verbose) 
our $TEST_LEVEL = 2;
#our $TEST_LEVEL = 5;

$MyLASTPIECE_IO::TEST_LEVEL = $TEST_LEVEL ; # Added on 2020/01/27. # RESURRECTED on 2020/02/04.


		### END of "(IMPORTED w/ modifications from '../../../lastpiece.alpha.pl' on 2020/01/30) (HD)" ###

our $EPS = 1E-07; # A "small" number to be used for checking the consistency. # (Added on 2020/02/01.) #


	####### Take in the Command-Line Arguments. #######

		# Modified in creation of 'cmpt_muP0.pl' (on 2020/02/03) (CL1) #
##unless (@ARGV > 8) {
##	print STDERR "\nUsage:  [perl] $program_name {initial_time (non-negative number)} {final_time (positive number, greater than initial_time)} {number_of_sub_time_intervals (positive integer)} {upper-bound_of_number_of_sites} {input_file_for_indel_parameters} {input_file_for_raw_indel_rates} {output_file_for_case_i} {output_file_for_case_ii} {output_file_for_case_iii}";
#
unless (@ARGV > 7) {
	print STDERR "\nUsage:  [perl] $program_name {initial_time (non-negative number)} {final_time (positive number, greater than initial_time)} {number_of_sub_time_intervals (positive integer)} {upper-bound_of_number_of_sites} {SIGMA_BE (= 0,1, or 2)} {input_file_for_indel_parameters} {input_file_for_raw_indel_rates} {output_file}";
		# End of "Modified in creation of 'cmpt_muP0.pl' (on 2020/02/03) (CL1)" #
	print STDERR " .\n\n";
	##print STDERR " [ options ] ,\n";
	##print STDERR "   where the [options] can consist of any (or none) of the following:\n\n";
	##print STDERR "            --MIN_WD_SIZE=integer      for the minimum window size for detecting purge-candidates (default=$MIN_WD_SIZE_PCAND);\n";
	##print STDERR "           --XLEVEL=integer       for the level of MSA space exploration (from =1 (performing only multiple-\"shift\"s) to =5 (performing all programed moves) (default=$XLEVEL);\n"; # ADDED on 2019/08/01.
	##print STDERR "\n"; # ADDED in REVISION (on 2019/10/18). (ARGV2) #

	die "$!\n\n";
}

		# Modified in creation of 'cmpt_muP0.pl' (on 2020/02/03) (CL2) #
##our ($TIME_INIT, $TIME_FIN, $CT_SB_INTVLS, $UB_CT_SITES_CMPT) = @ARGV[0 .. 3];
##my ($infile_id_prms, $infile_id_rates, $ofile_muP_case_i, $ofile_muP_case_ii, $ofile_muP_case_iii) = @ARGV[4 .. 8];
our ($TIME_INIT, $TIME_FIN, $CT_SB_INTVLS, $UB_CT_SITES_CMPT, $SIGMA_BE) = @ARGV[0 .. 3+1];
my ($infile_id_prms, $infile_id_rates, $ofile_muP0) = @ARGV[5 .. 7];
		# End of "Modified in creation of 'cmpt_muP0.pl' (on 2020/02/03) (CL2)" #

	####### END of "Take in the Command-Line Arguments." #######


	### Check the existence & permission mode of the input files. ###
	### (Added on 2020/02/03.) ###

foreach my $infile ($infile_id_prms, $infile_id_rates) {
    unless ( ( -f $infile ) and ( -r $infile ) ) {
	print STDOUT "[ERROR '$program_name'] CanNOT FIND a readable input file, \"$infile\": $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
    }
}
	### End of "(Added on 2020/02/03.)" ###

	### END of "(IMPORTED w/ modifications from '../muP_case_i2iii/cmpt_muP_case_i2iii.pl' on 2020/02/03) (HD & CL & CHK)" ###


	### Declare the Beginning of the Script ###
	### (Selectively IMPORTED w/ modifications from '../../../lastpiece.alpha.pl' on 2020/01/30) (DC) ###
	### (=> IMPORTED w/ modifications from '../muP_case_i2iii/cmpt_muP_case_i2iii.pl' on 2020/02/03) (DC) ###

print STDOUT "\n<<< Program name: $program_name (path: $program_path) >>>\n\n";

			# (Newly added on 2020/01/30) (DC1) #

print STDOUT "\n<< Input Files: >>\n\n";

print STDOUT " (for indel parameters) $infile_id_prms ,\n";
print STDOUT " (for raw indel rates)  $infile_id_rates .\n\n";

	# Modified in creation of 'cmpt_muP0.pl' (on 2020/02/03).
##print STDOUT "\n<< Output Files: >>\n\n";
##
##print STDOUT " (for muP of case-i segments)  $ofile_muP_case_i ,\n";
##print STDOUT " (for muP of case-ii segments) $ofile_muP_case_ii ,\n";
##print STDOUT " (for muP of case-iii segments) $ofile_muP_case_ii .\n\n";
#
print STDOUT "\n<< Output File: >>\n\n"; 

print STDOUT " (for muP0 under \"base\" rate operator)  $ofile_muP0 .\n\n";
	# End of "Modified in creation of 'cmpt_muP0.pl' (on 2020/02/03)".

			# End of "(Newly added on 2020/01/30) (DC1)" #

print STDOUT "\n<< Global Variables specifying the levels of tests: >>\n\n"; # Added on 2020/01/30. (DC2) #

print STDOUT "\n", '$TEST_LEVEL = ', $TEST_LEVEL, " .\n\n";

print STDOUT '$MyLASTPIECE_IO::TEST_LEVEL  = ', $MyLASTPIECE_IO::TEST_LEVEL , " .\n\n"; # Added on 2020/02/04.

			# Added on 2020/02/01. #
print STDOUT "\n<< A \"small\" number to be used for checking the consistency: >>\n\n"; #
print STDOUT '$EPS = ', $EPS, " .\n";
			# End of "Added on 2020/02/01." #


print STDOUT "\n<< Parameters necessary for recursive computations of the multiplication factors: >>\n\n";

print STDOUT '($TIME_INIT, $TIME_FIN) = ', "($TIME_INIT, $TIME_FIN) .\n\n";
print STDOUT '#{Sub-time-intervals} = ', $CT_SB_INTVLS, " .\n\n";

print STDOUT '{Upper-bound of #{sites} for computation} = ', $UB_CT_SITES_CMPT, " .\n\n"; 

print STDOUT '$SIGMA_BE = ', $SIGMA_BE, " .\n\n"; # Added in creation of 'cmpt_muP0.pl' (on 2020/02/03). #

	### End of "(Selectively IMPORTED w/ modifications from '../../../lastpiece.alpha.pl' on 2020/01/30) (DC) ###
	### (=> IMPORTED w/ modifications from '../muP_case_i2iii/cmpt_muP_case_i2iii.pl' on 2020/02/03) (DC)" ###


	##### (IMPORTED from '../muP_case_i2iii/cmpt_muP_case_i2iii.pl' on 2020/02/03) (Pre 00) #####

	## (MODIFIED on 2020/01/31) (00-IN) ##

print STDOUT "\n<<< Preliminary Process 00-IN: Reading in the quantities necessary for the computations >>>\n\n";

	## 00-IN-A: Read in the input file for indel parameters. ##

print STDOUT "\n<< (Pre 00-IN-A): Reading in the indel model parameters >>\n\n";

our ($TOTRATE_INS, $TOTRATE_DEL, $CUTOFF_INS_LEN, $CUTOFF_DEL_LEN)
   = readin_indel_params ($infile_id_prms);

unless (defined $CUTOFF_DEL_LEN) {
	print STDOUT "[ERROR '$program_name'] The subroutine, 'readin_indel_params', FAILED on the input file, \"$infile_id_prms\" $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}
(print STDOUT "\n ... The subroutine, 'readin_indel_params', Ended on the input file, \"$infile_id_prms\" ...\n\n") if ($TEST_LEVEL > 1);

##my ($infile_id_prms, $infile_id_rates, $ofile_muP_case_i, $ofile_muP_case_ii, $ofile_muP_case_iii) = @ARGV[4 .. 8];

print STDOUT "\n[ Indel Model Parameters retrieved: ]\n\n";

print STDOUT '($TOTRATE_INS, $TOTRATE_DEL) = ', "($TOTRATE_INS, $TOTRATE_DEL) .\n";
print STDOUT '($CUTOFF_INS_LEN, $CUTOFF_DEL_LEN) = ', "($CUTOFF_INS_LEN, $CUTOFF_DEL_LEN) .\n\n";


	## Preliminary 00-B: Read in the input file for raw indel rates. ##

print STDOUT "\n<< (Pre 00-IN-B): Reading in the raw indel rates >>\n\n";

our ($rf_RATES_INS, $rf_RATES_DEL, $rf_RATES_INS_BD, $rf_RATES_DEL_BD)
    = readin_indel_rates ($infile_id_rates);

unless (defined $rf_RATES_DEL_BD) {
	print STDOUT "[ERROR '$program_name'] The subroutine, 'readin_indel_rates', FAILED on the input file, \"$infile_id_rates\" $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}
(print STDOUT "\n ... The subroutine, 'readin_indel_rates', Ended on the input file, \"$infile_id_rates\" ...\n\n") if ($TEST_LEVEL > 1);


if ($TEST_LEVEL > 2) {

    print STDOUT "\n[ Raw Indel Rates retrieved: ]\n\n";

    print STDOUT join ("\t", 'Length', 'Ins', 'Del', 'Ins(bd)', 'Del(bd)'), "\n\n";

    my $maxlen = 0;
    foreach my $rf_rates ($rf_RATES_INS, $rf_RATES_DEL, $rf_RATES_INS_BD, $rf_RATES_DEL_BD) {
	my $len = @{$rf_rates} - 1;
	if ($maxlen < $len) { $maxlen = $len; }
    }

    for (my $l=1; $l <= $maxlen; $l++) {
	my @out_rates = ();
	foreach my $rf_rates ($rf_RATES_INS, $rf_RATES_DEL, $rf_RATES_INS_BD, $rf_RATES_DEL_BD) {
	    my $rate = $rf_rates->[$l];
	    unless (defined $rate) { $rate = 'undef'; }
	    push @out_rates, $rate;
	}
	print STDOUT join ("\t", $l, @out_rates), "\n";
    }
    print STDOUT "\n\n";

} # End of "if ($TEST_LEVEL > 2) {...}"


=OBSOLETE as of 2020/01/31. (00-IN)
	## Preliminary 00-A: Read in the input file for indel parameters. ##

print STDOUT "\n<< Indel Model Parameters retrieved: >>\n\n";

print STDOUT '$TOTRATE_INS = ', $TOTRATE_INS, " ,\n";
print STDOUT '$TOTRATE_DEL = ', $TOTRATE_DEL, " .\n\n";

	## Preliminary 00-B: Read in the input file for raw indel rates. ##

=cut # End of "OBSOLETE as of 2020/01/31. (00-IN)"

	## End of "(MODIFIED on 2020/01/31) (00-IN)" ##


	## Preliminary 00-C: Compute Other Important Parameters ##
	### (IMPORTED w/ modifications from '../../../lastpiece.alpha.pl' on 2020/01/30) (00-C) ###

print STDOUT "\n<<< Preliminary Process 00-C: Computing Other Important Parameters >>>\n\n";

our $TOTRATE_INDEL = $TOTRATE_INS + $TOTRATE_DEL;
our $SIZE_FULL_INTVL = $TIME_FIN - $TIME_INIT;
our $SIZE_SB_INTVL = $SIZE_FULL_INTVL /$CT_SB_INTVLS;

		# Modified on 2020/01/31. (00-C-1) #
our @TIME_LAPSES_CMPT = (); # Values of the time lapse (= t_F - t_I) at which the factors will be computed.
for (my $i = 0; $i <= $CT_SB_INTVLS; $i++) {
	$TIME_LAPSES_CMPT[$i] = $i * $SIZE_SB_INTVL;
} 
=OBSOLETE as of 2020/01/31. (A)
our @TIME_LAPSES = (); # Values of the time lapse (= t_F - t_I) to be examined.
for (my $i = 0; $i <= $CT_SB_INTVLS; $i++) {
	$TIME_LAPSES[$i] = $i * $SIZE_SB_INTVL;
} 
=cut # End of "OBSOLETE as of 2020/01/31. (A)"
		# End of "Modified on 2020/01/31. (00-C-1)" #

=DISABLED on 2020/01/31. (1)
our @E_INDEL_CTS = (); # Expected counts of indels. #
foreach my $tlapse (@TIME_LAPSES) { push @E_INDEL_CTS, $TOTRATE_INDEL * $tlapse; }
=cut # End of "DISABLED on 2020/01/31. (1)"

print STDOUT '$TOTRATE_INDEL = ', $TOTRATE_INDEL, " \;\n";
print STDOUT '$SIZE_FULL_INTVL = ', $SIZE_FULL_INTVL, " \;\n";
print STDOUT '$SIZE_SB_INTVL = ', $SIZE_SB_INTVL, " .\n\n";

		# Modified on 2020/01/31. (00-C-2) #
print STDOUT '@TIME_LAPSES_CMPT = (', join (', ', @TIME_LAPSES_CMPT), ") .\n\n";
##print STDOUT '@TIME_LAPSES = (', join (', ', @TIME_LAPSES), ") .\n\n";
		# End of "Modified on 2020/01/31. (00-C-2)" #

=DISABLED on 2020/01/31. (2)
print STDOUT '@E_INDEL_CTS = (', join (', ', @E_INDEL_CTS), ") .\n\n";
=cut # End of "DISABLED on 2020/01/31. (2)"

print STDOUT "\n";

	### End of "(IMPORTED w/ modifications from '../../../lastpiece.alpha.pl' on 2020/01/30) (00-C)" ###

	##### END of "(IMPORTED from '../muP_case_i2iii/cmpt_muP_case_i2iii.pl' on 2020/02/03) (Pre 00)" #####


	### CREATED by modifying "Main Process II" of '../muP_case_i2iii/cmpt_muP_case_i2iii.pl',
	###  in creation of 'cmpt_muP0.pl' (on 2020/02/03). (MAIN)

	###  Main Process: Compute multiplication factors for a fixed value of $SIGMA_BE, via forward extention. ###

print STDOUT "\n<<<<< MAIN Process: Compute multiplication factors for \$SIGMA_BE = $SIGMA_BE, via forward extention. >>>>>\n\n";


		# Modified for attempting to fix a bug (on 2020/02/03). (1) #
##my @muP0s = (); # The $L_I th element = {a 2D array of muP0 [($l, t_j) | (L_I = $L_I, t_I) ] }
my @muP0s = ([]); # The $L_I th element = {a 2D array of muP0 [($l, t_j) | (L_I = $L_I, t_I) ] } 
		# End of "Modified for attempting to fix a bug (on 2020/02/03). (1)" #


for (my $L_I = 1; $L_I <= $UB_CT_SITES_CMPT; $L_I++) { # MAIN Outermost for-loop (over initial site-counts). #

   (print STDOUT "\n{{{ \$L_I = $L_I }}}\n\n") if ($TEST_LEVEL >1);

		# Modified for attempting to fix a bug (on 2020/02/03). (2) #
    ##my @muP0s_fxd_L_I = (); # The $j th element = {an array of muP0 [($l, t_j) | (L_I = $L_I, t_I) ] }
    ##$muP0s[$L_I] = \@muP0s_fxd_L_I;
    my $muP0s_fxd_L_I = $muP0s[$L_I] = [];
		# End of "Modified for attempting to fix a bug (on 2020/02/03). (2)" #

		# Initialize the array of muP0s (w/ L_I = $L_I) for zero time-interval. #
    ##my @init_muP0s_fxd_L_I = (); # the $l th element = muP0 [ ($l, t_I) | (L_I = $L_I, t_I) ].
    my @init_muP0s_fxd_L_I = (0); # the $l th element = muP0 [ ($l, t_I) | (L_I = $L_I, t_I) ]. # Modified for attempting to fix a bug (on 2020/02/03) (5) #
    for (my $l=1; $l <= $UB_CT_SITES_CMPT; $l++) { # 1st preliminary for-loop (over site-counts). #
	$init_muP0s_fxd_L_I[$l] = ($l == $L_I) ? 1 : 0 ;
    } # End of the 1st preliminary for-loop (over site-counts). #

		# Modified for attempting to fix a bug (on 2020/02/03). (3) #
    ##$muP0s_fxd_L_I[0] = \@init_muP0s_fxd_L_I; # Initialize the 2D array for fixed $L_I.
    $muP0s_fxd_L_I->[0] = \@init_muP0s_fxd_L_I; # Initialize the 2D array for fixed $L_I.
		# End of "Modified for attempting to fix a bug (on 2020/02/03). (3)" #

   if ($TEST_LEVEL >3) {
	print STDOUT '@init_muP0s_fxd_L_I = (', join (", ", @init_muP0s_fxd_L_I), ") .\n\n";

	print STDOUT '$muP0s_fxd_L_I->[0] = ', $muP0s_fxd_L_I->[0], " ,\n";
	print STDOUT '@{$muP0s_fxd_L_I->[0]} = (', join (", ", @{$muP0s_fxd_L_I->[0]}), ") .\n\n";

	print STDOUT '$muP0s[$L_I]->[0] = ', $muP0s[$L_I]->[0], " ,\n";
	print STDOUT '@{$muP0s[$L_I]->[0]} = (', join (", ", @{$muP0s[$L_I]->[0]}), ") .\n\n\n"; 

   } # End of "if ($TEST_LEVEL >3) {...}"

    for (my $jt = 1; $jt <= $CT_SB_INTVLS; $jt++) { # Main outer for-loop (over time-points). #

		# Modified for attempting to fix a bug (on 2020/02/03). (3+1) #
	##my $rf_prev_muP0s = $muP0s_fxd_L_I[$jt-1]; # the $l th element = muP [ ($l, t_{j-1}) | (L_I = $L_I, t_I) ].
	##my $rf_curr_muP0s = $muP0s_fxd_L_I[$jt] = []; # the $l th element = muP [ ($l, t_j) | (L_I = $L_I, t_I) ].
	my $rf_prev_muP0s = $muP0s_fxd_L_I->[$jt-1]; # the $l th element = muP [ ($l, t_{j-1}) | (L_I = $L_I, t_I) ].
	##my $rf_curr_muP0s = $muP0s_fxd_L_I->[$jt] = []; # the $l th element = muP [ ($l, t_j) | (L_I = $L_I, t_I) ].
	my $rf_curr_muP0s = $muP0s_fxd_L_I->[$jt] = [0]; # the $l th element = muP [ ($l, t_j) | (L_I = $L_I, t_I) ]. # Further modified to fix a bug (on 2020/02/03). (6) #
		# End of "Modified for attempting to fix a bug (on 2020/02/03). (3+1)" #


	for (my $lj = 1; $lj <= $UB_CT_SITES_CMPT; $lj++) { # Main middle for-loop (over site-counts) #

	    my $curr_muP0 = (1 - $SIZE_SB_INTVL * $TOTRATE_INDEL * $lj) * $rf_prev_muP0s->[$lj];

	    ##my $ub_inslen = ($CUTOFF_INS_LEN < $lj) ? $CUTOFF_INS_LEN : $lj ;
	    my $ub_inslen = ($CUTOFF_INS_LEN < $lj -1) ? $CUTOFF_INS_LEN : $lj -1; # Modified in creation of 'cmpt_muP0.pl' (on 2020/02/03). (M1) #

	    $ub_inslen++;

	    my $sum_ins = 0;

	    for (my $l = 1; $l < $ub_inslen; $l++) { # 1st inner for-loop (over insertion lengths). #
		    ##$sum_ins += ($lj - $l + 1) * $rf_RATES_INS->[$l] * $rf_prev_muP0s->[$lj - $l];
		    $sum_ins += ($lj - $l - 1 + $SIGMA_BE) * $rf_RATES_INS->[$l] * $rf_prev_muP0s->[$lj - $l]; # Modified in creation of 'cmpt_muP0.pl' (on 2020/02/03). (M2) #
	    } # End of the 1st inner for-loop (over insertion lengths). #

	    my $ub_dellen = ($CUTOFF_DEL_LEN < $UB_CT_SITES_CMPT - $lj) ?  $CUTOFF_DEL_LEN : $UB_CT_SITES_CMPT - $lj ;
	    $ub_dellen++;

	    my $sum_del = 0;

	    for (my $l=1; $l < $ub_dellen; $l++) { # 2nd inner for-loop (over deletion lengths). #
		    $sum_del += $rf_RATES_DEL->[$l] * $rf_prev_muP0s->[$lj + $l];
	    } # End of the 2nd inner for-loop (over deletion lengths). #
	    
	    $curr_muP0 += $SIZE_SB_INTVL * ( $sum_ins + ($lj + 1) * $sum_del );

	    $rf_curr_muP0s->[$lj] = $curr_muP0;

	} # End of the main middle for-loop (over site-counts). #

    }	# End of the main outer for-loop (over time-points). #

   if ($TEST_LEVEL >3) {

	print STDOUT "[ After the computation: ]\n\n";

	print STDOUT '$muP0s[$L_I]->[0] = ', $muP0s[$L_I]->[0], " ,\n";
	print STDOUT '@{$muP0s[$L_I]->[0]} = (', join (", ", @{$muP0s[$L_I]->[0]}), ") .\n\n\n"; 

   } # End of "if ($TEST_LEVEL >3) {...}"


} # END of the MAIN Outermost for-loop (over initial site-counts). #

	### END of "CREATED by modifying "Main Process II" of '../muP_case_i2iii/cmpt_muP_case_i2iii.pl',
	###  in creation of 'cmpt_muP0.pl' (on 2020/02/03). (MAIN)"


	### Finishing Process: Output the Results. ###

print STDOUT "\n<<<<< Finishing Process: Output the Results. >>>>>\n\n";

unless ( print_muP0s ($ofile_muP0, '#{sites}_Init:#{sites}_Fin\Time_Lapse', @TIME_LAPSES_CMPT, $UB_CT_SITES_CMPT, @muP0s) ) {
	print STDOUT "[ERROR '$program_name'] The subroutine, 'print_muP0s', FAILED to output to the file, \"$ofile_muP0\" $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}
(print STDOUT "\n ... The subroutine, 'print_muP0s', Ended outputting to the file, \"$ofile_muP0\" ...\n\n") if ($TEST_LEVEL > 1);


print STDOUT "\n\n === All The Processes of '$program_name' FINISHED !! ===\n\n\n";


