#!/usr/bin/perl
#
###################################################################################
#	
#	LASTPIECE_P (Local Alignment-STate Probabilities that Insertion-type and dEletion-type gaps Co-Exist, type P):
#	    A package of programs to compute the probabilities of gapped segments in each of which an insertion-type gap and a deletion-type gap co-exist, (which were referred to as case-(iv) gapped segments by Ezawa (2016a?)),
#	    under a stochastic model of sequence evolution with biologically realistic insertions/deletions.
#	   (Written almost exclusively in Perl.)
#
#	Version 0.3: Copyright (C) 2020 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
#
###################################################################################
#
# cmpt_muP_case_i2iii.pl:
#	This servant script computes the multiplication factors
#	of gapped segments classified as case-(i), case-(ii), and case-(iii).
#	[CREATION STARTED on 2020/01/30, ENDED on 2020/02/01.]
#	[2 BUGs FIXED on 2020/02/02]:
#	(1) The case-(i) factors at time-interval = 0, which were missed by mistake,
#		 are now correctly output;
#	(2) Now, the case-(ii) factors are correctly rescaled,
#		whereas the case-(iii) were rescaled by mistake before the BUG fixation.
#
#	[MAJOR REVISION on 2020/03/01]:
#	In this revised version, the RESCALING of the case-(ii) factors is UNDONEs;
#	instead, the RESCALING WILL BE PERFORMED within the master script, '../../../lastpiece.alpha.pl'.
#

		### (IMPORTED w/ modifications from '../../../lastpiece.alpha.pl' on 2020/01/30) (HD) ###

use strict;
use warnings;

use MyFile_manipu qw(
	open_input_file2
	open_output_file
	list_sessions_ext
	);

use MyLASTPIECE_IO qw(
	readin_indel_params
	readin_indel_rates
	print_muP_case_i
	print_muP_case_ii_or_iii 
	);
	# Added 'readin_indel_params' & 'readin_indel_rates' on 2020/01/31. #
	# Added 'print_muP_case_i' & 'print_muP_case_ii_or_iii' on 2020/02/01. #

##use MyLASTPIECE_Sppl qw(
##	cal_delta_exp_del_5p2
##	);


##sub create_dir_rw (\$); # Added on 2020/01/29.


$| = 1; # Flush the buffer for standard output as soon as it receives data.

my $bin_perl = "/usr/bin/perl"; # The Perl interpreter. # (Added on 2020/01/29). #

my $program_path = $0;
my ($program_name) = ($program_path =~ m!([^/]+)$!);

	# Controlling the level of test outputs. #
	# (0: silent; 1: very quiet; 2: quiet; 3: medium; 4: verbose; 5: very verbose) 
our $TEST_LEVEL = 2;
##our $TEST_LEVEL = 3;

#$MyLASTPIECE_IO::TEST_LEVEL = $TEST_LEVEL ; # Added on 2020/01/27.


		### END of "(IMPORTED w/ modifications from '../../../lastpiece.alpha.pl' on 2020/01/30) (HD)" ###

our $EPS = 1E-07; # A "small" number to be used for checking the consistency. # (Added on 2020/02/01.) #


	####### Take in the Command-Line Arguments. #######

unless (@ARGV > 8) {
	print STDERR "\nUsage:  [perl] $program_name {initial_time (non-negative number)} {final_time (positive number, greater than initial_time)} {number_of_sub_time_intervals (positive integer)} {upper-bound_of_number_of_sites} {input_file_for_indel_parameters} {input_file_for_raw_indel_rates} {output_file_for_case_i} {output_file_for_case_ii} {output_file_for_case_iii}";
	print STDERR " .\n\n";
	##print STDERR " [ options ] ,\n";
	##print STDERR "   where the [options] can consist of any (or none) of the following:\n\n";
	##print STDERR "            --MIN_WD_SIZE=integer      for the minimum window size for detecting purge-candidates (default=$MIN_WD_SIZE_PCAND);\n";
	##print STDERR "           --XLEVEL=integer       for the level of MSA space exploration (from =1 (performing only multiple-\"shift\"s) to =5 (performing all programed moves) (default=$XLEVEL);\n"; # ADDED on 2019/08/01.
	##print STDERR "\n"; # ADDED in REVISION (on 2019/10/18). (ARGV2) #

	die "$!\n\n";
}

our ($TIME_INIT, $TIME_FIN, $CT_SB_INTVLS, $UB_CT_SITES_CMPT) = @ARGV[0 .. 3];
my ($infile_id_prms, $infile_id_rates, $ofile_muP_case_i, $ofile_muP_case_ii, $ofile_muP_case_iii) = @ARGV[4 .. 8];

	####### END of "Take in the Command-Line Arguments." #######


	### Check the existence & permission mode of the input files. ###
	### (Added on 2020/02/03.) ###

foreach my $infile ($infile_id_prms, $infile_id_rates) {
    unless ( ( -f $infile ) and ( -r $infile ) ) {
	print STDOUT "[ERROR '$program_name'] CanNOT FIND a readable input file, \"$infile\": $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
    }
}
	### End of "(Added on 2020/02/03.)" ###


	### Declare the Beginning of the Script ###
	### (Selectively IMPORTED w/ modifications from '../../../lastpiece.alpha.pl' on 2020/01/30) (DC) ###

print STDOUT "\n<<< Program name: $program_name (path: $program_path) >>>\n\n";

			# (Newly added on 2020/01/30) (DC1) #

print STDOUT "\n<< Input Files: >>\n\n";

print STDOUT " (for indel parameters) $infile_id_prms ,\n";
print STDOUT " (for raw indel rates)  $infile_id_rates .\n\n";

print STDOUT "\n<< Output Files: >>\n\n";

print STDOUT " (for muP of case-i segments)  $ofile_muP_case_i ,\n";
print STDOUT " (for muP of case-ii segments) $ofile_muP_case_ii ,\n";
print STDOUT " (for muP of case-iii segments) $ofile_muP_case_ii .\n\n";

			# End of "(Newly added on 2020/01/30) (DC1)" #

print STDOUT "\n<< Global Variables specifying the levels of tests: >>\n\n"; # Added on 2020/01/30. (DC2) #

print STDOUT "\n", '$TEST_LEVEL = ', $TEST_LEVEL, " .\n\n";


			# Added on 2020/02/01. #
print STDOUT "\n<< A \"small\" number to be used for checking the consistency: >>\n\n"; #
print STDOUT '$EPS = ', $EPS, " .\n";
			# End of "Added on 2020/02/01." #


print STDOUT "\n<< Parameters necessary for recursive computations of the multiplication factors: >>\n\n";

print STDOUT '($TIME_INIT, $TIME_FIN) = ', "($TIME_INIT, $TIME_FIN) .\n\n";
print STDOUT '#{Sub-time-intervals} = ', $CT_SB_INTVLS, " .\n\n";

print STDOUT '{Upper-bound of #{sites} for computation} = ', $UB_CT_SITES_CMPT, " .\n\n"; 

	### End of "(Selectively IMPORTED w/ modifications from '../../../lastpiece.alpha.pl' on 2020/01/30) (DC)" ###



	## (MODIFIED on 2020/01/31) (00-IN) ##

print STDOUT "\n<<< Preliminary Process 00-IN: Reading in the quantities necessary for the computations >>>\n\n";

	## 00-IN-A: Read in the input file for indel parameters. ##

print STDOUT "\n<< (Pre 00-IN-A): Reading in the indel model parameters >>\n\n";

our ($TOTRATE_INS, $TOTRATE_DEL, $CUTOFF_INS_LEN, $CUTOFF_DEL_LEN)
   = readin_indel_params ($infile_id_prms);

unless (defined $CUTOFF_DEL_LEN) {
	print STDOUT "[ERROR '$program_name'] The subroutine, 'readin_indel_params', FAILED on the input file, \"$infile_id_prms\" $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}
(print STDOUT "\n ... The subroutine, 'readin_indel_params', Ended on the input file, \"$infile_id_prms\" ...\n\n") if ($TEST_LEVEL > 1);

##my ($infile_id_prms, $infile_id_rates, $ofile_muP_case_i, $ofile_muP_case_ii, $ofile_muP_case_iii) = @ARGV[4 .. 8];

print STDOUT "\n[ Indel Model Parameters retrieved: ]\n\n";

print STDOUT '($TOTRATE_INS, $TOTRATE_DEL) = ', "($TOTRATE_INS, $TOTRATE_DEL) .\n";
print STDOUT '($CUTOFF_INS_LEN, $CUTOFF_DEL_LEN) = ', "($CUTOFF_INS_LEN, $CUTOFF_DEL_LEN) .\n\n";


	## Preliminary 00-B: Read in the input file for raw indel rates. ##

print STDOUT "\n<< (Pre 00-IN-B): Reading in the raw indel rates >>\n\n";

our ($rf_RATES_INS, $rf_RATES_DEL, $rf_RATES_INS_BD, $rf_RATES_DEL_BD)
    = readin_indel_rates ($infile_id_rates);

unless (defined $rf_RATES_DEL_BD) {
	print STDOUT "[ERROR '$program_name'] The subroutine, 'readin_indel_rates', FAILED on the input file, \"$infile_id_rates\" $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}
(print STDOUT "\n ... The subroutine, 'readin_indel_rates', Ended on the input file, \"$infile_id_rates\" ...\n\n") if ($TEST_LEVEL > 1);


if ($TEST_LEVEL > 2) {

    print STDOUT "\n[ Raw Indel Rates retrieved: ]\n\n";

    print STDOUT join ("\t", 'Length', 'Ins', 'Del', 'Ins(bd)', 'Del(bd)'), "\n\n";

    my $maxlen = 0;
    foreach my $rf_rates ($rf_RATES_INS, $rf_RATES_DEL, $rf_RATES_INS_BD, $rf_RATES_DEL_BD) {
	my $len = @{$rf_rates} - 1;
	if ($maxlen < $len) { $maxlen = $len; }
    }

    for (my $l=1; $l <= $maxlen; $l++) {
	my @out_rates = ();
	foreach my $rf_rates ($rf_RATES_INS, $rf_RATES_DEL, $rf_RATES_INS_BD, $rf_RATES_DEL_BD) {
	    my $rate = $rf_rates->[$l];
	    unless (defined $rate) { $rate = 'undef'; }
	    push @out_rates, $rate;
	}
	print STDOUT join ("\t", $l, @out_rates), "\n";
    }
    print STDOUT "\n\n";

} # End of "if ($TEST_LEVEL > 2) {...}"


=OBSOLETE as of 2020/01/31. (00-IN)
	## Preliminary 00-A: Read in the input file for indel parameters. ##

print STDOUT "\n<< Indel Model Parameters retrieved: >>\n\n";

print STDOUT '$TOTRATE_INS = ', $TOTRATE_INS, " ,\n";
print STDOUT '$TOTRATE_DEL = ', $TOTRATE_DEL, " .\n\n";

	## Preliminary 00-B: Read in the input file for raw indel rates. ##

=cut # End of "OBSOLETE as of 2020/01/31. (00-IN)"

	## End of "(MODIFIED on 2020/01/31) (00-IN)" ##


	## Preliminary 00-C: Compute Other Important Parameters ##
	### (IMPORTED w/ modifications from '../../../lastpiece.alpha.pl' on 2020/01/30) (00-C) ###

print STDOUT "\n<<< Preliminary Process 00-C: Computing Other Important Parameters >>>\n\n";

our $TOTRATE_INDEL = $TOTRATE_INS + $TOTRATE_DEL;
our $SIZE_FULL_INTVL = $TIME_FIN - $TIME_INIT;
our $SIZE_SB_INTVL = $SIZE_FULL_INTVL /$CT_SB_INTVLS;

		# Modified on 2020/01/31. (00-C-1) #
our @TIME_LAPSES_CMPT = (); # Values of the time lapse (= t_F - t_I) at which the factors will be computed.
for (my $i = 0; $i <= $CT_SB_INTVLS; $i++) {
	$TIME_LAPSES_CMPT[$i] = $i * $SIZE_SB_INTVL;
} 
=OBSOLETE as of 2020/01/31. (A)
our @TIME_LAPSES = (); # Values of the time lapse (= t_F - t_I) to be examined.
for (my $i = 0; $i <= $CT_SB_INTVLS; $i++) {
	$TIME_LAPSES[$i] = $i * $SIZE_SB_INTVL;
} 
=cut # End of "OBSOLETE as of 2020/01/31. (A)"
		# End of "Modified on 2020/01/31. (00-C-1)" #

=DISABLED on 2020/01/31. (1)
our @E_INDEL_CTS = (); # Expected counts of indels. #
foreach my $tlapse (@TIME_LAPSES) { push @E_INDEL_CTS, $TOTRATE_INDEL * $tlapse; }
=cut # End of "DISABLED on 2020/01/31. (1)"

print STDOUT '$TOTRATE_INDEL = ', $TOTRATE_INDEL, " \;\n";
print STDOUT '$SIZE_FULL_INTVL = ', $SIZE_FULL_INTVL, " \;\n";
print STDOUT '$SIZE_SB_INTVL = ', $SIZE_SB_INTVL, " .\n\n";

		# Modified on 2020/01/31. (00-C-2) #
print STDOUT '@TIME_LAPSES_CMPT = (', join (', ', @TIME_LAPSES_CMPT), ") .\n\n";
##print STDOUT '@TIME_LAPSES = (', join (', ', @TIME_LAPSES), ") .\n\n";
		# End of "Modified on 2020/01/31. (00-C-2)" #

=DISABLED on 2020/01/31. (2)
print STDOUT '@E_INDEL_CTS = (', join (', ', @E_INDEL_CTS), ") .\n\n";
=cut # End of "DISABLED on 2020/01/31. (2)"

print STDOUT "\n";

	### End of "(IMPORTED w/ modifications from '../../../lastpiece.alpha.pl' on 2020/01/30) (00-C)" ###


	###  Main Process I: Compute multiplication factors for cases (i) and (ii), via backward extention. ###

print STDOUT "\n<<<<< Main Process I: Compute multiplication factors for cases (i) and (ii), via backward extention. >>>>>\n\n";

my @muP_cases_i_and_ii = (); # The ($CT_SB_INTVLS - $j) th element = {an array of muP [(L_F =0, t_F) | ($l, t_j) ] }

		# Initialize the array of muPs for zero time-interval. #
my @init_prev_muPs = (); # the $l th element = muP [ (L_F=0, t_F) | ($l, t_F) ] = muP [ ($l, t_I) | (L_I = 0, t_I) ].
for (my $l=0; $l <= $UB_CT_SITES_CMPT; $l++) { # 1st preliminary for-loop (over site-counts). #
	$init_prev_muPs[$l] = ($l > 0) ? 0 : 1 ;
} # End of the 1st preliminary for-loop (over site-counts). #

$muP_cases_i_and_ii[0] = \@init_prev_muPs; # Initialize the 2D array for cases (i) and (ii).

for (my $jt = 1; $jt <= $CT_SB_INTVLS; $jt++) { # 1st main outer for-loop (over time-points). #

    my $rf_prev_muPs = $muP_cases_i_and_ii[$jt-1]; # the $l th element = muP [ (L_F=0, t_F) | ($l, t_{j+1}) ].

    my $rf_curr_muPs = $muP_cases_i_and_ii[$jt] = []; # the $l th element = muP [ (L_F=0, t_F) | ($l, t_j) ].

    for (my $lj = 0; $lj <= $UB_CT_SITES_CMPT; $lj++) { # 1st middle for-loop (over site-counts) #

	    my $curr_muP = (1 - $SIZE_SB_INTVL * $TOTRATE_INDEL * $lj) * $rf_prev_muPs->[$lj];

	    my $ub_inslen = ($CUTOFF_INS_LEN < $UB_CT_SITES_CMPT - $lj) ? $CUTOFF_INS_LEN : $UB_CT_SITES_CMPT - $lj ;
	    $ub_inslen++;

	    my $sum_ins = 0;

	    for (my $l = 1; $l < $ub_inslen; $l++) { # 1st inner for-loop (over insertion lengths). #
		$sum_ins += $rf_RATES_INS->[$l] * $rf_prev_muPs->[$lj + $l];
	    } # End of the 1st inner for-loop (over insertion lengths). #

	    my $ub_dellen = ($CUTOFF_DEL_LEN < $lj) ?  $CUTOFF_DEL_LEN : $lj ;
	    $ub_dellen++;

	    my $sum_del = 0;

	    for (my $l=1; $l < $ub_dellen; $l++) { # 2nd inner for-loop (over deletion lengths). #
		$sum_del += ($lj - $l + 1) * $rf_RATES_DEL->[$l] * $rf_prev_muPs->[$lj - $l];
	    } # End of the 2nd inner for-loop (over deletion lengths). #

	    
	    $curr_muP += $SIZE_SB_INTVL * ( ($lj + 1) * $sum_ins + $sum_del );

	    $rf_curr_muPs->[$lj] = $curr_muP;

    } # End of the 1st middle for-loop (over site-counts). #

	
    ##$rf_prev_muPs = $rf_curr_muPs; # Update the array of previous muPs. # DISABLED (on 2020/02/03) because it has no meaning. #

}	# End of the 1st main outer for-loop (over time-points). #



	###  Main Process II: Compute multiplication factors for cases (i) and (iii), via forward extention. ###

print STDOUT "\n<<<<< Main Process II: Compute multiplication factors for cases (i) and (iii), via forward extention. >>>>>\n\n";

my @muP_cases_i_and_iii = (); # The $j th element = {an array of muP [($l, t_j) | (L_I = 0, t_I) ] }

$muP_cases_i_and_iii[0] = \@init_prev_muPs; # Initialize the 2D array for cases (i) and (iii).


for (my $jt = 1; $jt <= $CT_SB_INTVLS; $jt++) { # 2nd main outer for-loop (over time-points). #

    my $rf_prev_muPs = $muP_cases_i_and_iii[$jt-1]; # the $l th element = muP [ ($l, t_{j-1}) | (L_I = 0, t_I) ].

    my $rf_curr_muPs = $muP_cases_i_and_iii[$jt] = []; # the $l th element = muP [ ($l, t_j) | (L_I = 0, t_I) ].

    for (my $lj = 0; $lj <= $UB_CT_SITES_CMPT; $lj++) { # 2nd middle for-loop (over site-counts) #

	    my $curr_muP = (1 - $SIZE_SB_INTVL * $TOTRATE_INDEL * $lj) * $rf_prev_muPs->[$lj];

	    my $ub_inslen = ($CUTOFF_INS_LEN < $lj) ? $CUTOFF_INS_LEN : $lj ;
	    $ub_inslen++;

	    my $sum_ins = 0;

	    for (my $l = 1; $l < $ub_inslen; $l++) { # 3rd inner for-loop (over insertion lengths). #
		$sum_ins += ($lj - $l + 1) * $rf_RATES_INS->[$l] * $rf_prev_muPs->[$lj - $l];
	    } # End of the 3rd inner for-loop (over insertion lengths). #

	    my $ub_dellen = ($CUTOFF_DEL_LEN < $UB_CT_SITES_CMPT - $lj) ?  $CUTOFF_DEL_LEN : $UB_CT_SITES_CMPT - $lj ;
	    $ub_dellen++;

	    my $sum_del = 0;

	    for (my $l=1; $l < $ub_dellen; $l++) { # 5th inner for-loop (over deletion lengths). #
		$sum_del += $rf_RATES_DEL->[$l] * $rf_prev_muPs->[$lj + $l];
	    } # End of the 5th inner for-loop (over deletion lengths). #

	    
	    $curr_muP += $SIZE_SB_INTVL * ( $sum_ins + ($lj + 1) * $sum_del );

	    $rf_curr_muPs->[$lj] = $curr_muP;

    } # End of the 2nd middle for-loop (over site-counts). #

	
    ##$rf_prev_muPs = $rf_curr_muPs; # Update the array of previous muPs. # DISABLED (on 2020/02/03) because it has no meaning. #

}	# End of the 2nd main outer for-loop (over time-points). #


	### Main Process III: Rescale the case-(ii) factors. ###

print STDOUT "\n\n<<<<< Main Process III: Rescale the case-(ii) factors. >>>>>\n\n\n";


			# (Added in MAJOR REVISION (on 2020/03/01).) (Main III) #

print STDOUT "# NOTE: As of March 1st, 2020, this script (i.e., '$program_name') NO LONGER rescales the case-(ii) factors;\n";
print STDOUT "#       The RESCALING SHOULD BE performed by the master script, 'lastpiece.alpha.pl',\n";
print STDOUT "#        when outputting the LOGARITHMs of the case-(ii) factors .\n";
print STDOUT "#\n\n\n";

			# End of "(Added in MAJOR REVISION (on 2020/03/01).) (Main III)" #

=DISABLED in MAJOR REVISION (on 2020/03/01). (Main III)
for (my $jt = 1; $jt <= $CT_SB_INTVLS; $jt++) { # 3rd main outer for-loop (over time-points). #

		## A BUG fixed on 2020/02/02. (2) ##
    my $rf_curr_muPs = $muP_cases_i_and_ii[$jt]; # the $l th element = muP [ (L_F=0, t_F) | ($l, t_j) ].
    ##my $rf_curr_muPs = $muP_cases_i_and_iii[$jt]; # the $l th element = muP [ ($l, t_j) | (L_I = 0, t_I) ].
		## End of "A BUG fixed on 2020/02/02. (2)" ##

    my $base_mlt = exp ( $SIZE_SB_INTVL * $jt * $TOTRATE_INDEL );
    my $multiplier = 1;

    for (my $lj = 1; $lj <= $UB_CT_SITES_CMPT; $lj++) { # 3rd middle for-loop (over site-counts) #

	$multiplier *= $base_mlt;
	$rf_curr_muPs->[$lj] *= $multiplier;

    } # End of the 3rd middle for-loop (over site-counts). #

}	# End of the 3rd main outer for-loop (over time-points). #

=cut # END of "DISABLED in MAJOR REVISION (on 2020/03/01). (Main III)"


	### Trailing Process (1): Check the consistency between case-(i) factors computed in two different ways. ### 

print STDOUT "\n<<< Trailing Process (1): Check the consistency between case-(i) factors computed in two different ways. >>>\n\n";

my @inconsistent = (); # element = [$jt, $muP_case_i_A, $muP_case_ii_A, $abs_diff] for the time-interval index at which the factors are inconsistent. #


my @muP_case_i = (); # The $jt th element is the factor for case (i) at the $jt th time-interval.

if ($TEST_LEVEL > 2) {  # (Added on 2020/02/02) (1) #
	print STDOUT join ("\t", 'Time-interval-index', 'muP[case(i)]_A', 'muP[case(i)]_B', '|diff|'), "\n\n";
} # End of "if ($TEST_LEVEL > 2) { # (Added on 2020/02/02) (1) # ...}"

		# A BUG fixed on 2020/02/02. (1) #
for (my $jt = 0; $jt <= $CT_SB_INTVLS; $jt++) { # 1st trailing for-loop (over time-points). #
##for (my $jt = 1; $jt <= $CT_SB_INTVLS; $jt++) { # 1st trailing for-loop (over time-points). #
		# End of "A BUG fixed on 2020/02/02. (1)" #

	my $muP_case_i_A = $muP_cases_i_and_ii[$jt]->[0];
	my $muP_case_i_B = $muP_cases_i_and_iii[$jt]->[0];

	my $diff = $muP_case_i_A - $muP_case_i_B;
	my $abs_diff = ($diff > 0) ? $diff : -$diff ;

	if ($abs_diff > $EPS) {
	    push @inconsistent, [$jt, $muP_case_i_A, $muP_case_i_B, $abs_diff ];
	}
   if ($TEST_LEVEL > 2) { # (Added on 2020/02/02) (2) #
	print STDOUT join ("\t", $jt, $muP_case_i_A, $muP_case_i_B, $abs_diff ), "\n";
   } # End of "if ($TEST_LEVEL > 2) { # (Added on 2020/02/02) (2) # ...}"

	$muP_case_i[$jt] = $muP_case_i_A;

}	# End of the 1st trailing for-loop (over time-points). #

if (@inconsistent >0) {
	my $ct_icst = @inconsistent;
	print STDOUT "[ERROR '$program_name'] Found $ct_icst time-intervals at which the case-(i) factors are INCONSISTNT: $? !\n\n";
	print STDOUT join ("\t", 'Time-interval-index', 'muP[case(i)]_A', 'muP[case(i)]_B', '|diff|'), "\n\n";
	foreach my $entry (@inconsistent) { print STDOUT join ("\t", @{$entry}), "\n"; }
	print STDOUT "\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}



	### Finishing Process: Output the Results. ###

print STDOUT "\n<<<<< Finishing Process: Output the Results. >>>>>\n\n";

print STDOUT "\n<<< (F-1): Case-(i) >>>\n\n";

unless ( print_muP_case_i ($ofile_muP_case_i, '#{Anc_sites}:#{dec_sites}\Time_Lapse', @TIME_LAPSES_CMPT, @muP_case_i) ) {
	print STDOUT "[ERROR '$program_name'] The subroutine, 'print_muP_case_i', FAILED to output to the file, \"$ofile_muP_case_i\" $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}
(print STDOUT "\n ... The subroutine, 'print_muP_case_i', Ended outputting to the file, \"$ofile_muP_case_i\" ...\n\n") if ($TEST_LEVEL > 1);


print STDOUT "\n<<< (F-2): Case-(ii) >>>\n\n";

unless ( print_muP_case_ii_or_iii ($ofile_muP_case_ii, '#{Anc_sites}\Time_Lapse', @TIME_LAPSES_CMPT, @muP_cases_i_and_ii) ) {
	print STDOUT "[ERROR '$program_name'] The subroutine, 'print_muP_case_ii_or_iii', FAILED to output to the file, \"$ofile_muP_case_ii\" $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}
(print STDOUT "\n ... The subroutine, 'print_muP_case_ii_or_iii', Ended outputting to the file, \"$ofile_muP_case_ii\" ...\n\n") if ($TEST_LEVEL > 1);


print STDOUT "\n<<< (F-3): Case-(iii) >>>\n\n";

unless ( print_muP_case_ii_or_iii ($ofile_muP_case_iii, '#{Dec_sites}\Time_Lapse', @TIME_LAPSES_CMPT, @muP_cases_i_and_iii) ) {
	print STDOUT "[ERROR '$program_name'] The subroutine, 'print_muP_case_ii_or_iii', FAILED to output to the file, \"$ofile_muP_case_iii\" $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}
(print STDOUT "\n ... The subroutine, 'print_muP_case_ii_or_iii', Ended outputting to the file, \"$ofile_muP_case_iii\" ...\n\n") if ($TEST_LEVEL > 1);

		### RESTART FROM HERE ###


print STDOUT "\n\n === All The Processes of '$program_name' FINISHED !! ===\n\n\n";





