#!/usr/bin/perl
#
###################################################################################
#	
#	LASTPIECE_P (Local Alignment-STate Probabilities that Insertion-type and dEletion-type gaps Co-Exist, type P):
#	    A package of programs to compute the probabilities of gapped segments in each of which an insertion-type gap and a deletion-type gap co-exist, (which were referred to as case-(iv) gapped segments by Ezawa (2016a?)),
#	    under a stochastic model of sequence evolution with biologically realistic insertions/deletions.
#	   (Written almost exclusively in Perl.)
#
#	Version 0.3: Copyright (C) 2020 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
#
###################################################################################
#
# wgtd_summs_3Darrays_for_mfacs.pl:
#	This servant script performs the element-wise weighted summations of two or more 3D-arrays of multiplication factors,
#	each of the form:
#	muP_i [ (L_F = $LF, t_F) | (L_I = $LI, t_I) ], with (i = 1, 2, ...),
#	that the number of sites is L_I at the initial time (=t_I) and L_F at the final time (=t_F),
#	with 1 <= $LI, $LF <= $UB_CT_SITES,
#	read-out in the form: $muP_i[$LI]->[$LF][$jt], where $jt = (t_F - t_I)/$SIZE_SB_INTVL.
#	The element-wise weighted summations are given by:
#
#	$muP_wsum[$LI]->[$LF][$jt] = Sum_{i=1}^{n} wgt_i * $muP_i[$LI]->[$LF][$jt] .
#
#	The script was created by taking advantage of the main framework of another servant script,
#	'graft_convolution.pl' (in "../../").
#	[CREATION STARTED on 2020/02/29, ENDED on 2020/03/01.]
#
# graft_convolution.pl:
#	This servant script performs the "graft-convolution" of TWO multiplication factors of the form:
#	muP_{1st or 2nd} [ (L_F, t_F)  | (L_I, t_I) ],
#	that the number of sites is L_I at the initial time (=t_I) and L_F at the final time (=t_F),
#	with 1 <= L_I, L_F <= $UB_CT_SITES.
#	Here, the "graft-convolution" means the following operation:
#	muP_{Output} [ (L_F, t_F) | (L_I, t_I) ]
#	     = \sum_{Lm = 1}^{$UB_CT_SITES} \int_{t_I}^{t_F} d{}tm  [
#	    muP_{1st} [ (Lm, tm)  | (L_I, t_I) ] * muP_{2nd} [ (L_F, t_F)  | (Lm, tm) ] 
#		].
#	The script was created by taking advantage of the main framework of another servant script,
#	'cmpt_mu_Dcr2Adel.pl' (in "./2nd_ii/mu_Dcr2Adel/").
#	[CREATION STARTED on 2020/02/14, ENDED on 2020/02/dd.]
#	
#
# cmpt_mu_Dcr2Adel.pl
#	This servant script computes the multiplication factor,
#	mu_{Dcr->Adel} [ (L_A, x(=L_A), [t_I, t_I+dt_I]) ; (L_D, [t_F-dt_F, t_F]) ](sigmaBE=1),
#	for the creation of a "D"-region within [t_I, t_I + dt_I]
#	and the deletion of an "A"-region within [t_F - dt_F, t_F],
#	where L_D is the number of sites belonging to the "D"-region at the "final" time (t_F),
#	and L_A is the number of sites belonging to the "A"-region at the "initial" time (t_I).
#	The script was created by taking advantage of the main framework of another servant script,
#	'cmpt_muP_2nd_i.pl' (in "../../2nd_i/PreFinal/").
#	[CREATION STARTED on 2020/02/11, ENDED on 2020/02/dd.]
#
# cmpt_muP_2nd_i.pl:
#	This servant script computes the multiplication factor,
#	muP_{case (iv)}^{2nd (i)} [(L_DF, t_F) | (L_AI, t_I) ],
#	which is the total contributions from ALL indel histories belonging to the class, "2nd (i)".
#	Here, L_DF is the final number of sites in the "D"-region, 
#	L_AI is the initial number of sites in the "A"-region,
#	and t_I and t_F are the initial and final time-points, respectively.
#	The script was created by taking advantage of the main framework of another servant script,
#	'cmpt_mu_A2empty.pl' (in "../mu_A2empty/").
#	[CREATION STARTED on 2020/02/08, ENDED on 2020/02/11.]
#
# cmpt_mu_A2empty.pl:
#	This servant script computes the "extended" multiplication factors
#	for 'A -> {emptyset}':
#	mu_{A->emptyset} [ (L_I, t_I); (0, 0) ; t_2](sigmaBE=2),
#	where L_I is the number of sites in the "A"-region at the 'initial' time t_I,
#	and t_2 is some time-point after the deletion of the "A"-region.
#	(Actually, a "D"-region will be created somewhere in [t_2, t_2 + Delta{}t_2].)
#	The script was created by taking advantage of the main framework of another servant script,
#	'cmpt_mu_Dcr.pl' (in "../../Basic/mu_Dcr/"),
#	as well as of the main process of 'cmpt_mu_Adel.pl' (in "../../Basic/mu_Adel/").
#	[CREATION STARTED on 2020/02/07, ENDED on 2020/02/08.]
#
# cmpt_mu_Dcr.pl:
#	This servant script computes the multiplication factors
#	for the creation of a "D"-region:
#	mu_{D-cr} [ (x, [t_1, t_1 + Delta{}t]) ; (L, t_2) ],
#	where x is the position at which the "D"-region is created,
#	[t_1, t_1 + Delta{}t] is the time-interval within which the D-creation occurred,
#	and L is the number of sites in the "D"-region at the time-point t_2 ( > t_1).
#	The script was created by taking advantage of the main framework of another servant script,
#	'cmpt_muP0.pl' (in "../muP0/") .
#	[CREATION STARTED on 2020/02/03, ENDED on 2020/02/04; Modified on 2020/02/05]
#


	##### (IMPORTED w/ modifications from '../mu_A2empty/cmpt_mu_A2empty.pl' on 2020/02/08) (HD & CL & CHK) #####

use strict;
use warnings;

use MyFile_manipu qw(
	open_input_file2
	open_output_file
	list_sessions_ext
	);

use MyLASTPIECE_IO qw(
	readin_indel_params
	readin_2B_graft_convoluted_1st
	readin_2B_graft_convoluted_2nd
	);
=A BUG fixed on 2020/02/16.
	readin_2B_graft_convoluted_2st
	);
=cut # End of "A BUG fixed on 2020/02/16."
	#
	# print_muP_case_i
	# print_muP0s
	# readin_indel_rates
	# print_muP_case_ii_or_iii 
	# readin_muP0s
	# readin_muP_case_i
	# readin_muP_case_ii_or_iii
	# readin_mu_Adel
	#
	# REMOVED 'readin_muP_case_ii_or_iii' & 'readin_mu_Adel', in creation of 'graft_convolution.pl' (on 2020/02/15). #
	# ADDED 'readin_2B_graft_convoluted_1st' & 'readin_2B_graft_convoluted_2st', in creation of 'graft_convolution.pl' (on 2020/02/15). #
	# => The 'readin_2B_graft_convoluted_2st' was CORRECTED to be 'readin_2B_graft_convoluted_2nd' (on 2020/02/16).

##use MyLASTPIECE_Sppl qw(
##	cal_delta_exp_del_5p2
##	);


##sub create_dir_rw (\$); # Added on 2020/01/29.


$| = 1; # Flush the buffer for standard output as soon as it receives data.

my $bin_perl = "/usr/bin/perl"; # The Perl interpreter. # (Added on 2020/01/29). #

my $program_path = $0;
my ($program_name) = ($program_path =~ m!([^/]+)$!);

	# Controlling the level of test outputs. #
	# (0: silent; 1: very quiet; 2: quiet; 3: medium; 4: verbose; 5: very verbose) 
##our $TEST_LEVEL = 2;
our $TEST_LEVEL = 5;

$MyLASTPIECE_IO::TEST_LEVEL = $TEST_LEVEL ; # Added on 2020/01/27. # RESURRECTED on 2020/02/04.


		### END of "(IMPORTED w/ modifications from '../../../lastpiece.alpha.pl' on 2020/01/30) (HD)" ###

=For the predecessor, 'cmpt_mu_Dcr.pl'. (GL)
our $EPS = 1E-07; # A "small" number to be used for checking the consistency. # (Added on 2020/02/01.) #

our $IF_DRESSED = 0; # Whether the input indel rates are raw (=0) or dressed (=1). (The default value = 0.)  # Added in creation of 'cmpt_mu_Dcr.pl' (on 2020/02/04). #
=cut # End of "For the predecessor, 'cmpt_mu_Dcr.pl'. (GL)"

	####### Take in the Command-Line Arguments. #######


		# Modified in creation of 'wgtd_summs_3Darrays_for_mfacs.pl' (on 2020/02/29) (CL1) #

unless (@ARGV > 6) {
	print STDERR "\nUsage:  [perl] $program_name {initial_time (non-negative number)} {final_time (positive number, greater than initial_time)} {number_of_sub_time_intervals (positive integer)} {upper-bound_of_number_of_sites} {weights,assigned,to,the,input,3D-arrays} {input_file_for_muP_1} {input_file_for_muP_2} .. {input_file_for_muP_\#{weights}} {output_file} ";
	print STDERR "\n\n";
	print STDERR "\t\t[NOTE: The number of weights MUST EQUAL the number of input files!!]\n\n";
	die "$!\n\n";
}


our ($TIME_INIT, $TIME_FIN, $CT_SB_INTVLS, $UB_CT_SITES_CMPT) = @ARGV[0 .. 3];

my @weights = split /\,/, $ARGV[3+1];
my $ct_weights = @weights;

unless (@ARGV == 6 + $ct_weights) {
	print STDERR "\nUsage:  [perl] $program_name {initial_time (non-negative number)} {final_time (positive number, greater than initial_time)} {number_of_sub_time_intervals (positive integer)} {upper-bound_of_number_of_sites} {weights,assigned,to,the,input,3D-arrays} {input_file_for_muP_1} {input_file_for_muP_2} .. {input_file_for_muP_\#{weights}} {output_file} ";
	print STDERR "\n\n";
	print STDERR "\t\t[NOTE: The number of weights MUST EQUAL the number of input files!!]\n\n";
	die "$!\n\n";
}

my @infiles_muPs = @ARGV[5 .. 3+1+$ct_weights];
my $ofile_muPs_sum = $ARGV[5 + $ct_weights]; 

=For the predecessor, 'graft_convolution.pl'. (CL1)
unless (@ARGV > 7) {
	print STDERR "\nUsage:  [perl] $program_name {initial_time (non-negative number)} {final_time (positive number, greater than initial_time)} {number_of_sub_time_intervals (positive integer)} {upper-bound_of_number_of_sites} {input_file_for_indel_parameters} {input_file_for_muPs_2B_graft-convoluted_1st} {input_file_for_muPs_2B_graft-convoluted_2nd}  {output_file} ";
	print STDERR "\n\n";
	die "$!\n\n";
}
=cut # End of "For the predecessor, 'graft_convolution.pl'. (CL1)"
=For the predecessor, 'cmpt_mu_Dcr.pl'. (CL1)
unless (@ARGV > 8) {
	print STDERR "\nUsage:  [perl] $program_name {initial_time (non-negative number)} {final_time (positive number, greater than initial_time)} {number_of_sub_time_intervals (positive integer)} {upper-bound_of_number_of_sites} {SIGMA_BE (MUST ALWAYS BE 2!!)} {input_file_for_indel_parameters} {input_file_for_raw_indel_rates} {input_file for muP0 (w/ SIGMA_BE=2)} {output_file} "; # CORRECTED on 2020/02/06. # 
		# End of "Modified in creation of 'cmpt_muP0.pl' (on 2020/02/03) (CL1)" #
	print STDERR " [ options ] ,\n";
	print STDERR "   where the [options] can be either of the following:\n\n";
	print STDERR "            --RAW      when using the RAW indel rates;\n";
	print STDERR "            --DRESSED      when using the DRESSED indel rates.\n";
	print STDERR "		  (NOTE1: If both options are used, the '--RAW' will be prioritized.)\n"; # (The '\n";' was added in debugging on 2020/02/05.)
	print STDERR "		  (NOTE2: If no options are used, the '--RAW' will be assumed.)\n"; # (The '\n";' was added in debugging on 2020/02/05.)
	##print STDERR "           --XLEVEL=integer       for the level of MSA space exploration (from =1 (performing only multiple-\"shift\"s) to =5 (performing all programed moves) (default=$XLEVEL);\n"; # ADDED on 2019/08/01.
	print STDERR "\n";

	die "$!\n\n";
}
=cut # End of "For the predecessor, 'cmpt_mu_Dcr.pl'. (CL1)"

		# End of "Modified in creation of 'wgtd_summs_3Darrays_for_mfacs.pl' (on 2020/02/29) (CL1)" #


=For the predecessor, 'graft_convolution.pl'. (CL2)
		# Modified in creation of 'graft_convolution.pl' (on 2020/02/15) (CL2) #

##our ($TIME_INIT, $TIME_FIN, $CT_SB_INTVLS, $UB_CT_SITES_CMPT, $SIGMA_BE) = @ARGV[0 .. 3+1];
##my ($infile_id_prms, $infile_mu_Dcr, $infile_mu_Adel, $ofile_mu_Dcr2Adel) = @ARGV[5 .. 8];
our ($TIME_INIT, $TIME_FIN, $CT_SB_INTVLS, $UB_CT_SITES_CMPT) = @ARGV[0 .. 3];
my ($infile_id_prms, $infile_muPs_in_1st, $infile_muPs_in_2nd, $ofile_muPs_out) = @ARGV[3+1 .. 7];
		#
		# End of "Modified in creation of 'graft_convolution.pl' (on 2020/02/15) (CL2)" #
=cut # End of "For the predecessor, 'graft_convolution.pl'. (CL2)"

=For the predecessor, 'cmpt_mu_Dcr.pl'. (CL2)
		# Added in creation of 'cmpt_mu_Dcr.pl' (on 2020/02/04). (CL) #
my ($if_dressed, $if_raw) = (0, 0);

for (my $k=9; $k < @ARGV; $k++) { # Preliminary for-loop (over additional command-line arguments). #

	my $clarg = $ARGV[$k];
	unless ($clarg =~ s/^(\-)+//) { next; } # Skip if the argument lacks preceding '-'s. #

	my ($option, $value) = split /\=/, $clarg;
	unless ((defined $value) and ($value ne '')) {
	     print STDOUT "[CAUTION '$program_name'] Found an UNEXPECTED FORMAT of the commmand-line argument, ", $ARGV[$k], ": $! ! ... I will SKIP it ...\n\n";
	     next; # Skip an argument in an unexpected format.
	}
	my $org_option = $option;
	$option = uc $option; # Capitalize the option, in advance.

	if ($option eq 'RAW') {
	    $if_raw = 1;

	} elsif ($option eq 'DRESSED') {
	    $if_dressed = 1;

	} else {
	    print STDOUT "[CAUTION '$program_name'] Found an UNEXPECTED OPTION, '$org_option', within the cmmmand-line argument, ", $
ARGV[$k], ": $! ! ... I will SKIP it ...\n\n";
	    next; # Skip an argument with an unexpected option.
	}

} # End of the "Preliminary for-loop (over additional command-line arguments)." #

if ($if_dressed and !($if_raw)) {
	$IF_DRESSED = 1;

	print STDOUT "\n[INFORMATION '$program_name'] AS YOU SPECIFIED, we set \$IF_DRESSED to be 1 (true) ...\n\n";
}
		# End of "Added in creation of 'cmpt_mu_Dcr.pl' (on 2020/02/04). (CL)" #
=cut # End of "For the predecessor, 'cmpt_mu_Dcr.pl'. (CL2)"


	####### END of "Take in the Command-Line Arguments." #######



	### Check the existence & permission mode of the input files. ###

##foreach my $infile ($infile_id_prms, $infile_muPs_in_1st, $infile_muPs_in_2nd) { # Yet-Yet Further Modified in creation of 'graft_convolution.pl' (on 2020/02/15). (CHK) #
foreach my $infile (@infiles_muPs) { # Yet-Yet-Yet Further Modified in creation of 'wgtd_summs_3Darrays_for_mfacs.pl' (on 2020/02/29). (CHK) #
    unless ( ( -f $infile ) and ( -r $infile ) ) {
	print STDOUT "[ERROR '$program_name'] CanNOT FIND a readable input file, \"$infile\": $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
    }
}


	### Declare the Beginning of the Script ###

print STDOUT "\n<<< Program name: $program_name (path: $program_path) >>>\n\n";

		# MODIFIED in creation of 'wgtd_summs_3Darrays_for_mfacs.pl' (on 2020/02/29). (DC1) #

print STDOUT '#{Input files} (= #{weights}) = ', $ct_weights, " .\n\n";

print STDOUT "\n<< Input Files and Weights: >>\n\n";

print STDOUT join ("\t", 'No.', 'weight', '{path/to/the/input/file}'), "\n\n";

for (my $k =0; $k < $ct_weights; $k++) {

	print STDOUT join ("\t", $k+1, $weights[$k], $infiles_muPs[$k]), "\n\n";

}

=For the predecessor, 'graft_convolution.pl'. (DC1)
print STDOUT "\n<< Input Files: >>\n\n";

		# Modified in creation of 'graft_convolution.pl' (on 2020/02/15). (DC1) #
		#
##print STDOUT " (for indel parameters) $infile_id_prms ,\n";
##print STDOUT " (for multiplication factors for the creation of a \"D\"-region)  $infile_mu_Dcr ,\n";
##print STDOUT " (for mutiplication factors for the creation of an \"A\"-region) $infile_mu_Adel .\n\n";
print STDOUT " (for indel parameters) $infile_id_prms ,\n";
print STDOUT " (for 1st set of multiplication factors to be graft-convoluted)  $infile_muPs_in_1st ,\n";
print STDOUT " (for 2nd set of multiplication factors to be graft-convoluted)  $infile_muPs_in_2nd .\n\n";
		#
		# End of "Modified in creation of 'graft_convolution.pl' (on 2020/02/15). (DC1)" #
=cut # END of "For the predecessor, 'graft_convolution.pl'. (DC1)"

		# END of "MODIFIED in creation of 'wgtd_summs_3Darrays_for_mfacs.pl' (on 2020/02/29). (DC1)" #

print STDOUT "\n<< Output File: >>\n\n"; 

		# Modified in creation of 'wgtd_summs_3Darrays_for_mfacs.pl' (on 2020/02/29). (DC2) #
		#
##print STDOUT " (for multiplication factors resulting from the graft-convolution)  $ofile_muPs_out .\n\n";
print STDOUT " (for element-wise weighted summations of the input multiplication factors)  $ofile_muPs_sum .\n\n";
		#
		# End of "Modified in creation of 'wgtd_summs_3Darrays_for_mfacs.pl' (on 2020/02/29). (DC2)" #


print STDOUT "\n<< Global Variables specifying the levels of tests: >>\n\n"; # Added on 2020/01/30. (DC2) #

print STDOUT "\n", '$TEST_LEVEL = ', $TEST_LEVEL, " .\n\n";

print STDOUT '$MyLASTPIECE_IO::TEST_LEVEL  = ', $MyLASTPIECE_IO::TEST_LEVEL , " .\n\n"; # Added on 2020/02/04.


print STDOUT "\n<< Parameters necessary for recursive computations of the multiplication factors: >>\n\n";

print STDOUT '($TIME_INIT, $TIME_FIN) = ', "($TIME_INIT, $TIME_FIN) .\n\n";
print STDOUT '#{Sub-time-intervals} = ', $CT_SB_INTVLS, " .\n\n";

print STDOUT '{Upper-bound of #{sites} for computation} = ', $UB_CT_SITES_CMPT, " .\n\n"; 



	## Preliminary 00-C: Compute Other Important Parameters ##

print STDOUT "\n<<< Preliminary Process 00-C: Computing Other Important Parameters >>>\n\n";

our $SIZE_FULL_INTVL = $TIME_FIN - $TIME_INIT;
our $SIZE_SB_INTVL = $SIZE_FULL_INTVL /$CT_SB_INTVLS;

our @TIME_LAPSES_CMPT = (); # Values of the time lapse (= t_F - t_I) at which the factors will be computed.
for (my $i = 0; $i <= $CT_SB_INTVLS; $i++) {
	$TIME_LAPSES_CMPT[$i] = $i * $SIZE_SB_INTVL;
} 

print STDOUT '$SIZE_FULL_INTVL = ', $SIZE_FULL_INTVL, " \;\n";
print STDOUT '$SIZE_SB_INTVL = ', $SIZE_SB_INTVL, " .\n\n";

print STDOUT '@TIME_LAPSES_CMPT = (', join (', ', @TIME_LAPSES_CMPT), ") .\n\n";

print STDOUT "\n";


=For the predecessor, 'graft_convolution.pl'. (00-IN-A)

print STDOUT "\n<<< Preliminary Process 00-IN: Reading in the quantities necessary for the computations >>>\n\n";

	## 00-IN-A: Read in the input file for indel parameters. ##
	## (IMPORTED from 'cmpt_mu_Dcr.pl', in creation of 'cmpt_muP_2nd_i.pl' (on 2020/02/10). (00-IN)

print STDOUT "\n<< (Pre 00-IN-A): Reading in the indel model parameters >>\n\n";


our ($TOTRATE_INS, $TOTRATE_DEL, $CUTOFF_INS_LEN, $CUTOFF_DEL_LEN)
   = readin_indel_params ($infile_id_prms);

unless (defined $CUTOFF_DEL_LEN) {
	print STDOUT "[ERROR '$program_name'] The subroutine, 'readin_indel_params', FAILED on the input file, \"$infile_id_prms\" $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}
(print STDOUT "\n ... The subroutine, 'readin_indel_params', Ended on the input file, \"$infile_id_prms\" ...\n\n") if ($TEST_LEVEL > 1);


print STDOUT "\n[ Indel Model Parameters retrieved: ]\n\n";

print STDOUT '($TOTRATE_INS, $TOTRATE_DEL) = ', "($TOTRATE_INS, $TOTRATE_DEL) .\n";
print STDOUT '($CUTOFF_INS_LEN, $CUTOFF_DEL_LEN) = ', "($CUTOFF_INS_LEN, $CUTOFF_DEL_LEN) .\n\n";

	## End of "(IMPORTED from 'cmpt_mu_Dcr.pl', in creation of 'cmpt_muP_2nd_i.pl' (on 2020/02/10). (00-IN)"

=cut # End of "For the predecessor, 'graft_convolution.pl'. (00-IN-A)"



my $ct_tlapses = @TIME_LAPSES_CMPT;
##my $ct_tlapses = $CT_SB_INTVLS +1;
my $cap_ct_sites = $UB_CT_SITES_CMPT +1;


=For the predecessor, 'cmpt_mu_Dcr.pl'. (00-IN)
	## (MODIFIED on 2020/01/31) (00-IN) ##

print STDOUT "\n<<< Preliminary Process 00-IN: Reading in the quantities necessary for the computations >>>\n\n";

	## 00-IN-A: Read in the input file for indel parameters. ##

print STDOUT "\n<< (Pre 00-IN-A): Reading in the indel model parameters >>\n\n";

	# [ Omitted Actual Commands ... ] #


	## Preliminary 00-B: Read in the input file for indel rates. ##

		# Modified in creation of 'cmpt_mu_Dcr.pl' (on 2020/02/04). (00-B-1) #
if ($IF_DRESSED) {
	print STDOUT "\n<< (Pre 00-IN-B): Reading in the DRESSED indel rates >>\n\n";
} else {
	print STDOUT "\n<< (Pre 00-IN-B): Reading in the RAW indel rates >>\n\n";
}
	# [ Omitted Actual Commands ... ] #


	## 00-IN-C: Read in the input file for muP0s (w/ $SIGMA_BE = 2). ##
	## (Added in creation of 'cmpt_mu_Dcr.pl' on 2020/02/04&05). (00-IN-C) ##

print STDOUT "\n<< (Pre 00-IN-C): Reading in the muP0s (w/ \$SIGMA_BE = 2) >>\n\n";

	# [ Omitted Actual Commands ... ] #

	## End of "(MODIFIED on 2020/01/31) (00-IN)" ##

=cut # END of "For the predecessor, 'cmpt_mu_Dcr.pl'. (00-IN)"


=For the predecessor, 'graft_convolution.pl'. (00-IN-H1&2)

	### Preliminary 00-IN-H1: Reading in the 1st set of multiplication factors to be graft-convoluted.
	### (ADDED in creation of 'graft_convolution.pl' on 2020/02/15). (00-IN-H1) ###

print STDOUT "\n<< (Pre 00-IN-H1): Reading in the 1st set of multiplication factors to be graft-convoluted >>\n\n";

my ($rf_muPs_1st) =  readin_2B_graft_convoluted_1st ($infile_muPs_in_1st, $CT_SB_INTVLS, $UB_CT_SITES_CMPT);

	# $rf_muPs_1st->[$LI][$LF][$jt]
	# (w/ 0 < $LI, $LF <= $ub_ct_sites, & 0 <= $jt <= $ct_sb_intvls)
	# is the multiplication factor, in the 1st set to be graft-convoluted,
	# at the $jt th time-lapse with $L_I sites initially and $L sites finally, both in between the PASs.


unless (defined $rf_muPs_1st) {
	print STDOUT "[ERROR '$program_name'] The subroutine, 'readin_2B_graft_convoluted_1st', FAILED on the input file, \"$infile_muPs_in_1st\" $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}
(print STDOUT "\n ... The subroutine, 'readin_2B_graft_convoluted_1st', Ended on the input file, \"$infile_muPs_in_1st\" ...\n\n") if ($TEST_LEVEL > 1);

		# Check the consistency of @{$rf_muPs_1st}. # (Added on 2020/02/15.) #

for (my $LI = 1; $LI < $cap_ct_sites; $LI++) { # Preceding outer for-loop (over the initial site-counts). #

	my $muPs_1st_fxd_LI = $rf_muPs_1st->[$LI];

   for (my $LF = 1; $LF < $cap_ct_sites; $LF++) { # Preceding middle for-loop (over the final site-counts). #
	my $ct_tlapses_in = @{$muPs_1st_fxd_LI->[$LF]};

	unless ($ct_tlapses_in == $ct_tlapses) {
		print STDOUT "[ERROR '$program_name'] The number of time-lapses (=$ct_tlapses_in) in the input ", '@{$rf_muPs_1st->[$LI][$LF]} (w/ ($LI, $LF) =', "($LI, $LF) )  CLASHES with the number of time-lapses for the output (=$ct_tlapses) : $? !\n";
		print STDERR "[ERROR '$program_name'] $!\n";
		die ;
	}
   } # End of the Preceding inner for-loop (over the final site-counts). #

}  # End of the Preceding outer for-loop (over the initial site-counts). #

		# End of "Check the consistency of @{$rf_muPs_1st}. # (Added on 2020/02/15.)" #


if ($TEST_LEVEL >= 5) { # (00-IN-H1) #

	print STDOUT "\n\n[[ Multiplication factors, muP_{1st} [ (L_F, t_F) | (L_I, t_I) ], retrieved (& stored in ", '@{$rf_muPs_1st}', "): ]]\n\n";

	print STDOUT join ("\t", '$LI:$LF\$jt', 0 .. $CT_SB_INTVLS), "\n\n";


	for (my $LI = 1; $LI < $cap_ct_sites; $LI++) { # Test outer for-loop (over the initial site-counts).

	    for (my $LF = 1; $LF < $cap_ct_sites; $LF++) { # Test middle for-loop (over the final site-counts).
		##print STDOUT join ("\t", "$LI:$$LF", @{$rf_muPs_1st->[$LI][$LF]}) , "\n"; # A BUG fixed on 2020/02/16. #
		print STDOUT join ("\t", "$LI:$LF", @{$rf_muPs_1st->[$LI][$LF]}) , "\n";

	    } # End of the test middle for-loop (over the final site-counts).

	} # End of the Test outer for-loop (over the initial site-counts).

	print STDOUT "\n\n\n";

} # End of "if ($TEST_LEVEL >= 5) { # (00-IN-H1) # ...}"

	### END of "(ADDED in creation of 'graft_convolution.pl' on 2020/02/15). (00-IN-H1)" ###


	### Preliminary 00-IN-H2: Reading in the 2nd set of multiplication factors to be graft-convoluted.
	### (ADDED in creation of 'graft_convolution.pl' on 2020/02/15). (00-IN-H2) ###

##print STDOUT "\n<< (Pre 00-IN-H1): Reading in the 1st set of multiplication factors to be graft-convoluted >>\n\n";
print STDOUT "\n<< (Pre 00-IN-H2): Reading in the 2nd set of multiplication factors to be graft-convoluted >>\n\n"; # CORRECTED on 2020/02/28.

my ($rf_muPs_2nd) =  readin_2B_graft_convoluted_2nd ($infile_muPs_in_2nd, $CT_SB_INTVLS, $UB_CT_SITES_CMPT);

	# $rf_muPs_2nd->[$LF][$LI][$jt]
	# (w/ 0 < $LI, $LF <= $ub_ct_sites, & 0 <= $jt <= $ct_sb_intvls)
	# is the multiplication factor, in the 2nd set to be graft-convoluted,
	# at the $jt th time-lapse with $L_I sites initially and $L sites finally, both in between the PASs.


unless (defined $rf_muPs_2nd) {
	print STDOUT "[ERROR '$program_name'] The subroutine, 'readin_2B_graft_convoluted_2nd', FAILED on the input file, \"$infile_muPs_in_2nd\" $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}
(print STDOUT "\n ... The subroutine, 'readin_2B_graft_convoluted_2nd', Ended on the input file, \"$infile_muPs_in_2nd\" ...\n\n") if ($TEST_LEVEL > 1);

		# Check the consistency of @{$rf_muPs_2nd}. # (Added on 2020/02/15.) #

for (my $LF = 1; $LF < $cap_ct_sites; $LF++) { # Preceding outer for-loop (over the final site-counts). #

	my $muPs_2nd_fxd_LF = $rf_muPs_2nd->[$LF];

   for (my $LI = 1; $LI < $cap_ct_sites; $LI++) { # Preceding middle for-loop (over the initial site-counts). #
	my $ct_tlapses_in = @{$muPs_2nd_fxd_LF->[$LI]};

	unless ($ct_tlapses_in == $ct_tlapses) {
		print STDOUT "[ERROR '$program_name'] The number of time-lapses (=$ct_tlapses_in) in the input ", '@{$rf_muPs_2nd->[$LF][$LI]} (w/ ($LI, $LF) =', "($LI, $LF) )  CLASHES with the number of time-lapses for the output (=$ct_tlapses) : $? !\n";
		print STDERR "[ERROR '$program_name'] $!\n";
		die ;
	}
   } # End of the Preceding inner for-loop (over the initial site-counts). #

}  # End of the Preceding outer for-loop (over the finaal site-counts). #

		# End of "Check the consistency of @{$rf_muPs_2nd}. # (Added on 2020/02/15.)" #


if ($TEST_LEVEL >= 5) { # (00-IN-H2) #

	print STDOUT "\n\n[[ Multiplication factors, muP_{2nd} [ (L_F, t_F) | (L_I, t_I) ], retrieved (& stored in ", '@{$rf_muPs_2nd}', "): ]]\n\n";

	print STDOUT join ("\t", '$LI:$LF\$jt', 0 .. $CT_SB_INTVLS), "\n\n";


	for (my $LI = 1; $LI < $cap_ct_sites; $LI++) { # Test outer for-loop (over the initial site-counts).

	    for (my $LF = 1; $LF < $cap_ct_sites; $LF++) { # Test middle for-loop (over the final site-counts).
		##print STDOUT join ("\t", "$LI:$$LF", @{$rf_muPs_2nd->[$LF][$LI]}) , "\n"; # A BUG fixed on 2020/02/16. #
		print STDOUT join ("\t", "$LI:$LF", @{$rf_muPs_2nd->[$LF][$LI]}) , "\n";

	    } # End of the test middle for-loop (over the final site-counts).

	} # End of the Test outer for-loop (over the initial site-counts).

	print STDOUT "\n\n\n";

} # End of "if ($TEST_LEVEL >= 5) { # (00-IN-H1) # ...}"

	### END of "(ADDED in creation of 'graft_convolution.pl' on 2020/02/15). (00-IN-H1)" ###

=cut # END of "For the predecessor, 'graft_convolution.pl'. (00-IN-H1&2)"




	### Preliminary 00-OUT: Initialize the output 3D-array, @muP_wsum. ###
	### (ADDED in creation of 'wgtd_summs_3Darrays_for_mfacs.pl' on 2020/02/29). (00-OUT) ###

print STDOUT "\n<< (Pre 00-OUT): Initialize the output 3D-array, \@muP_wsum >>\n\n";

my @muP_wsum = ([]);

	# muP_i [ (L_F = $LF, t_F) | (L_I = $LI, t_I) ], with (i = 1, 2, ...),
	# that the number of sites is L_I at the initial time (=t_I) and L_F at the final time (=t_F),
	# with 1 <= $LI, $LF <= $UB_CT_SITES,
	# read-out in the form: $muP_i[$LI]->[$LF][$jt], where $jt = (t_F - t_I)/$SIZE_SB_INTVL.
	# The element-wise weighted summations are given by:
	#
	# $muP_wsum[$LI]->[$LF][$jt] = Sum_{i=1}^{n} wgt_i * $muP_i[$LI]->[$LF][$jt] .

for (my $LI = 1; $LI < $cap_ct_sites; $LI++) {
    my @array_2D = ([]);
    for (my $LF = 1; $LF < $cap_ct_sites; $LF++) {
	my @vec_init = ();
	for (my $jt = 0; $jt < $ct_tlapses; $jt++) { $vec_init[$jt] = 0; } # 
	 $array_2D[$LF] = \@vec_init;
    }
    $muP_wsum[$LI] = \@array_2D;
}

	### END of "(ADDED in creation of 'wgtd_summs_3Darrays_for_mfacs.pl' on 2020/02/29). (00-OUT)" ###


	###  Main Process: Compute multiplication factors, 
	### muP_wsum [ (L_F, t_F) | (L_I, t_I) ], via Element-wise Weighted Summations.
	### (ADDED in creation of 'wgtd_summs_3Darrays_for_mfacs.pl' (on 2020/02/29).) (MAIN) ###

print STDOUT "\n<<<<<  Main Process: Compute multiplication factors, muP_wsum [ (L_F, t_F) | (L_I, t_I) ], via Element-wise Weighted Summations >>>>>\n\n\n";

for (my $k=0; $k <$ct_weights; $k++) { # MAIN Outer for-loop (over the input files). #

	my $number = $k + 1;
	my $infile_muPs = $infiles_muPs[$k];
	my $wgt = $weights[$k];

   (print STDOUT "\n<< The $number th input file (w/ weight=$wgt) >>\n\n") if ($TEST_LEVEL > 0);

		# (IMPORTED w/ modifications from 'readin_2B_graft_convoluted_1st' in "MyLASTPIECE_IO.pm",
		# in creation of 'wgtd_summs_3Darrays_for_mfacs.pl' (on 2020/02/29).) (MAIN) #

			### RESTART FROM HERE ###

	my $fh_in = open_input_file2 ($infile_muPs);
	unless (defined $fh_in) {
	    ##print STDOUT "[ERROR 'readin_2B_graft_convoluted_1st'] The subroutine, 'open_input_file2', FAILED to open a file-handle for the file \"$infile_muPs\": $! !\n";
	    print STDOUT "[ERROR '$program_name'] The subroutine, 'open_input_file2', FAILED to open a file-handle for the file \"$infile_muPs\": $! !\n"; # CORRECTED on 2020/03/01. (1) #
	    return (undef);
	}

	while (my $line = <$fh_in>) { # Main middle while-loop (over input lines). #

	   my ($LI, $LF) = ($line =~ m/^(\d+)\:(\d+)\t/);
	   unless (defined $LI) { next; } # Skip lines without multiplication factors. #

	   unless ((0 < $LI) and ($LI < $cap_ct_sites) and (0 < $LF) and ($LF < $cap_ct_sites)) {
		my $ub_ct_sites = $UB_CT_SITES_CMPT;
		##print STDOUT "[CAUTION 'readin_2B_graft_convoluted_1st'] In the $number th input file, Found a row whose (\$LI, \$LF) = ($LI, $LF) is NOT WITHIN the specified range, [1, $ub_ct_sites] x [1, $ub_ct_sites]: $! ! ... I will SKIP this row ...\n\n";
		print STDOUT "[CAUTION '$program_name'] In the $number th input file, Found a row whose (\$LI, \$LF) = ($LI, $LF) is NOT WITHIN the specified range, [1, $ub_ct_sites] x [1, $ub_ct_sites]: $! ! ... I will SKIP this row ...\n\n"; # CORRECTED on 2020/03/01. (2) #
		next;
	   }

	   ##$line =~ s/\n$//;
	   $line =~ s/\s+$//; # Modified on 2020/02/08. #

	   my @fields = split /\t/, $line;
	   my $label = shift @fields; 
	   unless (@fields == $ct_tlapses) {
		my $ct_flds = @fields;
		##print STDOUT "[CAUTION 'readin_2B_graft_convoluted_1st'] In the $number th input file, in the row labelled '$label', the number of fields (=$ct_flds) is NOT EQUAL to the specified number of Time-Lapses (=$ct_tlapses): $! ! ... I will SKIP this row ...\n\n";
		print STDOUT "[CAUTION '$program_name'] In the $number th input file, in the row labelled '$label', the number of fields (=$ct_flds) is NOT EQUAL to the specified number of Time-Lapses (=$ct_tlapses): $! ! ... I will SKIP this row ...\n\n"; # CORRECTED on 2020/03/01. (3) #
		next;
	   }

	   my $vec_muPs_wsum = $muP_wsum[$LI]->[$LF];

		# muP_i [ (L_F = $LF, t_F) | (L_I = $LI, t_I) ], with (i = 1, 2, ...),
		# that the number of sites is L_I at the initial time (=t_I) and L_F at the final time (=t_F),
		# with 1 <= $LI, $LF <= $UB_CT_SITES,
		# read-out in the form: $muP_i[$LI]->[$LF][$jt], where $jt = (t_F - t_I)/$SIZE_SB_INTVL.
		# The element-wise weighted summations are given by:
		#
		# $muP_wsum[$LI]->[$LF][$jt] = Sum_{i=1}^{n} wgt_i * $muP_i[$LI]->[$LF][$jt] .

	   for (my $jt =0; $jt < $ct_tlapses; $jt++) { $vec_muPs_wsum->[$jt] += $wgt * $fields[$jt]; }

	}	# End of the main middle while-loop (over input lines). #

	close ($fh_in);

		# END of "(IMPORTED w/ modifications from 'readin_2B_graft_convoluted_1st' in "MyLASTPIECE_IO.pm",
		# in creation of 'wgtd_summs_3Darrays_for_mfacs.pl' (on 2020/02/29).) (MAIN)" #

} # End of the MAIN Outer for-loop (over the input files). #

	### END of "(ADDED in creation of 'wgtd_summs_3Darrays_for_mfacs.pl' (on 2020/02/29).) (MAIN)" ###

## my @weights = split /\,/, $ARGV[3+1];
## my $ct_weights = @weights;
## my @infiles_muPs = @ARGV[5 .. 3+1+$ct_weights];
## my $ofile_muPs_sum = $ARGV[5 + $ct_weights]; 


	###  Post-Main Process: Output the computed multiplication factors,  muP_wsum [ (L_F, t_F) | (L_I, t_I) ]. ###
	### (ADDED in creation of 'graft_convolution.pl' (on 2020/02/15&16).) (MAIN) ###
	### (=> Modified in creation of 'wgtd_summs_3Darrays_for_mfacs.pl' (on 2020/02/29).) (Post-MAIN) ###

print STDOUT "\n<<<<<  Post-Main Process: Output the computed multiplication factors, muP_wsum [ (L_F, t_F) | (L_I, t_I) ] >>>>>\n\n\n";

		# (Preliminary) Open a file-handle for outputting to the output file. #

my $fh_out = open_output_file ($ofile_muPs_sum, 0);
unless (defined $fh_out) {
	print STDOUT "[ERROR '$program_name'] The subroutine 'open_output_file' FAILED to open a file-handle for the file \"$ofile_muPs_sum\": $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}

##print $fh_out join ("\t", '#{A_sites}_Init:#{D_sites}_Fin\Time_Lapse', @TIME_LAPSES_CMPT), "\n\n";
##print $fh_out join ("\t", '#{sites}_Init(A):#{sites}_Fin(D)\Time_Lapse', @TIME_LAPSES_CMPT), "\n\n";
print $fh_out join ("\t", '#{sites}_Init:#{sites}_Fin\Time_Lapse', @TIME_LAPSES_CMPT), "\n\n";

for (my $LI = 1; $LI < $cap_ct_sites; $LI++) { # Main Outermost for-loop (over initial site-counts). #

    my $muPs_wsum_fxd_LI = $muP_wsum[$LI];

	# muP_i [ (L_F = $LF, t_F) | (L_I = $LI, t_I) ], with (i = 1, 2, ...),
	# that the number of sites is L_I at the initial time (=t_I) and L_F at the final time (=t_F),
	# with 1 <= $LI, $LF <= $UB_CT_SITES,
	# read-out in the form: $muP_i[$LI]->[$LF][$jt], where $jt = (t_F - t_I)/$SIZE_SB_INTVL.
	# The element-wise weighted summations are given by:
	#
	# $muP_wsum[$LI]->[$LF][$jt] = Sum_{i=1}^{n} wgt_i * $muP_i[$LI]->[$LF][$jt] .

    for (my $LF = 1; $LF < $cap_ct_sites; $LF++) { # Main outer for-loop (over final site-counts). #

	my $vec_muPs_wsum = $muPs_wsum_fxd_LI->[$LF];

	print $fh_out join ("\t", "$LI:$LF", @{$vec_muPs_wsum}), "\n" ;

    } # End of the Main outer for-loop (over final site-counts). #

} # End of the Main Outermost for-loop (over initial site-counts). #


close ($fh_out);

print STDOUT "\n\n === All The Processes of '$program_name' FINISHED !! ===\n\n\n";

		### END of "(ADDED in creation of 'graft_convolution.pl' (on 2020/02/15&16).) (MAIN) ###
		### (=> Modified in creation of 'wgtd_summs_3Darrays_for_mfacs.pl' (on 2020/02/29).) (Post-MAIN) ###





