package MyTreeManipu_VI;
#
###################################################################################
#	DENSERM_P: A package to DEtect Negative SElection on Recurrent Mutations
#			(almost exclusively via Perl modules and scripts).
#	Version 0.3: Copyright (C) 2012 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
###################################################################################
#
# MyTreeManipu_VI.pm:
#	This module conducts algebraically advanced tree manipulations,
#	such as those related to the estimate of branch lengths.
#	The first few subroutines intend to emulate the 'fitch' program
#	in the Phylip package, which conducts Fitch-Margoliash algorithm of
#	the branch length estimation (and the tree topology inference).
#
#	Most of the subroutines in this module are based on algorithms or methods described in:
#	"Molecular Evolution and Phylogenetics",
#	written by Nei M, Kumar S (Oxford University Press, Inc.; New York, NW, USA; 2000).
#
# MyTreeManipu_V.pm:
#	This package conducts sophisticated tree manipulations.
#	Specifically, it has subroutines related to phylogenetic reconciliations.
#
#
# MyTreeManipu_II.pm: This package can process more general types of trees,
#	(i.e. NOT limited to binary trees with trichotomous top nodes).
#	It is also consistent with the format of the taxonomy tree which are handled by
#	the module 'MyTaxonomy.pm'.
#
# MyTreeManipu.pm: This package performs basic tree manipulations.
#
#  Subroutines whose substitutes are available in this module ('MyTreeManipu_II.pm')
#	or in 'MyTaxonomy.pm':
#   our @EXPORT = qw(retrieve_top_node confine_outgroup print_unrooted);
#   our @EXPORT_OK = qw(switch_topnode  print_subtree topo_unrooted topo_subtree
#		list_sequences fetch_node_series dist_nodes
#		    );
#
# Subroutines whose substitutes are not available yet:
#   our @EXPORT_OK = ( retrieve_bootstrap  prod_bootstrap );
#

use strict;
use warnings;

use MyArrayManipu qw(
	diff_array
	);
 	## union_array nr_array compare_arrays
	## union_array

use MyFile_manipu qw(
	open_input_file
	);

use MyLinearAlgebra qw(
	solve_lin_eq_lu
	);

use MyProbDistri qw(
	cum_gaussian
	);

use MyTaxonomy qw(
	fetch_ext_offsprings_nr
	fetch_common_ancestors
	);
	## fetch_ancestry
	## fetch_last_common_ancestor fetch_last_common_ancestor_ov2
	## fetch_all_offsprings

use MyTreeMap qw(
	fractionate_by_ndepth
	);
	## assign_gnode2sp_lca
	## fetch_splca_comps assign_gnode2sp_lca_comp

use MyTreeManipu_II qw(
	br_draw_subtree
	);
=set aside for later use
use MyTreeManipu_II qw(
	make_node2parent shift_topnode move_topnode segregate_subset_nodes
	);
	## input_tree_file read_tree_file input_subtree copy_tree
	## output_tree draw_subtree
	## br_output_tree br_draw_subtree topo_subtree
	## make_node2children make_node2parent
	## sum_branch_lengths intra_lineage_distance inter_node_distance
	## fetch_boot_series extract_hidden_bootstrap extract_hidden_bootstrap2
	## resolve_multipartite
use MyTreeManipu_IV qw(
	construct_neat_subtree
	);
use MyTreeManipu_V qw(
	fetch_species_ancestor make_gpair2sp_lca
	examine_node_states inspect_med_branches
	reroot_moltree_heur fractionate_spe_dup_gpairs
	resolve_mlt_gnode_with_sptree resolve_multifurcation
	assign_gnode2spe_dup annotate_spe_dup
	sep_genesets_by_spnode_bu sep_genesets_by_spnode_td
	separate_into_ortho_sets collect_og_phylogene_sets
	find_closest_outgroups bch_attach_closest_ogs
	);
=cut

use Exporter;

our @ISA = qw(Exporter);

our @EXPORT_OK = qw(
	readin_phylip_dists extract_phylip_dists
	coeffs_leastsq_brlen leastsq_brlen
	assign_trivial_brlen leastsq_brlen_nn
	extended_upgma_tree extended_wtd_upgma_tree
	level_sglnode_upgma var_nodelevels_ext_wgt_upgma
	approx_var_brlen_ext_wgt_upgma
	brlen_ext_wgt_upgma apprx_bootspp_by_z_test
	) ;


our $TOP_NODE = $MyTreeManipu_II::TOP_NODE;
our $INDX_LENGTH = $MyTreeManipu_II::INDX_LENGTH;
our $INDX_BOOT = $MyTreeManipu_II::INDX_BOOT;

our $MAX_SPDEPTH = 1E+6; # The maximum depth in species trees,
			# which is assigned to the common ancestor of an empty set.
			# (An actual depth cannot be that large.)

our $UL_ITMLT = 20; # An upper limit to conduct iterative multiplications
		   # for calculating a power.


=Descriptions
### 5 important hash tables which describe a tree ###
#
# %node2parent = (node_ID => parent_ID, ...)  # top_node_ID => undef
# %node2children = (node_ID => [a list of child ID's], ...) # external_node_ID => undef
# %name2id = (node_name => node_ID, ...) # including $TOP_NODE => top_node_ID
# %id2name = (node_ID => node_name, ...)  # NOT including top_node_ID => $TOP_NODE,
#				 # because we can always identify the top-node by
#				# checking if $node2parent{$node_id} is defined or not.
#
# %branch_attributes = (ID of the nearest downstream node => [length, bootstrap, ...], ...)
#	--- swapped when the top node is shifted.
#
=cut


# readin_phylip_dists:
#	This subroutine reads in an output file
#	of a distance estimation program in the Phylip package.
#	In order to save as much memory as possible,
#	it will just read in the lower triangular part
#	of the distance matrix.
#
# INPUT: \$infile, where
#	$infile is the path to the input file for the distance matrix.
#
# OUTPUT: (\@seqnames, \@dists), where
#	$seqnames[$i] is the name of the $i th sequence;
#	$dists[$i]->[$j] is the distance between the $i th and the $j th sequences,
#	with (0 <= $j < $i < #{sequences}).
#	NOTE that $dists[0] is the reference to an empty array.
#
sub readin_phylip_dists (\$) {

	my ($infile) = @_;
	local $/ = "\n";

	my $fh = open_input_file (${$infile});

	my $header = <$fh>;
	my ($ct_seqs) = ($header =~ m/^\s+(\d+)/ );
	unless (defined $ct_seqs) {
	    print STDERR "[ERROR 'readin_phylip_dists'] ILLEGAL HEADER for the input file '${$infile}':\n";
	    print STDERR "    '${$infile}' .\n\n";
	    return (undef, undef);
	}

	my @seqnames = ();
	my @dists = ();
	my $curr_indx = -1;

	while (my $in = <$fh>) { # The while-loop #

	   if ($in =~ m/^\S/) { # The beginning of a row in the matrix
		$curr_indx++;
		my @fields = split /\s+/, $in;
		my $seqname = shift @fields;
		push @seqnames, $seqname;
		my @nec_dists = ($curr_indx < @fields) ?
			@fields[0 .. $curr_indx-1] : @fields ; 
		push @dists, \@nec_dists;
	   } else {
		my $last_dists = $dists[$#dists];
		my $ct_stored = @{$last_dists};
		($curr_indx <= $ct_stored) and next; # we don't have to take in the distances any longer.
		my @fields = split ' ' , $in;
		my $ct_nec = $curr_indx - $ct_stored;
		my @nec_dists = ($ct_nec < @fields) ?
			@fields[0 .. $ct_nec-1] : @fields ;
		push @{$last_dists}, @nec_dists;
	   }

	}		# End of the while-loop #
	close ($fh);

	unless (($ct_seqs == @seqnames) and ($ct_seqs == @dists)
	  and ($ct_seqs == $curr_indx + 1) and ($curr_indx == @{$dists[$#dists]})) {
	    my $ct_outseq = @seqnames;
	    my $size_dists = @dists;
	    my $size_last_dists = @{$dists[$#dists]};
	    print STDERR "[ERROR 'readin_phylip_dists'] FAILED to take in the input file '${$infile}':\n";
	    print STDERR "   {extected sequence count} = $ct_seqs ,\n";
	    print STDERR "   {actual input sequence count} = $ct_outseq ,\n";
	    print STDERR "   {actual size of the distance matrix) = $size_dists ,\n";
	    print STDERR "   {last index for the distance matrix) = $curr_indx ,\n";
	    print STDERR "   {size of the last row of the distance matrix} = $size_last_dists .\n\n";
	    return (undef, undef);
	}

	return (\@seqnames, \@dists);

}		# End of the subroutine 'readin_phylip_dists' #


# extract_phylip_dists:
#	This subroutine reads in an output chunk of a distance matrix
#	following the format of the Phylip package.
#	In order to save as much memory as possible,
#	it will just read in the lower triangular part
#	of the distance matrix.
#
# INPUT: (\*${$fh_in}, $ct_seqs), where
#	$fh_in is the filehandle providing a distance matrix;
#	$ct_seqs >0 gives the number of sequences,
#	if it is negative or undefined, the information is expected to be
#	extracted from the input.
#
# OUTPUT: (\@seqnames, \@dists, $distance_type), where
#	$seqnames[$i] is the name of the $i th sequence;
#	$dists[$i]->[$j] is the distance between the $i th and the $j th sequences,
#	with (0 <= $j < $i < #{sequences}).
#	NOTE that $dists[0] is the reference to an empty array.
#	$distance_type tells what type of distances are used when
#	the distance matrix is NOT available.
#
sub extract_phylip_dists (\*$) {

	my ($fh_in, $ct_seqs) = @_;
	local $/ = "\n";

=for DEBUGGING the subroutine
	print STDERR "\n< Subroutine 'extract_phylip_dists' with the initial \$ct_seqs = $ct_seqs >\n\n";
=cut

	my $prev_line;
	while ($prev_line = <$fh_in>) { ($prev_line =~ m/\S/) and last; } # Skip leading blank lines

	(defined $prev_line) or (return (undef, undef)); # Reached the end of the input file

		# Examine the signs that the distance matrix is NOT AVAILABLE #
	if ($prev_line =~ m/DISTANCES\s*$/) {
	    $prev_line =~ s/\s+$//;
	    return ([], [], $prev_line);
	}

		## ADDED on Tuesday, May 22, 2007 ##
	(uc ($prev_line) =~ m/NOT_AVAILABLE/)
		and (return ([], [], 'NOT_AVAILABLE'));

=for DEBUGGING the subroutine
	print STDERR "The first line: $prev_line .\n\n";
=cut

	unless ((defined $ct_seqs) and ($ct_seqs>=0)) {
	   while (1) { # The preliminary while-loop #
		my $in ;
		if (defined $prev_line) {
		   $in = $prev_line;
		   $prev_line = undef;
		} else {
		   $in = <$fh_in>;
		}
		(defined $in) or last;
		($ct_seqs) = ($in  =~ m/^\s+(\d+)/ );
		(defined $ct_seqs) or next;
		last;
	    }			# End of the preliminary while-loop #

=for DEBUGGING the subroutine
	    print STDERR "Fetched \$ct_seqs = $ct_seqs.\n\n";
=cut
	}
	unless ((defined $ct_seqs) and ($ct_seqs>0)) {
	    print STDERR "[ERROR 'extract_phylip_dists'] FAILED to EXTRACT the SEQUENCE COUNT !!\n\n";
	    return (undef, undef);
	}


	my @seqnames = ();
	my @dists = ();
	my $curr_indx = -1;

	while (1) { # The main while-loop #
	   my $in;
	   if (defined $prev_line) {
		$in = $prev_line;
		$prev_line = undef;
	   } else {
		$in = <$fh_in>;
	   }
	   (defined $in) or last;
	   ($in =~ m/\S/) or last;
		# The next blank line marks the end of the distance matrix. #

	   if ($in =~ m/^\S/) { # The beginning of a row in the matrix
		$curr_indx++;
		my @fields = split /\s+/, $in;
		my $seqname = shift @fields;
		push @seqnames, $seqname;
		my @nec_dists = ($curr_indx < @fields) ?
			@fields[0 .. $curr_indx-1] : @fields ; 
		push @dists, \@nec_dists;
	   } else {
		my $last_dists = $dists[$#dists];
		my $ct_stored = @{$last_dists};
		($curr_indx <= $ct_stored) and next; # we don't have to take in the distances any longer.
		my @fields = split ' ' , $in;
		my $ct_nec = $curr_indx - $ct_stored;
		my @nec_dists = ($ct_nec < @fields) ?
			@fields[0 .. $ct_nec-1] : @fields ;
		push @{$last_dists}, @nec_dists;
	   }
	}		# End of the main while-loop #

	unless (($ct_seqs == @seqnames) and ($ct_seqs == @dists)
	  and ($ct_seqs == $curr_indx + 1) and ($curr_indx == @{$dists[$#dists]})) {
	    my $ct_outseq = @seqnames;
	    my $size_dists = @dists;
	    my $size_last_dists = @{$dists[$#dists]};
	    print STDERR "[ERROR 'extract_phylip_dists'] FAILED to take in a distance matrix:\n";
	    print STDERR "   {extected sequence count} = $ct_seqs ,\n";
	    print STDERR "   {actual input sequence count} = $ct_outseq ,\n";
	    print STDERR "   {actual size of the distance matrix) = $size_dists ,\n";
	    print STDERR "   {last index for the distance matrix) = $curr_indx ,\n";
	    print STDERR "   {size of the last row of the distance matrix} = $size_last_dists .\n\n";
	    return (undef, undef);
	}

	return (\@seqnames, \@dists);

}		# End of the subroutine 'extract_phylip_dists' #


# mypower:
#	This subroutine calculates power, basically using
#	iterative multiplication. 
#
# INPUT: ($base, $exponent).
#
# OUTPUT: $power = $base ** $exponent.
#
sub mypower ($$) {

	my ($base, $exponent) = @_;

	($exponent ==0) and (return 1); ## DEFINE  0 * log 0 as 1 # 
	(($base ==0) and ($exponent >0)) and (return 0);
	if ($base ==0) {
	    print STDERR "[ERROR 'mypower'] ILLEGAL INPUT: (base, exponent) = \($base, $exponent\) !\n";
	    return undef;
	}

	my $power = 1;
	if ($exponent == int $exponent) {
	    my $abs_e = ($exponent>0) ? $exponent : - $exponent;
	    if ($abs_e <= $UL_ITMLT) {
		for (1 .. $abs_e) { $power *= $base; }
		($exponent<0) and ($power = 1/$power); 
	    } else {
		my $abs_b = ($base > 0) ? $base : - $base;
		$power = $abs_b ** $exponent;
		($exponent % 2) and ($base<0) and ($power = -$power);
	    }
	} elsif ($base <0) {
	    print STDERR "[ERROR 'mypower'] ILLEGAL INPUT: (base, exponent) = \($base, $exponent\) !\n";
	    return undef;
	} else {
	   $power = $base ** $exponent;
	}

	return $power;

}		# End of the subroutine 'mypower' #


# assign_branch_indices:
#	This subroutine assigns indices for the nodes immediately below the branches
#	from 0 upwards in a sloppy manner.
#
# INPUT: (\%node2parent, %null_br), where
#	%node2parent = ($node_id => $parent_id, ...);
#	%null_br =  (ID of the node immediately below the branch => some defined value, ...)
#	for the branches whose distances are definitely zero.
#
# OUTPUT: \%br2indx, where
#	%br2indx = (ID of the node immediately below the branch => $indx, ...)
#	for the branches whose distances may be nonzero.
#
sub assign_branch_indices (\%\%) {

	my ($node2parent, $null_br) = @_;

	my $indx = 0;
	my %br2indx;
	foreach my $nodeid (sort { $a <=> $b} keys %{$node2parent}) {
	    (defined $node2parent->{$nodeid}) or next; # skip the node NOT supporting a branch
	    (defined $null_br->{$nodeid}) and next; # skip the node whose supporting branch is known to be zero in length
	    $br2indx{$nodeid} = $indx;
	    $indx++;
	}

	return \%br2indx;

}		# end of the subroutine 'assign_branch_indices' #


# coeffs_leastsq_brlen:
#	This satellite subroutine of 'leastsq_brlen'
#	constructs the coefficient matrix and the right-hand-side vector
#	that are necessary for the least-square estimation of the branch lengths.
#
# INPUT: (\@dists, \@nodeids, \%br2indx, \%node2parent, $power, $buffer), where
#	$dists[$i]->[$j] is the distance between the $i th and $j th sequences
#	(with 0 <= $j < $i <= #{sequences});
#	$nodeids[$i] is the node ID of the $i th sequence;
#	%br2indx = (ID of the node immediately below the branch => $indx, ...);
#	%node2parent = ($node_id => $parent_node_id, ...);
#	$power and $buffer corresponds, respectively, to 'p' and 'buffer'
#	in the description of the subroutine 'leastsq_brlen'.
#
# OUTPUT: (\@coeffs, \@rhs), where
#	$coeffs[$m]->[$n] is the coefficient acting on the $n th branch length
#		to form the equation for the $m th branch length;
#	$rhs[$m] is the right hand side of the equation for the $m th branch length.
#
sub coeffs_leastsq_brlen (\@\@\%\%$$) {

	my ($dists, $nodeids, $br2indx, $node2parent, $power, $buffer) = @_;


	if ($power <0) {
	    print STDERR "[ERROR 'coeffs_leastsq_brlen'] ILLEGAL INPUT: NEGATIVE POWER currently NOT ALLOWED!\n";
	    return (undef, undef);
	}

=for TESTING the subroutine
	print STDERR "\n\< Subroutine 'coeffs_leastsq_brlen' \>\n\n";
=cut

	my $ct_br = scalar keys %{$br2indx};
	my $ct_nodes = @{$nodeids}; 

		# Initialize the output arrays #
	my @coeffs = my @rhs = ();
	for (1 .. $ct_br) { # The double for-loop #
	   push @rhs, 0;
	   my @row = ();
	   for (1 .. $ct_br) { push @row, 0; }
	   push @coeffs, \@row;
	}		# End of the double for-loop #

		# Now construct the coefficients #
	my $regl = ($power == 0) ? 1 : $buffer ** $power;  # A 'regulator'

	for (my $i=1; $i<$ct_nodes; $i++) { # The outermost for-loop #
	    my $node1 = $nodeids->[$i];
	    my $row_dists = $dists->[$i];
	    for (my $j=0; $j<$i; $j++) { # The outer for-loop #
		my $node2 = $nodeids->[$j];
		my $dist = $row_dists->[$j]; # Distance
			# Reciprocal denominator
		my $pwdist = mypower ($dist, $power);
		unless (defined $pwdist) {
		    print STDERR "[ERROR 'coeffs_leastsq_brlen'] 'mypower' FAILED for the node pair \($node1, $node2\) !\n";
		    return (undef, undef);
		}
		my $rec_den = ($power == 0) ? 1 : 1/($regl + $pwdist); # Reciprocal denominator #

		my ($cm_anc, $lin1, $lin2) = fetch_common_ancestors ($node1, $node2, %{$node2parent});
		unless (defined $cm_anc) {
		    print STDERR "[ERROR 'coeffs_leastsq_brlen'] 'fetch_common_ancestors' FAILED for the node pair \($node1, $node2\) !\n";
		    return (undef, undef);
		}
=for TESTING the subroutine
		print STDERR "\n{ Node pair \($node1, $node2\) }\n\n";
		print STDERR "distance = $dist, powered_distance = $pwdist,\n";
		print STDERR "    reciprocal_denominator = $rec_den .\n\n";
		print STDERR "Branches on the path: ", join (', ', @{$lin1}, @{$lin2}), ".\n\n";
=cut
		my @path = ();
		foreach my $brid (@{$lin1}, @{$lin2}) { # The 1st inner for-loop #
		    my $br_indx = $br2indx->{$brid};
		    (defined $br_indx) or next; # If the index is NOT defined,
				# the branch should have a fixed (zero) length.
		    push @path, $br_indx;
		}				# End of the 1st inner for-loop #

		my $ct_on_path = @path;
		for (my $m=0; $m<$ct_on_path; $m++) { # The 2nd inner for-loop #
		    my $br1 = $path[$m];
		    $rhs[$br1] += $dist * $rec_den;
		    my $row = $coeffs[$br1];
		    for (my $n=0; $n<$ct_on_path; $n++) { $row->[$path[$n]] += $rec_den; }
		} 				# End of the 2nd inner for-loop #
	    }				# End of the outer for-loop #
	}			# End of the outermost for-loop #

	return (\@coeffs, \@rhs);

}		# end of the subroutine 'coeffs_leastsq_brlen' #


# leastsq_brlen:
#	This subroutine estimates branch lengths by means of
#	a least square method like that used for the program 'fitch'
#	in the Phylip package.
#	The subroutine minimizes the sum of squares:
#	Sum_{(i,j)} [Exp(i,j) - Obs(i,j)]**2 /[(Obs(i,j))**p + (buffer)**p],
#	where Obs(i,j) is the observed value of the distance between the sequences i and j;
#	Exp(i,j) = Sum_{br(m) in path(i,j)} l(m)  is the expected vale of the distance
#	under a given set {l(m)} of the branch lengths;
#	p is the 'power' parameter (p=2 for Fitch & Margoliash's algorithm,
#	p=0 for the Neighbor-Joining method);
#	buffer is an 'effective minimum scale', which avoids divergence
#	even when the observed distance is extremely small.
#
# INPUT: (\@seqnames, \@dists, \%name2id, \%node2parent, $power, $buffer), where
#	$seqnames[$i] is the name (label) of the $i th sequence;
#	$dists[$i]->[$j] is the distance between the $i th and $j th sequences
#	(with 0 <= $j < $i <= #{sequences});
#	%name2id = ($seq_name => $node_id, ...), with $seq_name has to belong to @seqnames;
#	%node2parent = ($node_id => $parent_node_id, ...);
#	$power and $buffer corresponds, respectively, to 'p' and 'buffer'
#	in the above description of the least .
#
# OUTPUT: \%br_attr, where
#	%br_attr = (ID of the node immediately below the branch => [branch length], ...).
#
sub leastsq_brlen (\@\@\%\%$$) {

	my ($seqnames, $dists, $name2id, $node2parent, $power, $buffer) = @_;

	if (@{$seqnames}<3) {
	    my $ct_seqs = @{$seqnames};
	    print STDERR "[ERROR 'leastsq_brlen'] The number of sequences ($ct_seqs) is NOT ENOUGH for this subroutine to work !\n";
	    return undef;
	}

		# Create an array, while checking the consistency
		# between the distance matrix and the tree.
	my @nodeids = (); # = ($node_id, ...)
	foreach my $seqname (@{$seqnames}) {
	   my $nodeid = $name2id->{$seqname};
	   unless (defined $nodeid) {
		print STDERR "[ERROR 'leastsq_brlen'] The sequence '$seqname' has NO NODE ID!\n";
		return undef;
	   }
	   push @nodeids, $nodeid
	}

	my %br_attr; # (ID of the node immediately below the branch => \@attr, ...)

	# %{$br2indx} = (ID of the node immediately below the branch => $indx, ...)
	#
	my $br2indx = assign_branch_indices (%{$node2parent}, %br_attr);

		## BUG FIXING on Monday, March 5, 2007 ##
		## In the case where all the branches are trivially zero ##
	(0 < scalar keys %{$br2indx}) or (return \%br_attr);

	my ($coeffs, $rhs) = coeffs_leastsq_brlen (@{$dists}, @nodeids,
		%{$br2indx}, %{$node2parent}, $power, $buffer);

	unless (defined $coeffs) {
	    print STDERR "[ERROR 'leastsq_brlen'] 'coeffs_leastsq_brlen' FAILED !\n";
	    return undef;
	}

	# @{$br_lens} = (branch lengths arranged in the order of the assigned index.
	my $br_lens = solve_lin_eq_lu (@{$coeffs}, @{$rhs});
	unless (defined $br_lens) {
	    print STDERR "[ERROR 'leastsq_brlen'] 'solve_lin_eq_lu' FAILED !\n";
	    return undef;
	}

		# Cast the estimated branch lengths into the form
		# commonly used in my modules.
	foreach my $dwnode (keys %{$br2indx}) { # The last foreach-loop #
	    my $indx = $br2indx->{$dwnode};
	    (defined $indx) or next;
	    my $br_len = $br_lens->[$indx];
	    unless (defined $br_len) {
		print STDERR "[ERROR 'leastsq_brlen'] The branch immediately above the node '$dwnode' has NO BRANCH LENGTH estimated !\n"; 
		return undef;
	    }
	    my @tmp = ();
	    $tmp[$INDX_LENGTH] = $br_len;
	    $br_attr{$dwnode} = \@tmp;
	}			# End of the last foreach-loop #

	return \%br_attr;

}		# end of the subroutine 'leastsq_brlen' #


# assign_trivial_brlen:
#	This subroutine skims through the distance matrix and assigns  
#	ZERO LENGTHS to the branches comprizing the path between
#	the sequence pairs with NON-POSITIVE distances.
#
# INPUT: (\@dists, \@nodeids, \%node2parent, \%br_attr), where
#	$dists[$i]->[$j] is the distance between the $i th and $j th sequences
#	(with 0 <= $j < $i <= #{sequences});
#	$nodeids[$i] is the node ID of the $i th sequence;
#	%node2parent = ($node_id => $parent_node_id, ...);
#	%br_attr = (ID of the node immediately below the branch => \@attr, ...)
#	 with $attr[$INDX_LENGTH] is the branch length,
#	this hash is ASSUMED to be empty when input.
#
# OUTPUT: 1 if the subroutine succeeded,
#	  0 if it failed.
#
sub assign_trivial_brlen (\@\@\%\%) {

	my ($dists, $nodeids, $node2parent, $br_attr) = @_; 

	my $ct_seqs = @{$nodeids};

=for TESTING the subroutine
	print STDERR "\n\< Subroutine 'assign_trivial_brlen' \>\n\n";
=cut

	for (my $i=1; $i<$ct_seqs; $i++) { # The outer for-loop #
	    my $node1 = $nodeids->[$i];
	    my $row_dists = $dists->[$i];
	    for (my $j=0; $j < $i; $j++) { # The middle for-loop #
		my $node2 = $nodeids->[$j];
		my $dist = $row_dists->[$j];
		($dist >0) and next;

			# The pair has a non-positive distance #
		my ($cm_anc, $lin1, $lin2) = fetch_common_ancestors ($node1, $node2, %{$node2parent});
		unless (defined $cm_anc) {
		    print STDERR "[ERROR 'assign_trivial_brlen'] 'fetch_common_ancestors' FAILED for the node pair \($node1, $node2\) !\n";
		    return (undef, undef);
		}

=for TESTING the subroutine
		print STDERR "\n{ Node pair \($node1, $node2\) }\n\n";
		print STDERR "distance = $dist, powered_distance = $pwdist,\n";
		print STDERR "    reciprocal_denominator = $rec_den .\n\n";
		print STDERR "Branches on the path: ", join (', ', @{$lin1}, @{$lin2}), ".\n\n";
=cut

		foreach my $br_id (@{$lin1}, @{$lin2}) { # The inner foreach-loop #
		   (defined $br_attr->{$br_id}) and next;
			# Don't touch already given attributes

			# Assign zero for the branch length #
		   my @attr = ();
		   $attr[$INDX_LENGTH] = 0;
		   $br_attr->{$br_id} = \@attr;
		}			# End of the inner foreach-loop #
		
	    }			# End of the middle for-loop #
	}		# End of the outer for-loop #

	return 1;

}		# End of the subroutine 'assign_trivial_brlen' #


# leastsq_brlen_nn:
#	This revised version of the subroutine 'leastsq_brlen'
#	basically do the same job as the predecessor,
#	except that this version KEEPS branch lengths NONNEGATIVE.
#	It forcefully assigns zero length to the branches
#	whose lengths are estimated to be negative,
#	then recalculate the remaining lengths.
#
# leastsq_brlen:
#	This subroutine estimates branch lengths by means of
#	a least square method like that used for the program 'fitch'
#	in the Phylip package.
#	The subroutine minimizes the sum of squares:
#	Sum_{(i,j)} [Exp(i,j) - Obs(i,j)]**2 /[(Obs(i,j))**p + (buffer)**p],
#	where Obs(i,j) is the observed value of the distance between the sequences i and j;
#	Exp(i,j) = Sum_{br(m) in path(i,j)} l(m)  is the expected vale of the distance
#	under a given set {l(m)} of the branch lengths;
#	p is the 'power' parameter (p=2 for Fitch & Margoliash's algorithm,
#	p=0 for the Neighbor-Joining method);
#	buffer is an 'effective minimum scale', which avoids divergence
#	even when the observed distance is extremely small.
#
# INPUT: (\@seqnames, \@dists, \%name2id, \%node2parent, $power, $buffer), where
#	$seqnames[$i] is the name (label) of the $i th sequence;
#	$dists[$i]->[$j] is the distance between the $i th and $j th sequences
#	(with 0 <= $j < $i <= #{sequences});
#	%name2id = ($seq_name => $node_id, ...), with $seq_name has to belong to @seqnames;
#	%node2parent = ($node_id => $parent_node_id, ...);
#	$power and $buffer corresponds, respectively, to 'p' and 'buffer'
#	in the above description of the least .
#
# OUTPUT: \%br_attr, where
#	%br_attr = (ID of the node immediately below the branch => [branch length], ...).
#
sub leastsq_brlen_nn (\@\@\%\%$$) {

	my ($seqnames, $dists, $name2id, $node2parent, $power, $buffer) = @_;

	if (@{$seqnames}<3) {
	    my $ct_seqs = @{$seqnames};
	    print STDERR "[ERROR 'leastsq_brlen_nn'] The number of sequences ($ct_seqs) is NOT ENOUGH for this subroutine to work !\n";
	    return undef;
	}

		# Create an array, while checking the consistency
		# between the distance matrix and the tree.
	my @nodeids = (); # = ($node_id, ...)
	foreach my $seqname (@{$seqnames}) {
	   my $nodeid = $name2id->{$seqname};
	   unless (defined $nodeid) {
		print STDERR "[ERROR 'leastsq_brlen_nn'] The sequence '$seqname' has NO NODE ID!\n";
		return undef;
	   }
	   push @nodeids, $nodeid
	}

	my %br_attr; # (ID of the node immediately below the branch => \@attr, ...)

		# First assign zero to the lengths of branches comprising
		# the paths with distance zero.
		#
	assign_trivial_brlen (@{$dists}, @nodeids, %{$node2parent}, %br_attr);

	my $round = 0;
    while (1) { # An infinite while-loop ! #
	$round++;

=for TESTING the subroutine
	print STDERR "\n[ROUND $round]\n\n";
	print STDERR "Branches with ZERO lengths: ", join (', ', sort keys %br_attr), " .\n\n";
=cut

	# %{$br2indx} = (ID of the node immediately below the branch => $indx, ...)
	#
	my $br2indx = assign_branch_indices (%{$node2parent}, %br_attr);

		## BUG FIXING on Monday, March 5, 2007 ##
		## In the case where all the branches are trivially zero ##
	(0 < scalar keys %{$br2indx}) or (return \%br_attr);

	my ($coeffs, $rhs) = coeffs_leastsq_brlen (@{$dists}, @nodeids,
		%{$br2indx}, %{$node2parent}, $power, $buffer);

	unless (defined $coeffs) {
	    print STDERR "[ERROR 'leastsq_brlen_nn'] 'coeffs_leastsq_brlen' FAILED  in the $round th round!\n";
	    return undef;
	}

	# @{$br_lens} = (branch lengths arranged in the order of the assigned index.
	my $br_lens = solve_lin_eq_lu (@{$coeffs}, @{$rhs});
	unless (defined $br_lens) {
	    print STDERR "[ERROR 'leastsq_brlen_nn'] 'solve_lin_eq_lu' FAILED in the $round th round!\n";
	    return undef;
	}

		# Cast the estimated branch lengths into the form
		# commonly used in my modules.

		# First, look for NEGATIVE branch lengths,
		# and, if any, set them to zero.
		#
	my $ct_nega = 0;
	foreach my $dwnode (keys %{$br2indx}) { # The last foreach-loop #
	    my $indx = $br2indx->{$dwnode};
	    (defined $indx) or next;
	    my $br_len = $br_lens->[$indx];
	    unless (defined $br_len) {
		print STDERR "[ERROR 'leastsq_brlen_nn'] The branch immediately above the node '$dwnode' has NO BRANCH LENGTH estimated (in the $round th round) !\n"; 
		return undef;
	    }
	    ($br_len <0) or next;

	    $ct_nega++;
	    my @tmp = ();
	    $tmp[$INDX_LENGTH] = 0;
	    $br_attr{$dwnode} = \@tmp;
	}			# End of the last foreach-loop #

	($ct_nega>0) and next; # If there is any negative branch lengths,
				# recalculate the remaining ones.

		# Second, stuff the remaining branch lengths #
		#
	foreach my $dwnode (keys %{$br2indx}) { # The last foreach-loop #
	    (defined $br_attr{$dwnode}) and next;
	    my $indx = $br2indx->{$dwnode};
	    (defined $indx) or next;

	    my $br_len = $br_lens->[$indx];
	    my @tmp = ();
	    $tmp[$INDX_LENGTH] = $br_len;
	    $br_attr{$dwnode} = \@tmp;
	}			# End of the last foreach-loop #

	return \%br_attr;

    }			# End of the infinite while-loop # 

}		# end of the subroutine 'leastsq_brlen_nn' #


# extended_upgma_tree:
#	This subroutine implements an extended version of the
#	UPGMA (unweighted pair-group method using arithmetic averages) algorithm
#	for tree construction.
#	This extended algorithm allows a distance matrix to have undefined elements.
#
#	ALGORITHM:
#	(i) Compare ALL DEFINED pairwise distances, and
#	cluster the pair with the smallest distance,
#	creating a new node at the mid-point between the two taxa.
#	(NOTE that the distance will be defined between the new clusters,
#	not between the new nodes);
#
#	(ii) Update the distance matrix by replacing the clustered pair of taxa
#	with a new taxon defined by the clustered pair.
#	The distance between two clusters (A and B) is given by:
#	d (A,B) = Sum_(i,j) d (i, j) / #{pairs with defined distances},
#	where the summation is taken across the pairs (i,j) with i in A, j in B that have defined distances.
#	In the actual calculation, the update of the distance is performed as:
#	d (A, B1+B2) = [#{dd-pairs}_(A, B1) * d (A, B1) + #{dd-pairs}_(A, B2) * d (A, B2)] / #{dd-pairs}_(A,B) ,
#	where the cluster B is formed by merging the 2 clusters B1 and B2, and #{dd-pairs}_(A, B) is the number
#	of exterior taxon pairs with defined distances. NOTE the equation
#	#{dd_pairs}_(A, B} = #{dd_pairs}_(A, B1) + #{dd_pairs}_(A, B2) ;
#
#	(iii) Iterate (i) and (ii) untill we are left with a set of taxa between which NO DISTANCES are defined.
#
# INPUT: (\@taxa, \@pw_distances, $str_na), where
#	$taxa[$i] is the ID or name of the $i+1 th taxon ($i=0, ..., #{taxa} -1);
#	$pw_distance[$i]->[$j] (0<=$j< $i) is the pairwise distance between $taxa[$i] and $taxa[$j].
#	The element $pw_distance[0] is the reference to an empty array;
#	$str_na is the string indicating that the distance is unavailable
#	(or undefined) between the taxa.
#
# OUTPUT: (\@set_name2nodeid, \%nodeid2name, \%node2pa, \%node2chi, \%br_attr, \%node2ct_dd_pairs),where
#	%{$set_name2nodeid[$t]} = ($taxon => $node_id, ...) for $TOP_NODE and exterior taxa (given by @taxa)
#	 of the $t th tree,
#	%nodeid2name = ($node_ID => $taxon, ...) for exterior taxa,
#	%node2pa = ($node_ID => $parent_ID, ...) ,
#	%node2chi = ($node_ID => \@child_IDs, ...) ,
#	%br_attr = ($ID of the nearest downstream node => [$length, $bootstrap (=undef)], ...),
#	%node2ct_dd_pairs = ($node_id => #{pairs with defined distances} between the 2 sister clusters, ...)
#	for interior nodes.
#
#	CONVENTION for the NODE IDs:
#	The node ID $i+1 is assigned to the exterior taxon $taxa[$i] ($i = 0, ... #{taxa}-1),
#	then each interior taxon (or node) is assigned with a node ID that begins with $ct_taxa,
#	and that is incremented by 1 as the a new taxon (=cluster) is created.
#
sub extended_upgma_tree (\@\@$) {

	my ($taxa, $pw_distances, $str_na) = @_;

	my $ct_taxa = @{$taxa};
	unless ($ct_taxa>0) {
	    print STDERR "[ERROR 'extended_upgma_tree'] ILLEGAL INPUT: NO TAXA are supplied: $! !\n";
	    return (undef, undef, undef, undef, undef, undef);
	}

	unless ($ct_taxa == @{$pw_distances}) {
	    print STDERR "\n[ERROR 'extended_upgma_tree'] ILLEGAL INPUT: The number of ROWS (", scalar (@{$pw_distances}), ") CLASHES with the number of TAXA ($ct_taxa) : $! !\n";
	    return (undef, undef, undef, undef, undef, undef);
	}

		# Initialize the setting #
	my @mtrx_ct_dd_pairs = ();
	my %nodeid2name;
	for (my $i=0; $i<$ct_taxa; $i++) { # The preliminary outer for-loop #
	    $nodeid2name{$i+1} = $taxa->[$i];
	    my @row_ct_dd_pairs = ();
	    my $row_pw_distances = $pw_distances->[$i];
	    for (my $j = 0; $j<$i; $j++) { # Inner for-loop #
		my $pw_dist = $row_pw_distances->[$j];
		push @row_ct_dd_pairs, ((defined $pw_dist) and ($pw_dist ne $str_na)) ?  1 : 0;
	    }				# End of the inner for-loop #
	    push @mtrx_ct_dd_pairs, \@row_ct_dd_pairs;
	}			# End of the preliminary outer for-loop #
	my @sbjct_nodes = (1..$ct_taxa); # list of subject nodes in exactly the same order as in the distance matrix
	my $max_nodeid = $ct_taxa;

	my @pw_distances = @{$pw_distances};

	my %node2pa; my %node2chi; my %br_attr; my %node2ct_dd_pairs;

	my %node2level; # ($node_ID => 'level' from the bottom, ...), which is an auxiliary hash

	while (1) { # The main outer while-loop #

	    my $ct_sbjcts = @sbjct_nodes;

		# Search for the closest pair #

	    my @closest_pair = (); # = ($indx_node1, $indx_node2, $distance) where $indx_node1/2 is the index in the distance matrix.
	    for (my $i=0; $i<$ct_sbjcts; $i++) { # The 1st middle for-loop #
		my $row_pw_distances = $pw_distances[$i];
		for (my $j=0; $j<$i; $j++) { # inner for-loop #
		    my $pw_dist = $row_pw_distances->[$j];

			# Skip pairs with undefined or unavailable distances #
		    (defined $pw_dist) or next;
		    ($pw_dist eq $str_na) and next;

		    unless (@closest_pair>0) { # 'Initialize' the closest pair
			@closest_pair = ($i, $j, $pw_dist);
			next;
		    }
		    ($pw_dist < $closest_pair[2]) or next; # No need to update

			# Update the closest pair
		     @closest_pair = ($i, $j, $pw_dist);
		
		}			# End of the inner for-loop #
	    }				# End of the 1st middle for-loop #

	    (@closest_pair>0) or last; # We cannot cluster the nodes any more. #

		# Cluster the closest pair #

	    my ($i, $j, $pw_dist) = @closest_pair;
	    $max_nodeid++;

	    $node2ct_dd_pairs{$max_nodeid} = $mtrx_ct_dd_pairs[$i]->[$j];

	    ##my @clustered = @sbjct_nodes[$i,$j];
	    my @clustered = @sbjct_nodes[$j,$i]; # Just for a better visualization

	    $node2chi{$max_nodeid} = \@clustered;
	    foreach my $nodeid (@clustered) { # The 2nd middle foreach-loop #
		$node2pa{$nodeid} = $max_nodeid;
		my $br_len = my $level = $pw_dist/2;
		$node2level{$nodeid} = $level;
		my $children = $node2chi{$nodeid};
		((defined $children) and (@{$children}>0))
		    and ($br_len -= $node2level{$children->[0]});
		my @attr = ();
		$attr[$INDX_LENGTH] = $br_len;
		$br_attr{$nodeid} = \@attr;
	    }			# End of the 2nd middle foreach-loop #

	    unless ($ct_sbjcts > 2) { # We do not have the subjects to cluster any longer.
		@sbjct_nodes = ($max_nodeid);
		last;
	    }

		# Update the subject nodes &
		# matrices for distances & #{distance-defined pairs} #

	    my @new_sbjct_nodes = my @new_pw_distances = my @new_mtrx_ct_dd_pairs = ();

	    my @mgd_row_pw_distances = my @mgd_row_ct_dd_pairs = ();
		# merged information on the $i th and $j th subjects #

	    for (my $k=0; $k< $ct_sbjcts; $k++) { # The 3rd middle for-loop #
		(($k == $i) or ($k == $j)) and next; # skip the clustered subjects #

			# squeeze pairs of remaining taxa into the new arrays #

		push @new_sbjct_nodes, $sbjct_nodes[$k];

		my $row_pw_distances = $pw_distances[$k];
		my $row_ct_dd_pairs = $mtrx_ct_dd_pairs[$k];
		my @new_row_pw_distances = my @new_row_ct_dd_pairs = ();
		for (my $l=0; $l<$k; $l++) { # The inner for-loop #
		    (($l == $i) or ($l == $j)) and next; # Skip the clustered subjects

		    push @new_row_pw_distances, $row_pw_distances->[$l];
		    push @new_row_ct_dd_pairs, $row_ct_dd_pairs->[$l];
		}			# End of the inner for-loop #

		push @new_pw_distances, \@new_row_pw_distances;
		push @new_mtrx_ct_dd_pairs, \@new_row_ct_dd_pairs;

			# pairs containing a new (merged) taxon #

		my $dist1 = ($k>$i) ? $pw_distances[$k]->[$i] : $pw_distances[$i]->[$k];
		my $ct1 = ($k > $i) ? $mtrx_ct_dd_pairs[$k]->[$i] : $mtrx_ct_dd_pairs[$i]->[$k];
		my $dist2 = ($k>$j) ? $pw_distances[$k]->[$j] : $pw_distances[$j]->[$k];
		my $ct2 = ($k>$j) ? $mtrx_ct_dd_pairs[$k]->[$j] : $mtrx_ct_dd_pairs[$j]->[$k];
		my $sum_ct = $ct1 + $ct2;
		push @mgd_row_ct_dd_pairs, $sum_ct;
		if ($sum_ct>0) {
		   my $sum_dist = ($ct1>0) ? $ct1 * $dist1 : 0;
		   ($ct2>0) and ($sum_dist += $ct2 * $dist2);
		   push @mgd_row_pw_distances, $sum_dist/$sum_ct;
		} else {
		    push @mgd_row_pw_distances, undef;
		}

	    }			# End of the 3rd middle for-loop #

		# Finish updating the arrays #

	    push @new_sbjct_nodes, $max_nodeid;
	    push @new_pw_distances, \@mgd_row_pw_distances;
	    push @new_mtrx_ct_dd_pairs, \@mgd_row_ct_dd_pairs;

	    @sbjct_nodes = @new_sbjct_nodes;
	    @pw_distances = @new_pw_distances;
	    @mtrx_ct_dd_pairs = @new_mtrx_ct_dd_pairs;

	}	# End of the main outer while-loop #

		#  Create a hash %name2nodeid for each of the remaining subject nodes,
		# which will be the top-node of a tree made of nodes connected to it.

	my @set_name2nodeid = ();
	foreach my $topnode_id (@sbjct_nodes) { # The trailing foreach-loop #
	    my %name2nodeid = ($TOP_NODE => $topnode_id); 
	    my @curr_nodes = ($topnode_id);
	    while (@curr_nodes>0) { # The middle while-loop #
	        my @child_nodes = ();
		foreach my $nodeid (@curr_nodes) { # inner foreach-loop #

		    my $name = $nodeid2name{$nodeid};
		    (defined $name) and ($name2nodeid{$name} = $nodeid);

		    my $children = $node2chi{$nodeid};
		    (defined $children) and (push @child_nodes, @{$children});
		}			# End of the inner foreach-loop #

		@curr_nodes = @child_nodes;
	    }			# End of the middle while-loop #

	    push @set_name2nodeid, \%name2nodeid;
	}			# End of the trailing foreach-loop #


	return (\@set_name2nodeid, \%nodeid2name, \%node2pa, \%node2chi, \%br_attr, \%node2ct_dd_pairs);

}		# End of the subroutine 'extended_upgma_tree' #


# extended_wtd_upgma_tree:
#	This revised version of 'extended_upgma_tree' calculates WEIGHTED averages
#	of distances between cluster pairs. The weight for each pair is given by
#	the number of effective sites used for distance estimation, for example.
#	Do NOT CONFUSE this algorithm with the WPGMA (weighted pair-group method using arithmetic averages) algorithm. In WPGMA, distances are weighted so that the members
#	of the smaller cluster give larger contributions to the average, which is
#	considered inferior to UPGMA, in which all members give equal contributions.
#	I think the algorithm used here is MORE REASONABLE than the simple UPGMA WHEN 
#	HANDLING pairwise distances estimated from DIFFERENT NUMBERS OF EFFECTIVE SITES.
#	Although the use of such distance matrix is considered unconventional.
#
#	In prectice, the only revision from the predecessor 'extended_upgma_tree' is
#	in the step (ii), where #{dd-pairs}_(A,B) is replaced by
#	{total_weight}_(A,B} = Sum_{$i in A} Sum_{$j in B} {weight}_($i, $j),
#	with {weight}_($i, $j) being a weight factor between the taxa $i and $j,
#	which was received as an input. Like #{dd_pairs}, {weight} also follow the
#	addition law: {weight}_(A,B1 + B2) = {weight}_(A,b1) + {weight}_(A,B2).
#	We force {weight}_($i,$j) = 0 for the pairs ($i, $j) whose distances
#	are UNavailable or UNdefined.
#	
# extended_upgma_tree:
#	This subroutine implements an extended version of the
#	UPGMA (unweighted pair-group method using arithmetic averages) algorithm
#	for tree construction.
#	This extended algorithm allows a distance matrix to have undefined elements.
#
#	ALGORITHM:
#	(i) Compare ALL DEFINED pairwise distances, and
#	cluster the pair with the smallest distance,
#	creating a new node at the mid-point between the two taxa.
#	(NOTE that the distance will be defined between the new clusters,
#	not between the new nodes);
#
#	(ii) Update the distance matrix by replacing the clustered pair of taxa
#	with a new taxon defined by the clustered pair.
#	The distance between two clusters (A and B) is given by:
#	d (A,B) = Sum_(i,j) d (i, j) / #{pairs with defined distances},
#	where the summation is taken across the pairs (i,j) with i in A, j in B that have defined distances.
#	In the actual calculation, the update of the distance is performed as:
#	d (A, B1+B2) = [#{dd-pairs}_(A, B1) * d (A, B1) + #{dd-pairs}_(A, B2) * d (A, B2)] / #{dd-pairs}_(A,B) ,
#	where the cluster B is formed by merging the 2 clusters B1 and B2, and #{dd-pairs}_(A, B) is the number
#	of exterior taxon pairs with defined distances. NOTE the equation
#	#{dd_pairs}_(A, B} = #{dd_pairs}_(A, B1) + #{dd_pairs}_(A, B2) ;
#
#	(iii) Iterate (i) and (ii) untill we are left with a set of taxa between which NO DISTANCES are defined.
#
# INPUT: (\@taxa, \@pw_distances, \@pw_weights, $str_na), where
#	$taxa[$i] is the ID or name of the $i+1 th taxon ($i=0, ..., #{taxa} -1);
#	$pw_distance[$i]->[$j] (0<=$j< $i) is the pairwise distance between $taxa[$i] and $taxa[$j].
#	The element $pw_distance[0] is the reference to an empty array;
#	$pw_weights[$i]->[$j] (0<=$j<$i) is the weight factor assigned to
#	the pair ($taxa[$i], $taxa[$j]);
#	$str_na is the string indicating that the distance is UNavailable
#	(or UNdefined) between the taxa.
#
# OUTPUT: (\@set_name2nodeid, \%nodeid2name, \%node2pa, \%node2chi, \%br_attr, \%node2pw_weight),where
#	%{$set_name2nodeid[$t]} = ($taxon => $node_id, ...) for $TOP_NODE and exterior taxa (given by @taxa)
#	 of the $t th tree,
#	%nodeid2name = ($node_ID => $taxon, ...) for exterior taxa,
#	%node2pa = ($node_ID => $parent_ID, ...) ,
#	%node2chi = ($node_ID => \@child_IDs, ...) ,
#	%br_attr = ($ID of the nearest downstream node => [$length, $bootstrap (=undef)], ...),
#	%node2pw_weights = ($node_id => {total weights} between the 2 sister clusters, ...)
#	for interior nodes.
#
#	CONVENTION for the NODE IDs:
#	The node ID $i+1 is assigned to the exterior taxon $taxa[$i] ($i = 0, ... #{taxa}-1),
#	then each interior taxon (or node) is assigned with a node ID that begins with $ct_taxa,
#	and that is incremented by 1 as the a new taxon (=cluster) is created.
#
sub extended_wtd_upgma_tree (\@\@\@$) {

	my ($taxa, $pw_distances, $pw_weights, $str_na) = @_;

	my $ct_taxa = @{$taxa};
	unless ($ct_taxa>0) {
	    print STDERR "[ERROR 'extended_wtd_upgma_tree'] ILLEGAL INPUT: NO TAXA are supplied: $! !\n";
	    return (undef, undef, undef, undef, undef, undef);
	}

	unless ($ct_taxa == @{$pw_distances}) {
	    print STDERR "\n[ERROR 'extended_wtd_upgma_tree'] ILLEGAL INPUT: The number of ROWS (", scalar (@{$pw_distances}), ") for the distance matrix CLASHES with the number of TAXA ($ct_taxa) : $! !\n";
	    return (undef, undef, undef, undef, undef, undef);
	}
	unless ($ct_taxa == @{$pw_weights}) {
	    print STDERR "\n[ERROR 'extended_wtd_upgma_tree'] ILLEGAL INPUT: The number of ROWS (", scalar (@{$pw_weights}), ") for the weight matrix CLASHES with the number of TAXA ($ct_taxa) : $! !\n";
	    return (undef, undef, undef, undef, undef, undef);
	}

		# Initialize the setting #
	my @mtrx_pw_weights = ();
	my %nodeid2name;
	for (my $i=0; $i<$ct_taxa; $i++) { # The preliminary outer for-loop #
	    $nodeid2name{$i+1} = $taxa->[$i];
	    my @row_pw_weights = ();
	    my $row_pw_distances = $pw_distances->[$i];
	    my $row_pw_weights = $pw_weights->[$i];
	    for (my $j = 0; $j<$i; $j++) { # Inner for-loop #
		my $pw_dist = $row_pw_distances->[$j];
		my $pw_wgt = $row_pw_weights->[$j];

		if ((defined $pw_dist) and ($pw_dist ne $str_na)) {
		    unless (defined $pw_wgt) {
			print STDERR "\n[ERROR 'extended_wtd_upgma_tree'] ILLEGAL INPUT: The weight for the pair \($i, $j\) = \($taxa->[$i], $taxa->[$j]\) is NOT AVAILABLE !\n";
			return (undef, undef, undef, undef, undef, undef);
		    }
		    push @row_pw_weights,  $pw_wgt;
		} else {
		    if ((defined $pw_wgt) and ($pw_wgt ne $str_na) and ($pw_wgt != 0)) {
			print STDERR "\n[CAUTION 'extended_wtd_upgma_tree'] The pair \($i, $j\) = \($taxa->[$i], $taxa->[$j]\) has a DEFINED WEIGHT ($pw_wgt) !\n";
			print STDERR "  --- I will FORCEFULLY RE-ASSIGN the weight '0' to the pair ...\n\n";
		    }
		    push @row_pw_weights, 0;
		}
	    }				# End of the inner for-loop #
	    push @mtrx_pw_weights, \@row_pw_weights;
	}			# End of the preliminary outer for-loop #
	my @sbjct_nodes = (1..$ct_taxa); # list of subject nodes in exactly the same order as in the distance matrix
	my $max_nodeid = $ct_taxa;

	my @pw_distances = @{$pw_distances};

	my %node2pa; my %node2chi; my %br_attr; my %node2pw_weight;

	my %node2level; # ($node_ID => 'level' from the bottom, ...), which is an auxiliary hash

	while (1) { # The main outer while-loop #

	    my $ct_sbjcts = @sbjct_nodes;

		# Search for the closest pair #

	    my @closest_pair = (); # = ($indx_node1, $indx_node2, $distance) where $indx_node1/2 is the index in the distance matrix.
	    for (my $i=0; $i<$ct_sbjcts; $i++) { # The 1st middle for-loop #
		my $row_pw_distances = $pw_distances[$i];
		for (my $j=0; $j<$i; $j++) { # inner for-loop #
		    my $pw_dist = $row_pw_distances->[$j];

			# Skip pairs with undefined or unavailable distances #
		    (defined $pw_dist) or next;
		    ($pw_dist eq $str_na) and next;

		    unless (@closest_pair>0) { # 'Initialize' the closest pair
			@closest_pair = ($i, $j, $pw_dist);
			next;
		    }
		    ($pw_dist < $closest_pair[2]) or next; # No need to update

			# Update the closest pair
		     @closest_pair = ($i, $j, $pw_dist);
		
		}			# End of the inner for-loop #
	    }				# End of the 1st middle for-loop #

	    (@closest_pair>0) or last; # We cannot cluster the nodes any more. #

		# Cluster the closest pair #

	    my ($i, $j, $pw_dist) = @closest_pair;
	    $max_nodeid++;

	    $node2pw_weight{$max_nodeid} = $mtrx_pw_weights[$i]->[$j];

	    ##my @clustered = @sbjct_nodes[$i,$j];
	    my @clustered = @sbjct_nodes[$j,$i]; # Just for a better visualization

	    $node2chi{$max_nodeid} = \@clustered;
	    foreach my $nodeid (@clustered) { # The 2nd middle foreach-loop #
		$node2pa{$nodeid} = $max_nodeid;
		my $br_len = my $level = $pw_dist/2;
		$node2level{$nodeid} = $level;
		my $children = $node2chi{$nodeid};
		((defined $children) and (@{$children}>0))
		    and ($br_len -= $node2level{$children->[0]});
		my @attr = ();
		$attr[$INDX_LENGTH] = $br_len;
		$br_attr{$nodeid} = \@attr;
	    }			# End of the 2nd middle foreach-loop #

	    unless ($ct_sbjcts > 2) { # We do not have the subjects to cluster any longer.
		@sbjct_nodes = ($max_nodeid);
		last;
	    }

		# Update the subject nodes &
		# matrices for distances & #{distance-defined pairs} #

	    my @new_sbjct_nodes = my @new_pw_distances = my @new_mtrx_pw_weights = ();

	    my @mgd_row_pw_distances = my @mgd_row_pw_weights = ();
		# merged information on the $i th and $j th subjects #

	    for (my $k=0; $k< $ct_sbjcts; $k++) { # The 3rd middle for-loop #
		(($k == $i) or ($k == $j)) and next; # skip the clustered subjects #

			# squeeze pairs of remaining taxa into the new arrays #

		push @new_sbjct_nodes, $sbjct_nodes[$k];

		my $row_pw_distances = $pw_distances[$k];
		my $row_pw_weights = $mtrx_pw_weights[$k];
		my @new_row_pw_distances = my @new_row_pw_weights = ();
		for (my $l=0; $l<$k; $l++) { # The inner for-loop #
		    (($l == $i) or ($l == $j)) and next; # Skip the clustered subjects

		    push @new_row_pw_distances, $row_pw_distances->[$l];
		    push @new_row_pw_weights, $row_pw_weights->[$l];
		}			# End of the inner for-loop #

		push @new_pw_distances, \@new_row_pw_distances;
		push @new_mtrx_pw_weights, \@new_row_pw_weights;

			# pairs containing a new (merged) taxon #

		my $dist1 = ($k>$i) ? $pw_distances[$k]->[$i] : $pw_distances[$i]->[$k];
		my $wgt1 = ($k > $i) ? $mtrx_pw_weights[$k]->[$i] : $mtrx_pw_weights[$i]->[$k];
		my $dist2 = ($k>$j) ? $pw_distances[$k]->[$j] : $pw_distances[$j]->[$k];
		my $wgt2 = ($k>$j) ? $mtrx_pw_weights[$k]->[$j] : $mtrx_pw_weights[$j]->[$k];
		my $sum_wgt = $wgt1 + $wgt2;
		push @mgd_row_pw_weights, $sum_wgt;
		if ($sum_wgt>0) {
		   my $sum_dist = ($wgt1>0) ? $wgt1 * $dist1 : 0;
		   ($wgt2>0) and ($sum_dist += $wgt2 * $dist2);
		   push @mgd_row_pw_distances, $sum_dist/$sum_wgt;
		} else {
		    push @mgd_row_pw_distances, undef;
		}

	    }			# End of the 3rd middle for-loop #

		# Finish updating the arrays #

	    push @new_sbjct_nodes, $max_nodeid;
	    push @new_pw_distances, \@mgd_row_pw_distances;
	    push @new_mtrx_pw_weights, \@mgd_row_pw_weights;

	    @sbjct_nodes = @new_sbjct_nodes;
	    @pw_distances = @new_pw_distances;
	    @mtrx_pw_weights = @new_mtrx_pw_weights;

	}	# End of the main outer while-loop #

		#  Create a hash %name2nodeid for each of the remaining subject nodes,
		# which will be the top-node of a tree made of nodes connected to it.

	my @set_name2nodeid = ();
	foreach my $topnode_id (@sbjct_nodes) { # The trailing foreach-loop #
	    my %name2nodeid = ($TOP_NODE => $topnode_id); 
	    my @curr_nodes = ($topnode_id);
	    while (@curr_nodes>0) { # The middle while-loop #
	        my @child_nodes = ();
		foreach my $nodeid (@curr_nodes) { # inner foreach-loop #

		    my $name = $nodeid2name{$nodeid};
		    (defined $name) and ($name2nodeid{$name} = $nodeid);

		    my $children = $node2chi{$nodeid};
		    (defined $children) and (push @child_nodes, @{$children});
		}			# End of the inner foreach-loop #

		@curr_nodes = @child_nodes;
	    }			# End of the middle while-loop #

	    push @set_name2nodeid, \%name2nodeid;
	}			# End of the trailing foreach-loop #


	return (\@set_name2nodeid, \%nodeid2name, \%node2pa, \%node2chi, \%br_attr, \%node2pw_weight);

}		# End of the subroutine 'extended_wtd_upgma_tree' #


# level_sglnode_upgma:
#	This subroutine quickly estimates the level of a single node
#	in a UPGMA tree.
#	The subroutine ASSUME that the distances are ultra-metric.
#
# INPUT: ($node_id, \%node2chi, \%br_attr), where
#
# OUTPUT: $level = {summation of branch lengths along the path
#	from the input node to an exterior node under it}.
#
sub level_sglnode_upgma ($\%\%) {

	my ($node_id, $node2chi, $br_attr) = @_;

	my $level = 0;
	my $children = $node2chi->{$node_id};

	while ((defined $children) and (@{$children}>0)) { # The main while-loop #
	    my $child = $children->[0];
	    my $attr = $br_attr->{$child};
	    unless (defined $attr) {
		print STDERR "[ERROR 'level_sglnode_upgma'] ILLEGAL INPUT: NO ATTRIBUTES are assigned to the branch just above the node '$child': $! !\n";
		return undef;
	    }
	    my $br_len = $attr->[$INDX_LENGTH];
	    unless (defined $br_len) {
		print STDERR "[ERROR 'level_sglnode_upgma'] ILLEGAL INPUT: NO LENGTH is assigned to the branch just above the node '$child': $! !\n";
		return undef;
	    }

	    $level += $br_len;
	    $children = $node2chi->{$child};
	}			# End of the main while-loop #

	return $level;

}			# End of the subroutine 'level_sglnode_upgma' #


# var_nodelevels_ext_wgt_upgma:
#	This satellite subroutine of 'approx_var_brlen_ext_wgt_upgma'
#	extimates variances of node levels ( = the summation of branch lengths
#	in a path from the node in question to an exterior node under it)
#	under the weighted extended UPGMA model.
#	You can find details on the assumptions and algorithms
#	in the header part of 'approx_var_brlen_ext_wgt_upgma'.
#
# INPUT: (\%nodeid2mtrx_indx, \@mtrx_dist, \@mtrx_wgt, \@mtrx_se, $str_na,
#		$topnode_id, \%node2chi, \%node2pa, \%br_attr), where
#	%nodeid2mtrx_indx = ($node_ID => $indx for the matrices, ...);
#	$mtrx_dist/wgt/se[$k]->[$l] (0<= $l<$k<$ct_seqs) gives the distance/weight factor/standard error of the distance for the pair of sequences $k and $l;
#	$str_na indicates that the matrix element is not available;
#	%node2chi = ($node_ID => \@child_nodes, ...);
#	%node2pa  = ($node_ID => $parent_ID, ...);
#	%br_attr = ($ID of the node immediately below the branh => \@attributes, ...),
#	with $attribute[$INDX_LENGTH] giving branch length ;
#
# OUTPUT: (\%node2var_level, \%node2depth, \%node2level), where
#	%node2var_level = ($node_ID => variance of the level of the node, ...);
#	%node2depth = ($node_ID => $depth, ...),
#	with $depth defined as the number of 'ancestor' nodes
#	between it (excluding) and the top node (including);
#	%node2level = ($node_ID => level of the node, ...);
#
sub var_nodelevels_ext_wgt_upgma (\%\@\@\@$$\%\%\%) {

	my ($nodeid2mtrx_indx, $mtrx_dist, $mtrx_wgt, $mtrx_se, $str_na,
		$topnode_id, $node2chi, $node2pa, $br_attr) = @_;

=for DEBUGGING the subroutine 'var_nodelevels_ext_wgt_upgma'
	print STDERR "\n< Subroutine 'var_nodelevels_ext_wgt_upgma' >\n\n";
=cut
		 # Check decency of the inputs #

	my $ct_seqs = @{$mtrx_dist};
	unless ($ct_seqs>0) {
	    print STDERR "\n[ERROR 'var_nodelevels_ext_wgt_upgma'] NO INPUT SEQUENCES are available : $! !\n";
	    return (undef, undef, undef);
	}
	unless (($ct_seqs == @{$mtrx_wgt}) and ($ct_seqs == @{$mtrx_se})) {
	    print STDERR "\n[ERROR 'var_nodelevels_ext_wgt_upgma'] CLASHING MATRIX SIZES:  (distances, weights, standatd errors) = ($ct_seqs, ", scalar (@{$mtrx_wgt}), ", ", scalar (@{$mtrx_se}), "): $! !\n";
	    return (undef, undef, undef);
	}
	for (my $i=0; $i<$ct_seqs; $i++) { # The preliminary outer for-loop #
	    for (my $j=0; $j<$i; $j++) { # The preliminary for-loop #
		my $dist = $mtrx_dist->[$i][$j];
		((defined $dist) and ($dist ne $str_na)) or ($mtrx_wgt->[$i][$j] = 0);
	    }			# End of the preliminary inner for-loop #
	}			# End of the preliminary outer for-loop #


		# Estimate variances of the levels in a top-down manner #

	my %node2var_level; # ($node_ID => var(level($node_ID)), ...)
	my %node2depth;
	my %node2level;
	my $curr_depth = 0;
	my @curr_sbjcts = ($topnode_id);

	while (@curr_sbjcts>0) { # The outermost iterative while-loop #

	    my @next_sbjcts = ();
	    while (my $sbjct = shift @curr_sbjcts) { # The while-loop across subject nodes #
		$node2depth{$sbjct} = $curr_depth;

		my $children = $node2chi->{$sbjct};
		unless ((defined $children) and (@{$children}>0)) { # Exterior nodes #
		    $node2var_level{$sbjct} = 0;
		}
		my $level = level_sglnode_upgma ($sbjct, %{$node2chi}, %{$br_attr});
		unless (defined $level) {
		    print STDERR "[ERROR 'var_nodelevels_ext_wgt_upgma'] The subroutine 'level_sglnode_upgma' FAILED on the node '$sbjct': $! !\n";
		    return (undef, undef, undef);
		}

		$node2level{$sbjct} = $level;

		    ## MOVED from below as a DEBUGging measure (on Aug 29 (Fri), 2008)
		(defined $children) and (push @next_sbjcts, @{$children});
	        unless ($level>0) {
		    #  If the level of a node is 0,
		    # usually its variance is also 0.

		    $node2var_level{$sbjct} = 0;
	 	    next;
		}

		   # Initialize the variance of the level of the subject node
		my $var_level = 0;

		    ## MOVED to above as a DEBUGging measure (on Aug 29 (Fri), 2008)
		###push @next_sbjcts, @{$children};
		my $ct_chi = @{$children};
		my @sbcls_extnodes = ();
		foreach my $chi (@{$children}) { # The preliminary foreach-loop across sub-clusters
		    my $sb_extnodes= fetch_ext_offsprings_nr ($chi, %{$node2chi});
		    unless ((defined $sb_extnodes) and (@{$sb_extnodes}>0)) {
			print STDERR "[ERROR 'var_nodelevels_ext_wgt_upgma'] ANOMALOUS INPUT: NO EXTERNAL NODES under the node '$chi': $! !\n";
			return (undef, undef, undef);

		    }
		    push @sbcls_extnodes, $sb_extnodes;
		}		# End of the preliminary foreach-loop across sub-clusters

			# Preliminary double-summation to estimate
			# the total weight for the subject node
		my $tot_wgt = 0;
		for (my $i=0; $i<$ct_chi; $i++) { # Preliminary outer for-loop across the 1st sub-clusters
		    foreach my $extnode1 (@{$sbcls_extnodes[$i]}) { # middle foreach-loop across the 1st exterior nodes
			my $k = $nodeid2mtrx_indx->{$extnode1};
			unless (defined $k) {
			   print STDERR "[ERROR 'var_nodelevels_ext_wgt_upgma'] NO CORRESPONDING MATRIX INDEX for the exterior node '$extnode1' in the sub-cluster under the $i th child '", $children->[$i], "' of the node '$sbjct': $! !\n";
			    return (undef, undef, undef);
			}

			for (my $j=0; $j<$i;  $j++) { # Preliminary inner for-loop across the 2nd sub-clusters
			    foreach my $extnode2 (@{$sbcls_extnodes[$j]}) { # Inner foreach-loop across the 2nd exterior nodes
				my $l = $nodeid2mtrx_indx->{$extnode2};
				unless (defined $l) {
				   print STDERR "[ERROR 'var_nodelevels_ext_wgt_upgma'] NO CORRESPONDING MATRIX INDEX for the exterior node '$extnode2' in the sub-cluster under the $j th child '", $children->[$j], "' of the node '$sbjct': $! !\n";
				    return (undef, undef, undef);
				}
				my $wgt = ($k>$l) ?
				   $mtrx_wgt->[$k][$l] : $mtrx_wgt->[$l][$k] ;
				(defined $wgt) or next;
				($wgt eq $str_na) and next;
				($wgt>0) or next;
				$tot_wgt += $wgt;
			    }				# End of the inner foreach-loop  across the 2nd exterior nodes
			}			# End of the preliminary inner for-loop across the 2nd sub-clusters
		    }				# End of the middle foreach-loop across the 1st exterior nodes
		}			# End of the preliminary outer for-loop across the 1st sub-clusters

		for (my $i=0; $i<$ct_chi; $i++) { # The for-loop across sub-clusters #
		    my @curr_branches = ($children->[$i]);
		    while (@curr_branches>0) { # The middle iterative while-loop #
			my @next_branches = ();
			while (my $curr_branch = shift @curr_branches) { # The while-loop across branches
			    my $attr = $br_attr->{$curr_branch};
			    unless (defined $attr) {
				print STDERR "[ERROR 'var_nodelevels_ext_wgt'] NO ATTRIBUTES are assigned to the branch '$curr_branch' under the node '$sbjct': $! !\n";
				return (undef, undef);
			    }
			    my $br_len = $attr->[$INDX_LENGTH];

			    my $chi_brs = $node2chi->{$curr_branch};
			    my $dwstr_extnodes ; # exterior nodes downstream of the current branch
			    if ((defined $chi_brs) and (@{$chi_brs}>0)) {
				$dwstr_extnodes =
				fetch_ext_offsprings_nr ($curr_branch, %{$node2chi});
				push @next_branches, @{$chi_brs};
			    } else {
				$dwstr_extnodes = [$curr_branch];
			    }

			    my @exnode_pairs = (); # element = [$k, $l, $wgt]

			    foreach my $extnode1 (@{$dwstr_extnodes}) { # foreach-loop across downstream exterior nodes
				my $k = $nodeid2mtrx_indx->{$extnode1};
				##unless (defined $k) {
				   ## print STDERR "[ERROR 'var_nodelevels_ext_wgt_upgma'] NO CORRESPONDING MATRIX INDEX for the exterior node '$extnode1' downstream of the branch '$curr_branch' under the node '$sbjct': $! !\n";
				    ##return (undef, undef);
				##}

				for (my $j=0; $j<$ct_chi; $j++) { # For-loop across sibling sub-clusters #
				    ($j == $i) and next;
				    foreach my $extnode2 (@{$sbcls_extnodes[$j]}) { # For-loop across exterior nodes in a sibling-subcluster
					my $l = $nodeid2mtrx_indx->{$extnode2};
					##unless ($l) {
					  ##  print STDERR "[ERROR 'var_nodelevels_ext_wgt_upgma'] NO CORERSPONDING MATRIX INDEX for the exterior node '$extnode1' in thesibling sub-cluster under the child '$children->[$i]' of the node '$sbjct': $! !\n";
					   ## return (undef, undef);
					##}
					my $wgt = ($k>$l) ? $mtrx_wgt->[$k][$l] : $mtrx_wgt->[$l][$k];
					((defined $wgt) and ($wgt ne $str_na) and ($wgt > 0)) or next;
					push @exnode_pairs, [$k, $l, $wgt];
				    }	# End of the for-loop across exterior nodes in a sibling-subcluster #
				} 			# End of the for-loop acros sibling sub-clusters #
			    }			# End of the foreach-loop across downstream exterior nodes
			    unless (@exnode_pairs>0) {
				# NO contribution from the branch $curr_branch to the level of $sbjct #
				# Such cases can actually occur (and may not be rare) !!
				# BUG FIXED on July 10 (Thu), 2008 
				next;
				##print STDERR "[ERROR 'var_nodelevels_ext_wgt_upgma'] NO PAIRS of EXTERIOR NODES that are under the node '$sbjct' and whose path runs through the branch '$curr_branch': $! !\n";
				##return (undef, undef);
			    }

			    my @std_exnode_pairs = sort {$b->[2] <=> $a->[2] } @exnode_pairs;
			    @exnode_pairs = (); # Empty the original list #

			    my $br_contr = 0; # Contribution from this branch #

			    my $ct_pairs = @std_exnode_pairs;
			    for (my $p=0; $p<$ct_pairs; $p++) { # For-loop across pairs of exterior nodes
				my ($k, $l, $wgt) = @{$std_exnode_pairs[$p]};
				my $dist = ($k>$l) ? $mtrx_dist->[$k][$l] : $mtrx_dist->[$l][$k];
				unless ((defined $dist) and ($dist ne $str_na)) {
				    print STDERR "[ERROR 'var_nodelevels_ext_wgt_upgma'] NO DISTANCES are available for the node index pair \($k, $l\) under the node '$sbjct':  $! !\n";
				     return (undef, undef);
				}
				my $factor = $wgt;
				my $se = ($k>$l) ? $mtrx_se->[$k][$l] : $mtrx_se->[$l][$k];
				##if ((defined $se) and ($se ne $str_na) and ($dist>0)) {
				if ((defined $se) and ($se ne $str_na)) {
				    my $alt_factor = $wgt * $se;
				    ##$alt_factor *= $alt_factor/$dist;
				    $alt_factor *= $alt_factor/2/$level;
				    ($alt_factor > $factor) and ($factor = $alt_factor);
				}
				$br_contr += (1 + 2 * $p) * $factor;
			    }		# End of the for-loop across pairs of exterior nodes
			    $br_contr *= $br_len;
			    $var_level += $br_contr;

			}			# End of the while-loop across branches
			@curr_branches = @next_branches;
		    }			# End of the middle iteratie while-loop #

		}			# End of the for-loop across sub-clusters #

		    ## ADDED accompanying the BUG FIXATION on July 10 (Thu), 2008 ##
		unless ($tot_wgt>0) {
		    print STDERR "[ERROR 'var_nodelevels_ext_wgt_upgma'] NO CONTRUBUTION to the weight for the level of the subject '$sbjct' from its downstream branches !!\n\n";
		    return (undef, undef, undef);
		}

		$var_level /= 4 * $tot_wgt * $tot_wgt;

		$node2var_level{$sbjct} = $var_level;

		### RESTART FROM HERE ###
	    }			# End of the while-loop across subject nodes #

	    @curr_sbjcts = @next_sbjcts;
	    $curr_depth++;
	}		# End of the outermost iterative while-loop #

=for DEBUGGING the subroutine
	print STDERR "\n\n --- Subroutine 'var_nodelevels_ext_wgt_upgma' FINISHED ---\n\n";
=cut

	return (\%node2var_level, \%node2depth, \%node2level);

}		# End of the subroutine 'var_nodelevels_ext_wgt_upgma' #


# approx_var_brlen_ext_wgt_upgma:
#	This subroutine approximately estimates variances of branch lengths for the
#	trees constructed by 'extended_wtd_upgma_tree' of the module 'MyTreeManipu_IV.pm'.
#
# 	METHOD:
#	We make the following assumptions:
#	(i) Contributions that different sites give to variances are independent of each other;
#	(ii) Contributions to different branches are independent of each other;
#	(iii) Contributions from overlapping sites to the same branche are completely coherent to each other;
#	(iv) The sites in different alignments should show a maximum overlap, namely,
#	  the overlap size is exactly the same as the smaller number of effective sites.
#
#	From (i), Var (len(br1))_aln1 is proportional to #{eff_sites}_aln1;
#	From (ii), we can set Cov (len(br1), len(br2)) = 0 if br1 != br2;
#	From (iii) and (iv), it follows
#	 Cov ( wgt(kl)*len(br1), wgt(k'l')*len(br1) ) = Var ( min[wgt(kl), wgt(k'l')]*len(br1) ),
#	where the weight factor wgt(kl) = #{effective sites}
#	in the pairwise alignment between k and l;
#
#	Therefore, under the assumption of ultra-metricity,
#	we get for the variance of the level of the node p,
#	Var (level (p))
#	= Sum_{br under p} Sum_{(kl) passing and under p and passing br} (2* order(kl) -1) *
#			max [ (wgt(kl) *se(dist(kl) )**2 * len(br)/dist(kl), wgt(kl) * len(br) ] /
#		[2 * Sum_{(kl) passing and under p} wgt(kl)]**2 .
#
#	and the variance of the branch between the nodes p and ch is:
#	Var (level(p) - level (ch)) = Var (level (p) ) + Var (level (ch) ).
#	Here we set Cov (level (p), level (ch)) = 0, which is probably incorrect.
#	However, because the above covariance cannot be negative under the current assumption,
#	The above formula of the branch always overestimates the variance, thus giving
#	conservative tests for the significant branch lengths.
#
# INPUT: (\%nodeid2mtrx_indx, \@mtrx_dist, \@mtrx_wgt, \@mtrx_se, $str_na,
#		$topnode_id, \%node2chi, \%node2pa, \%br_attr), where
#	%nodeid2mtrx_indx = ($node_ID => $indx for the matrices, ...);
#	$mtrx_dist/wgt/se[$k]->[$l] (0<= $l<$k<$ct_seqs) gives the distance/weight factor/standard error of the distance for the pair of sequences $k and $l;
#	$str_na indicates that the matrix element is not available;
#	%node2chi = ($node_ID => \@child_nodes, ...);
#	%node2pa  = ($node_ID => $parent_ID, ...);
#	%br_attr = ($ID of the node immediately below the branh => \@attributes, ...),
#	with $attribute[$INDX_LENGTH] giving branch length ;
#
# OUTPUT: (\%var_brlen, \%node2var_level, \%node2depth, \%node2level), where
#	%var_brlen = ($ID of the node immediately below the branch => $var_brlen, ..);
#	%node2var_level = ($node_ID => variance of the level of the node, ...);
#	%node2depth = ($node_ID => $depth, ...),
#	with $depth defined as the number of 'ancestor' nodes
#	between it (excluding) and the top node (including);
#	%node2level = ($node_ID => $level, ...).
#
sub approx_var_brlen_ext_wgt_upgma (\%\@\@\@$$\%\%\%) {

	my ($nodeid2mtrx_indx, $mtrx_dist, $mtrx_wgt, $mtrx_se, $str_na,
		$topnode_id, $node2chi, $node2pa, $br_attr) = @_;

=for DEBUGGING the subroutine 'approx_var_brlen_ext_wgt_upgma'
	print STDERR "\n< Subroutine 'approx_var_brlen_ext_wgt_upgma' >\n\n";
	my $subtree = br_draw_subtree ($topnode_id, %{$node2chi}, %{$nodeid2mtrx_indx}, %{$br_attr}, 1, 1);
	print STDERR "{ Input tree (TOP_NODE = $topnode_id) }\n\n";
	print STDERR ${$subtree}, "\n\n";
	### RESTART ###
=cut

	my ($node2var_level, $node2depth, $node2level)
		= var_nodelevels_ext_wgt_upgma (%{$nodeid2mtrx_indx},
		@{$mtrx_dist}, @{$mtrx_wgt}, @{$mtrx_se}, $str_na,
		$topnode_id, %{$node2chi}, %{$node2pa}, %{$br_attr});

	unless (defined $node2var_level) {
	    print STDERR "\n[ERROR 'approx_var_brlen_ext_wgt_upgma'] The subroutine 'var_nodelevels_ext_wgt_upgma' FAILED : $! !\n";
	    return (undef, undef, undef, undef);
	}

=for DEBUGGING the subroutine 'approx_var_brlen_ext_wgt_upgma'
	print STDERR "\n{ Output of 'var_nodelevels_ext_wgt_upgma' }\n\n";
	print STDERR "\%\{\$node2var_level\} = (\n";
	foreach my $node (sort keys %{$node2var_level}) {
	    print STDERR ' ' x 5 , $node, ' => ', $node2var_level->{$node}, " ,\n";
	}
	print STDERR ' ' x 5, ");\n\n";
	print STDERR "\%\{\$node2depth\} = (\n";
	foreach my $node (sort keys %{$node2depth}) {
	    print STDERR ' ' x5, $node, ' => ', $node2depth->{$node}, " ,\n";
	}
	print STDERR ' ' x 5, ");\n\n";
	print STDERR "\%\{\$node2level\} = (\n";
	foreach my $node (sort keys %{$node2level}) {
	    print STDERR ' ' x 5, $node , '  => ', $node2level->{$node}, " ,\n";
	}
	print STDERR ' ' x 5, ");\n\n";
=cut
	my %var_brlen;

	##foreach my $node (%{$node2pa}) { # The main foreach-loop #
	foreach my $node (%{$node2var_level}) { # The main foreach-loop #
	    my $parent = $node2pa->{$node};
	    (defined $parent) or next;
	    ($node eq $topnode_id) and next; # TOP node should not be assigned with a branch length #
	    my $var_level = $node2var_level->{$node};
	    unless (defined $var_level) {
		print STDERR "[ERROR 'approx_var_brlen_ext_wgt_upgma'] NO VARIANCE of the level for the node '$node': $! !\n";
		return (undef, undef, undef, undef);
	    }
	    my $par_var_level = $node2var_level->{$parent};
	    unless (defined $par_var_level) {
		print STDERR "[ERROR 'approx_var_brlen_ext_wgt_upgma'] NO VARIANCE of the level for the parent '$parent' of the node '$node': $! !\n";
		return (undef, undef, undef, undef);
	    }

		# An approximation that ignores the covariance
		# between the two levels
	    $var_brlen{$node} = $var_level + $par_var_level;

	}			# End of the main foreach-loop #

=for DEBUGGING the subroutine
	print STDERR "\n\n --- Subroutine 'approx_var_brlen_ext_wgt_upgma' FINISHED ---\n\n";
=cut

	return (\%var_brlen, $node2var_level, $node2depth, $node2level);

}		# End of the subroutine 'approx_var_brlen_ext_wgt_upgma' #


# brlen_ext_wtd_upgma:
#	This subroutine assigns branch lengths to a given tree
#	by means of weighted extended UPGMA-like algorithm.
#	It can also handle multi-furcated trees.
#	The branch length is given by difference between
#	the levels of the nodes immediately upstream and downstream of the branch.
#	The level of a node is estimated as:
#	level ($node) = Sum_{(k,l) with LCA= node} wgt(k, l) * dist (k,l) /
#			Sum_{(k,l) with LCA = node) wgt (k,l) .
#	In this subroutine, we estimate the levels 
#	by merging clusters shareing the LCA in a bottom-up manner.
#	Namely, when merging B_1, B_2, ..., B_r into B
#	wgt (A, B) = Sum_{k=1..r} wgt (A, B_k),
#	dist (A, B) = Sum_{k=1..r} wgt (A, B_k) * dist (A, B_k) / wgt (A, B);
#	sum_wgt (LCA(B))  = Sum_{k=1..r} Sum_{l=1..k-1} wgt (B_k, B_l),
#	level (LCA(B)) = 1/2 * Sum_{k=1..r}Sum_{l=1..k-1} wgt(B_k, B_l) * dist (B_k, B_l) / sum_wgt (LCA(B)) .
#
# INPUT: (\%nodeid2mtrx_indx, \@mtrx_dist, \@mtrx_wgt, $str_na, 
#		\%name2nodeid, %node2chi, %node2pa), where
#	%nodeid2mtrx_indx = ($node_ID => $indx for the matrices, ...) usually for the exterior nodes;
#	$mtrx_dist/wgt[$k]->[$l] (0<= $l<$k<$ct_seqs) gives the distance/weight factor for the pair of sequences $k and $l;
#	$str_na indicates that the matrix element is not available;
#	%name2nodeid = ($name => $node_ID, ...) for the exterior nodes & top node;
#	%node2chi = ($node_ID => \@child_nodes, ...);
#	%node2pa  = ($node_ID => $parent_ID, ...);
#
# OUTPUT: (\%br_attr, \%nodeid2sum_wgt), where
#	%br_attr = ($ID of the node immediately below the branch => \@attributes, ...), with
#	$attributes[$INDX_LENGTH] givins the branch length
#	estimated by the weighted extended UPGMA-like algorithm;
#	%nodeid2sum_wgt = ($node_ID => $summ_weight, ...);
#
sub brlen_ext_wgt_upgma (\%\@\@$\%\%\%) {

	my ($nodeid2mtrx_indx, $mtrx_dist, $mtrx_wgt, $str_na,
		$name2nodeid, $node2chi, $node2pa) = @_;

		 # Check decency of the inputs #

	my $ct_seqs = @{$mtrx_dist};
	unless ($ct_seqs>0) {
	    print STDERR "\n[ERROR 'brlen_ext_wgt_upgma'] NO INPUT SEQUENCES are available : $! !\n";
	    return (undef, undef);
	}
	unless ($ct_seqs == @{$mtrx_wgt}) {
	    print STDERR "\n[ERROR 'brlen_ext_wgt_upgma'] CLASHING MATRIX SIZES:  (distances, weights) = ($ct_seqs, ", scalar (@{$mtrx_wgt}), "): $! !\n";
	    return (undef, undef);
	}

=for DEBUGGING the subroutine
	print STDERR "\n< Subroutine 'brlen_ext_wgt_upgma' >\n\n";
=cut

		# Prepare the Inital set of nodes & their depth #

	my $topnode_id = $name2nodeid->{$TOP_NODE};

	my $exnodes = fetch_ext_offsprings_nr ($topnode_id, %{$node2chi});
	my $ct_exnodes = @{$exnodes};
	unless ($ct_exnodes>0) {
	    print STDERR "\n[ERROR 'brlen_ext_wgt_upgma'] NO EXTERNAL NODES are present under the top node '$topnode_id': $! !\n";
	    return (undef, undef);
	}

	my @curr_nodes = sort {$a <=> $b} @{$exnodes}; # current set of subject nodes
	my %nodeid2mtrx_indx;

	my @mtrx_dist = my @mtrx_wgt = ();
	for (my $i=0; $i<$ct_exnodes; $i++) { # The preliminary outer for-loop #
	    my $exnode1 = $curr_nodes[$i];
	    my $mtrx_indx1 = $nodeid2mtrx_indx->{$exnode1};
	    unless (defined $mtrx_indx1) {
		print STDERR "\n[ERROR 'brlen_ext_wgt_upgma'] The external node '$exnode1' has NO corresponding MATRIX INDEX: $! !\n";
		return (undef, undef);
	    }
	    $nodeid2mtrx_indx{$exnode1} = $i;

	    my @row_dist = my @row_wgt = ();
	    for (my $j=0; $j<$i; $j++) { # The preliminary for-loop #
		my $exnode2 = $curr_nodes[$j];
		my $mtrx_indx2 = $nodeid2mtrx_indx->{$exnode2};
		unless (defined $mtrx_indx2) {
		    print STDERR "\n[ERROR 'brlen_ext_wgt_upgma'] The external node '$exnode2' has NO corresponding MATRIX INDEX: $! !\n";
		    return (undef, undef);
		}

		my ($indx1, $indx2) = ($mtrx_indx1 > $mtrx_indx2) ? 
			($mtrx_indx1, $mtrx_indx2) : ($mtrx_indx2, $mtrx_indx1);
		my $dist = $mtrx_dist->[$indx1][$indx2];
		my $wgt = ((defined $dist) and ($dist ne $str_na)) ?
			$mtrx_wgt->[$indx1][$indx2] : 0;
		((defined $wgt) and ($wgt ne $str_na)) or ($wgt = 0);
		push @row_dist, $dist;
		push @row_wgt, $wgt;
	    }			# End of the preliminary inner for-loop #
	    push @mtrx_dist, \@row_dist;
	    push @mtrx_wgt, \@row_wgt;
	}			# End of the preliminary outer for-loop #


	my $nodes_frc_by_depth = fractionate_by_ndepth (%{$node2chi}, %{$name2nodeid});
		# a list of nodes in  ascending order of depth #

	unless (defined $nodes_frc_by_depth) {
	    print STDERR "\n[ERROR 'brlen_ext_wgt_upgma'] 'fractionate_by_ndepth' FAILED : $! !\n";
	    return (undef, undef);
	}

	my $ct_frcs = @{$nodes_frc_by_depth};

=for DEBUGGING the subroutine
	print STDERR "\n{ Initial setting }\n\n";
	print STDERR "( Tree nodes classified by depth )\n\n";
	print STDERR "Depth\tList/of/nodes\n\n";
	for (my $d =0; $d<$ct_frcs; $d++) {
	    print STDERR $d, "\t", join ('/', @{$nodes_frc_by_depth->[$d]}), "\n";
	}
	print STDERR "\n( Node_IDs \[Matrix_index] )\n\n";
	foreach my $nodeid (@curr_nodes) {
	    print STDERR "\t", "$nodeid [", $nodeid2mtrx_indx{$nodeid}, "]";
	}
	print STDERR "\n";
	print STDERR "\n( Distance matrix )\n\n";
	print STDERR join ("\t", '#Node_ID', 'distances'), "\n\n";
	for (my $k=0; $k<$ct_exnodes; $k++) {
	    print STDERR join ("\t", $curr_nodes[$k], @{$mtrx_dist[$k]}), "\n";
	}
	print STDERR "\n( Weight matrix )\n\n";
	print STDERR join ("\t", '#Node_ID', 'weights'), "\n\n";
	for (my $k=0; $k<$ct_exnodes; $k++) {
	    print STDERR join ("\t", $curr_nodes[$k], @{$mtrx_wgt[$k]}), "\n";
	}
=cut

		## Assign levels and weights to nodes in a bottom-up manner ##

	my %nodeid2sum_wgt;
	my %nodeid2level;

		# All exterior nodes have level 0 #
	foreach my $node (@curr_nodes) { $nodeid2level{$node} = 0; }

	my $depth = $ct_frcs;
	while (my $sbset_nodes = pop @{$nodes_frc_by_depth}) { # The outermost while-loop #
	    $depth--;

=for DEBUGGING the subroutine
	print STDERR "\n[ Depth = $depth ]\n\n";
	print STDERR "Initial set of subjects:   ", join (' ', @{$sbset_nodes}), "\n\n";
=cut

	    while (@{$sbset_nodes}>0) { # The outer while-loop #
		my $nodeid = shift @{$sbset_nodes};
		my $parent = $node2pa->{$nodeid};


		# By the time we encounter the top-node,
		# we should already have finished the whole assignment.
		(defined $parent) or last;

		my $sibls = $node2chi->{$parent};
		my $ct_sibls = @{$sibls};

		my $ct_nodes = @curr_nodes;

=for DEBUGGING the subroutine
	print STDERR "{ Siblings:  ", join (', ', @{$sibls}), " }\n\n";
	print STDERR "\n( Node_IDs \[Matrix_index] )\n\n";
	foreach my $nodeid (@curr_nodes) {
	    print STDERR "\t", "$nodeid [", $nodeid2mtrx_indx{$nodeid}, "]";
	}
	print STDERR "\n";
	print STDERR "\n( Distance matrix )\n\n";
	print STDERR join ("\t", '#Node_ID', 'distances'), "\n\n";
	for (my $k=0; $k<$ct_nodes; $k++) {
	    print STDERR join ("\t", $curr_nodes[$k], @{$mtrx_dist[$k]}), "\n";
	}
	print STDERR "\n( Weight matrix )\n\n";
	print STDERR join ("\t", '#Node_ID', 'weights'), "\n\n";
	for (my $k=0; $k<$ct_nodes; $k++) {
	    print STDERR join ("\t", $curr_nodes[$k], @{$mtrx_wgt[$k]}), "\n";
	}
	print STDERR "\n";
=cut
			# calculate the level of the parent node #

		my $sum_wgt = my $wgtd_level = 0;
		my %sbl_indx2nodeid;
		for (my $i=0; $i<$ct_sibls; $i++) { # The 1st middle for-loop #
		    my $sbl_nodeid1 = $sibls->[$i];
		    my $mtrx_indx1 = $nodeid2mtrx_indx{$sbl_nodeid1};
		    unless (defined $mtrx_indx1) {
			print STDERR "[ERROR 'brlen_ext_wtd_upgma'] The 1st sibling node '$sbl_nodeid1' has NO CORRESPONDING MATRIX INDEX: $! !\n";
			return (undef, undef);
		    }
		    $sbl_indx2nodeid{$mtrx_indx1} = $sbl_nodeid1;

		    for (my $j=0; $j<$i; $j++) { # The inner for-loop #
			my $sbl_nodeid2 = $sibls->[$j];
			my $mtrx_indx2 = $nodeid2mtrx_indx{$sbl_nodeid2};
			unless (defined $mtrx_indx1) {
			    print STDERR "[ERROR 'brlen_ext_wtd_upgma'] The 2nd sibling node '$sbl_nodeid2' has NO CORRESPONDING MATRIX INDEX: $! !\n";
			    return (undef, undef);
			}
			my ($indx1, $indx2) = ($mtrx_indx1 > $mtrx_indx2) ? 
			($mtrx_indx1, $mtrx_indx2) : ($mtrx_indx2, $mtrx_indx1);
			my $wgt = $mtrx_wgt[$indx1]->[$indx2];
			my $dist = $mtrx_dist[$indx1]->[$indx2];
			($wgt ==0) and next;
			$sum_wgt += $wgt;
			$wgtd_level += $wgt * $dist;
		    }				# End of the inner for-loop #
		}			# End of the 1st middle for-loop #

		$nodeid2sum_wgt{$parent} = $sum_wgt;
		if ($sum_wgt>0) {
		   $nodeid2level{$parent} = $wgtd_level /2/$sum_wgt ;
		} else {
		   print STDERR "[ERROR 'brlen_ext_wtd_upgma'] NULL VALUE is assigned as the SUMMED WEIGHT of the parent node '$parent' of the node '$nodeid': $! !\n";
		   return (undef, undef);
		}

			# Update the matrices by merging the sibling nodes #

=for DEBUGGING the subroutine
	print STDERR "[ Update matrices ]\n\n";
=cut
		my @new_mtrx_dist = my @new_mtrx_wgt = ();
		my @mgd_row_dist = my @mgd_row_wgt = ();
		my @next_nodes = ();
		##my $ct_nodes = @curr_nodes;

		for (my $i=0; $i<$ct_nodes; $i++) { # The 2nd middle for-loop # 
		    my $nodeid1 = $curr_nodes[$i];
		    my $mtrx_indx1 = $nodeid2mtrx_indx{$nodeid1};
		    unless (defined $mtrx_indx1) {
			print STDERR "[ERROR 'brlen_ext_wtd_upgma'] The 1st current node '$nodeid1' has NO CORERSPONDING MATRIX INDEX: $! !\n";
			return (undef, undef);
		    }
		    (defined $sbl_indx2nodeid{$mtrx_indx1}) and next; # Skip siblings #
		    push @next_nodes, $nodeid1;
=for DEBUGGING the subroutine
	print STDERR "{ NodeID '$nodeid1' (indx: $mtrx_indx1) }\n\n";
=cut

			# unmerged nodes #

		    my @row_dist = my @row_wgt = ();
		    for (my $j=0; $j<$i; $j++) { # The 1st inner for-loop #
			my $nodeid2 = $curr_nodes[$j];
			my $mtrx_indx2 = $nodeid2mtrx_indx{$nodeid2};
			unless (defined $mtrx_indx2) {
			    print STDERR "[ERROR 'brlen_ext_wtd_upgma'] The 2nd current node '$nodeid2' has NO CORRESPONDING MATRIX INDEX: $! !\n";
			    return (undef, undef);
			}
			(defined $sbl_indx2nodeid{$mtrx_indx2}) and next; # Skip siblings;
			my ($indx1, $indx2) = ($mtrx_indx1 > $mtrx_indx2) ?
				($mtrx_indx1, $mtrx_indx2) : ($mtrx_indx2, $mtrx_indx1) ;
			my $wgt = $mtrx_wgt[$indx1]->[$indx2];
			my $dist = $mtrx_dist[$indx1]->[$indx2];
			push @row_wgt, $wgt;
			push @row_dist, $dist;
		    }			# End of the 1st inner for-loop #
		    push @new_mtrx_dist, \@row_dist;
		    push @new_mtrx_wgt, \@row_wgt;

			# merger of the sibling nodes #

=for DEBUGGING the subroutine
	print STDERR "{ vs Merged  nodes }\n\n";
=cut
		    my $mgd_wgt = my $mgd_wtd_dist = 0;
		    foreach my $sbl_indx (sort {$a <=> $b} keys %sbl_indx2nodeid) { # The 2nd inner foreach-loop #
			my ($indx1, $indx2) = ($sbl_indx > $mtrx_indx1) ?
			 ($sbl_indx, $mtrx_indx1) : ($mtrx_indx1, $sbl_indx) ; 
			my $wgt = $mtrx_wgt[$indx1]->[$indx2];
			($wgt>0) or next;
			$mgd_wgt += $wgt;
			my $dist = $mtrx_dist[$indx1]->[$indx2];

				# Bug FIXED on July 6 (Sun), 2008 #
			$mgd_wtd_dist += $wgt * $dist ;
			##$mgd_wtd_dist = $wgt * $dist ;
=for DEBUGGING the subroutine
	my $sbl_nodeid = $sbl_indx2nodeid{$sbl_indx};
	print STDERR "( vs $sbl_nodeid (indx: $sbl_indx) )  weight = $wgt,  distance = $dist ...\n";
=cut
		    }				# End of the 2nd inner foreach-loop #
=for DEBUGGING the subroutine
	print STDERR "( Total summ of $nodeid1 vs siblings )  weight = $mgd_wgt,  distance = $mgd_wtd_dist .\n\n";
=cut
		    if ($mgd_wgt>0) {
			push @mgd_row_wgt, $mgd_wgt;
			push @mgd_row_dist, $mgd_wtd_dist/$mgd_wgt;

		    } else { # NO link is found between the siblings and the 1st node #
			push @mgd_row_wgt, 0;
			push @mgd_row_dist, undef;
		    }
		}			# End of the 2nd middle for-loop #

		push @new_mtrx_wgt, \@mgd_row_wgt;
		push @new_mtrx_dist, \@mgd_row_dist;
		push @next_nodes, $parent;

		@mtrx_wgt = @new_mtrx_wgt;
		@mtrx_dist = @new_mtrx_dist;
		@curr_nodes = @next_nodes;
		$ct_nodes = @curr_nodes;
		%nodeid2mtrx_indx = ();
		for (my $k=0; $k<$ct_nodes; $k++) { $nodeid2mtrx_indx{$curr_nodes[$k]} = $k; }

			# remove the siblings from the set of remaining nodes

		(my $shared, my $ex_sbl, $sbset_nodes) = diff_array (@{$sibls}, @{$sbset_nodes});

=for DEBUGGING the subroutine
	print STDERR "\n --- The round for the siblings (", join (', ', @{$sibls}), ") ENDED ...\n\n";
	print STDERR "\n\nRemaining Subjects:  ", join ('  ', @{$sbset_nodes}), "\n\n";
	print STDERR "( Remaining Nodes [Matrx_indx] )\n\n";
	foreach my $nodeid (@curr_nodes) { print STDERR "\t", "$nodeid [", $nodeid2mtrx_indx{$nodeid}, "]"; }
	print STDERR "\n";
	print STDERR "\n( Updated Distance matrix )\n\n";
	print STDERR join ("\t", '#Node_ID', 'distances'), "\n\n";
	for (my $k=0; $k<$ct_nodes; $k++) {
	    print STDERR join ("\t", $curr_nodes[$k], @{$mtrx_dist[$k]}), "\n";
	}
	print STDERR "\n( Updated Weight matrix )\n\n";
	print STDERR join ("\t", '#Node_ID', 'weights'), "\n\n";
	for (my $k=0; $k<$ct_nodes; $k++) {
	    print STDERR join ("\t", $curr_nodes[$k], @{$mtrx_wgt[$k]}), "\n";
	}
	print STDERR "\n";
=cut

	    }		# End of the 1st outer while-loop #

	}			# End of the outermost while-loop #

=for DEBUGGING the subroutine
	print STDERR "\n\n --- ALL the iterative rounds ENDED !! ...\n\n";
	print STDERR join ("\t", '#Node_ID', 'Level'), "\n\n";
	foreach my $nodeid (sort {$a <=> $b} keys %nodeid2level) {
	    print STDERR join ("\t", $nodeid, $nodeid2level{$nodeid}), "\n";
	}
	print STDERR "\n\n";
=cut

	my %br_attr;

		## BUG FIXED on July 10 (Thu), 2008 ##
	foreach my $nodeid (keys %nodeid2level) { # The trailing foreach-loop #
	##foreach my $nodeid (keys %{$node2pa}) { # The trailing foreach-loop #
	    my $parent = $node2pa->{$nodeid};
	    (defined $parent) or next;
	    my $level = $nodeid2level{$nodeid};

		## BUG FIXED on July 10 (Thu), 2008 ##
	    (defined $level) or next; # level = 0 should also be assigned correctly
	    ##(defined $level) or ($level = 0);
	    my $pa_level = $nodeid2level{$parent};
	    unless (defined $pa_level) {
		print STDERR "[ERROR 'brlen_ext_wgt_upgma'] NO LEVE is assigned to the parent '$parent' of the node '$nodeid': $! !\n";
		return (undef, undef);
	    }
	    my @attr = ();
	    $attr[$INDX_LENGTH] = $pa_level - $level;
	    $br_attr{$nodeid} = \@attr;
	}			# End of the trailing foreach-loop #

=for DEBUGGING the subroutine
	print STDERR "\n\n --- 'brlen_ext_wgt_upgma' FINISHED !! ---\n\n";
=cut
	return (\%br_attr, \%nodeid2sum_wgt);

}		# End of the subroutine 'brlen_ext_wgt_upgma' #


# apprx_bootspp_by_z_test:
#	This subroutine approximates bootstrap support values
#	by applying the Z-test on each branch length.
#	More precisely, it assumes that Z = (branch length} / {standard error}
#	follows the normal distribution, and approximates the bootstrap support
#	with int (#{iterations} * P(Z>0)), where P(Z>0)  is the probability that
#	that Z is positive under the normal distribution of mean 0 and variance 1.
#
# INPUT: (\%node2chi, \%br_attr, \%var_brlen, $indx_var, $fl_stderr, $ct_boot_ite), where
#	%node2chi = ($nodeid => \@children,...) ;
#	%br_attr = ($ID of the node immediately below the branch => \@attr, ...), with
#	$attr[$INDX_LENGTH] giving the branch length;
#	%var_brlen = ($ID of the node immediately below the branch => $var, ...), with
#	variance = $var->[$indx_var] (for $indx_var >=0) or $var (for $indx_var < 0);
#	If $fl_stderr = 1, the "variance" above is actually the standard error;
#	$ct_boot_ite is the count of the virtual bootstrap iterations.
#	When $ct_boot_ite = 1, the subroutine calculates the probability itself.
#
# OUTPUT: \%apprx_boot, where
#	%apprx_boot = ($ID of the node immediately below the branch => {approximate bootstrap value}, ...) ;
#
sub apprx_bootspp_by_z_test (\%\%\%$$$) {

	my ($node2chi, $br_attr, $var_brlen, $indx_var, $fl_stderr, $ct_boot_ite) = @_;

	my %apprx_boot;

	foreach my $nodeid (keys %{$var_brlen}) { # The main foreach-loop #

	    my $chi = $node2chi->{$nodeid};
	    ((defined $chi) and (@{$chi}>0)) or next; # Skip exterior nodes
	    my $var = $var_brlen->{$nodeid};
	    ($indx_var >=0) and ($var = $var->[$indx_var]);

	    my $attr = $br_attr->{$nodeid};
	    my $br_len = $attr->[$INDX_LENGTH];
	    unless (defined $br_len) {
		print STDERR "[CAUTION 'apprx_bootspp_by_z_test'] NO BRANCH LENGTH is assigned to the node '$nodeid' !\n";
		next;
	    }
	    if ($br_len == 0) {
		$apprx_boot{$nodeid} = $ct_boot_ite/2;
		next;
	    }
	    unless ($var >=0) {
		print STDERR "[CAUTION 'apprx_bootspp_by_z_test'] NEGATIVE VALUE ($var) of the ", (($fl_stderr > 0) ? "standard error" : "variance" ), " for the node '$nodeid' : $! !\n";
		next;
	    }

	    my $se = $fl_stderr ? $var : sqrt ($var);

	    my $z = $br_len / $se ;

	    my $p = cum_gaussian ($z, 0, 1);
	    my $app_boot = $p * $ct_boot_ite;
	    ($ct_boot_ite == 1) or ($app_boot = int ($app_boot)) ;

	    $apprx_boot{$nodeid} = $app_boot;
	}				# End of the main foreach-loop #

	return \%apprx_boot;

}		# End of the subroutine 'apprx_bootspp_by_z_test' #


1;	# a "spell" which makes the module "use"-able and/or "require"-able #


